/*

    File: CModel.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>

#include "cmodel.h"

void CModel::Reset(void)
{
	memset(this,0,sizeof(CModel));
}

void CModel::Unload(void)
{
	if (vtx) free(vtx);
	if (plane) free(plane);
	if (edge) free(edge);
	if (tri) free(tri);

	Reset();
}


#include <stdio.h>
#include "../m_system3d/error.h"
#include "../m_foundation/const.h"

bool CModel::Load(char *filename)
{
	int file_vtxCount;
	int file_triCount;
	int t,v;

	//CUSTOM MDL FILE FORMAT

	Unload();

	strcpy(title, filename);

	char sig[4];

	FILE *File = fopen(filename, "rb");
	if (File == NULL)
	{
		Error("%s: unable to open model file", filename);
		return false;
	}

	fread(sig, 1, 4, File);
	if (memcmp(sig, "MMDL", 4))
	{
		Error("%s: bad signature on file", filename);
		return false;
	}
	fread(&file_vtxCount, sizeof(int), 1, File);
	fread(&file_triCount, sizeof(int), 1, File);

	if (file_vtxCount < 0)
	{
		Error("%s: cmodel has bad vertex count: %d", filename, file_vtxCount);
		fclose(File);
		return false;
	}
	if (file_triCount < 0)
	{
		Error("%s: cmodel has bad plane count: %d", filename, file_triCount);
		fclose(File);
		return false;
	}

	//allocate verticies
	if ((vtx = (vector_t *)malloc(sizeof(vector_t) * file_vtxCount))==NULL)
	{
		Error("%s: unable to allocate vertex array", filename);
		fclose(File);
		return false;
	}
	//read verticies
	if (fread(vtx, sizeof(vector_t), file_vtxCount, File) != (size_t)(file_vtxCount))
	{
		Error("%s: unable to read vertex array", filename);
		free(vtx);	vtx = NULL;
		fclose(File);
		return false;
	}
	//skip texture coordinates
	if (fseek(File, sizeof(float)*file_vtxCount*2, SEEK_CUR))
	{
		Error("%s: unable to skip texture coordinate array", filename);
		free(vtx);	vtx = NULL;
		fclose(File);
		return false;
	}
	//allocate triangles
	if ((tri = (int *)malloc(sizeof(int) * 3 * file_triCount))==NULL)
	{
		Error("%s: unable to allocate tri array", filename);
		free(vtx);	vtx = NULL;
		fclose(File);
		return false;
	}
	//read triangles
	if (fread(tri, sizeof(int)*3, file_triCount, File) != (size_t)(file_triCount))
	{
		Error("%s: unable to read tri array", filename);
		free(vtx);	vtx = NULL;
		free(tri);	tri = NULL;
		fclose(File);
		return false;
	}

	fclose(File);

	//calculate volume based on tris and vtxs
	volume = 0;
	for (t = 0; t < file_triCount*3; t+=3)
	{
		vector_t *pv[3] = { &vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]] };

		volume += pv[0]->x * pv[1]->y * pv[2]->z;
		volume += pv[0]->y * pv[1]->z * pv[2]->x;
		volume += pv[0]->z * pv[1]->x * pv[2]->y;
		volume -= pv[2]->x * pv[1]->y * pv[0]->z;
		volume -= pv[2]->y * pv[1]->z * pv[0]->x;
		volume -= pv[2]->z * pv[1]->x * pv[0]->y;
	}
	volume /= 6.f;
	if (volume < 0) volume = -volume;

	//calculate axial moment of inertia based on tris and vtxs
	axisMoment = vectorZero;
	for (t = 0; t < file_triCount*3; t+=3)
	{
		vector_t *pv[3] = { &vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]] };

		scalar_t det = 0;
		det += pv[0]->x * pv[1]->y * pv[2]->z;
		det += pv[0]->y * pv[1]->z * pv[2]->x;
		det += pv[0]->z * pv[1]->x * pv[2]->y;
		det -= pv[2]->x * pv[1]->y * pv[0]->z;
		det -= pv[2]->y * pv[1]->z * pv[0]->x;
		det -= pv[2]->z * pv[1]->x * pv[0]->y;

		scalar_t v0term =	pv[0]->x * pv[0]->x + pv[0]->x * pv[1]->x + pv[0]->x * pv[2]->x +
							pv[1]->x * pv[1]->x + pv[1]->x * pv[2]->x + pv[2]->x * pv[2]->x;
		scalar_t v1term =	pv[0]->y * pv[0]->y + pv[0]->y * pv[1]->y + pv[0]->y * pv[2]->y +
							pv[1]->y * pv[1]->y + pv[1]->y * pv[2]->y + pv[2]->y * pv[2]->y;
		scalar_t v2term =	pv[0]->z * pv[0]->z + pv[0]->z * pv[1]->z + pv[0]->z * pv[2]->z +
							pv[1]->z * pv[1]->z + pv[1]->z * pv[2]->z + pv[2]->z * pv[2]->z;

		axisMoment.x += det * v0term;
		axisMoment.y += det * v1term;
		axisMoment.z += det * v2term;
	}

	//NOTICE - inerta = Vector(
		//axisMoment.y + axisMoment.z,
		//axisMoment.x + axisMoment.z,
		//axisMoment.x + axisMoment.y);

	//alloc planes
	if ((plane = (plane_t *)malloc(sizeof(plane_t) * file_triCount))==NULL)
	{
		Error("%s: unable to allocate plane array", filename);
		free(tri);		tri = NULL;
		free(vtx);		vtx = NULL;
		fclose(File);
		return false;
	}
	for (v = 0; v < file_triCount; v++)
	{
		t = v*3;
		//VectorCopy(&vtx[tri[t]], &plane[v].pos);
		VectorPlaneUnitNormal(&vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]], &plane[v].normal);
		plane[v].dist_neg = -VectorDot(&plane[v].normal, &vtx[tri[t]]);
	}

	//calc bbox
	VectorSet(&bbox.min, fInf,fInf,fInf);
	VectorSet(&bbox.max, -fInf,-fInf,-fInf);
	//the max radius will always occur on a vertex of a convex model
	max_radius = -1;
	for (v = 0; v < file_vtxCount; v++)
	{
		//stretch bbox
		BoxStretch(&bbox, &vtx[v]);

		//stretch radius
		float l = VectorLength(&vtx[v]);
		if (l > max_radius) max_radius = l;
	}
	//the min radius will always occur on a plane of a convex model
	min_radius = fInf;
	for (v = 0; v < file_triCount; v++)
	{
		float l = PlaneVectorDistance(&plane[v], &vectorZero);
		if (l < min_radius) min_radius = l;
	}

	//calc edgeCount
	edgeCount = 0;
	for (int a = 0; a < file_triCount-1; a++)
	{
		for (int b = a+1; b < file_triCount; b++)
		{
			int count = 0;

			//if a and b share two verticies - create an edge
			for (v = 0; v < 3; v++)	//verticies of tri a
			{
				for (t = 0; t < 3; t++)	//verticies of tri b
				{
					if (tri[3*a+v] == tri[3*b+t]) count++;
				}
			}

			if (count == 2) edgeCount++;
		}
	}
	//allocate edge buffer
	if ((edge = (edge_t *)malloc(edgeCount * sizeof(edge_t)))==NULL)
	{
		Error("%s: unable to allocate edge array", filename);
		free(vtx);		vtx = NULL;
		free(tri);		tri = NULL;
		free(plane);	plane = NULL;
		return false;
	}
	//create edges
	int l = 0;
	for (a = 0; a < file_triCount-1; a++)
	{
		for (int b = a+1; b < file_triCount; b++)
		{
			edge_t e;
			int count = 0;

			//if a and b share two verticies - create an edge
			for (v = 0; v < 3; v++)	//verticies of tri a
			{
				for (t = 0; t < 3; t++)	//verticies of tri b
				{
					if (tri[3*a+v] == tri[3*b+t])
					{
						if (count < 2)
						{
							//store triangles
							e.t[0] = a;
							e.t[1] = b;
							//store edge
							e.v[count] = tri[3*a+v];
						}
						count++;
					}
				}
			}

			if (count == 2)
			{
				memcpy(&edge[l], &e, sizeof(edge_t));
				l++;
			}
		}
	}

	vtxCount = file_vtxCount;
	planeCount = file_triCount;
	triCount = file_triCount;

	return true;
}

void CModel::CalcInertia(const vector_t *scale, vector_t *vres)
{
	scalar_t scale_prod;
	vector_t scale_sq, inertia;

	scale_prod = scale->x * scale->y * scale->z;

	VectorSet(&scale_sq,
		scale->x * scale->x,
		scale->y * scale->y,
		scale->z * scale->z);

	//calculate inertia based on tris and vtxs
	VectorCopy(&vectorZero, &inertia);

	for (int t = 0; t < triCount*3; t+=3)
	{
		vector_t *pv[3] = { &vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]] };

/*
[ || || || ]
[          ]
[ v1 v2 v3 ]
[          ]
[ || || || ]

[ x1 x2 x3 ]
[          ]
[ y1 y2 y3 ]
[          ]
[ z1 z2 z3 ]
*/

		scalar_t det = 0;
		det += pv[0]->x * pv[1]->y * pv[2]->z;
		det += pv[0]->y * pv[1]->z * pv[2]->x;
		det += pv[0]->z * pv[1]->x * pv[2]->y;
		det -= pv[2]->x * pv[1]->y * pv[0]->z;
		det -= pv[2]->y * pv[1]->z * pv[0]->x;
		det -= pv[2]->z * pv[1]->x * pv[0]->y;
		if (det < 0) det = -det;

		scalar_t v0term =
			pv[0]->x * pv[0]->x +
			pv[0]->x * pv[1]->x +
			pv[0]->x * pv[2]->x +
			pv[1]->x * pv[1]->x +
			pv[1]->x * pv[2]->x +
			pv[2]->x * pv[2]->x;

		scalar_t v1term =
			pv[0]->y * pv[0]->y +
			pv[0]->y * pv[1]->y +
			pv[0]->y * pv[2]->y +
			pv[1]->y * pv[1]->y +
			pv[1]->y * pv[2]->y +
			pv[2]->y * pv[2]->y;

		scalar_t v2term =
			pv[0]->z * pv[0]->z +
			pv[0]->z * pv[1]->z +
			pv[0]->z * pv[2]->z +
			pv[1]->z * pv[1]->z +
			pv[1]->z * pv[2]->z +
			pv[2]->z * pv[2]->z;

		inertia.x += det * (v1term * scale_sq.y + v2term * scale_sq.z);
		inertia.y += det * (v0term * scale_sq.x + v2term * scale_sq.z);
		inertia.z += det * (v0term * scale_sq.x + v1term * scale_sq.y);
	}

	VectorScale(&inertia, (scale->x * scale->y * scale->z / 60.f), vres);
}


CModel *GetCModel(LinkedList *p, char *title)
{
	for (CModel *m = (CModel *)(p->head); m; m = (CModel *)(m->chain_next))
		if (!_stricmp(title, m->title)) break;

	return m;
}