/*

    File: Collision.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "../m_math3d/vector.h"
#include "../m_math3d/geometry.h"

#include "../m_foundation/const.h"
#include "../m_foundation/main.h"

#include "entitytree.h"
#include "collision.h"
#include "entity.h"
#include "world.h"

#include "main.h"		//debug

#define C_MODEL_COUNT		2

#define C_MODEL_VERTEX_MAX	100
#define C_MODEL_PLANE_MAX	100

static vector_t		cmodel_position[C_MODEL_COUNT];

static int			cmodel_vtxCount[C_MODEL_COUNT];
static vector_t		cmodel_vtx[C_MODEL_COUNT][C_MODEL_VERTEX_MAX];	//2400 bytes of collision verticies
static int			cmodel_planeCount[C_MODEL_COUNT];
static plane_t		cmodel_plane[C_MODEL_COUNT][C_MODEL_PLANE_MAX];	//4800 bytes of collision planes
static int			cmodel_edgeCount[C_MODEL_COUNT];
static edge_t		*cmodel_edge[C_MODEL_COUNT];

void LoadCollisionModel(
	const CModel *c_model,
	const vector_t *position,
	const quat_t *angle,
	const vector_t *scale,
	int cid)
{
	int v;

	//make sure its a valid collision model ID
	if (cid < 0 || cid >= C_MODEL_COUNT) return;

	//if no collision model data - bail out on the extensive tests
	if (!c_model) return;

	cmodel_vtxCount[cid] = c_model->vtxCount;
	if (cmodel_vtxCount[cid] > C_MODEL_VERTEX_MAX)
	{
		main_console->Print("Encountered collision model %s with vertex count %d greater than vertex buffer\n", c_model->title, c_model->vtxCount);
		return;
	}

	cmodel_planeCount[cid] = c_model->planeCount;
	if (cmodel_planeCount[cid] > C_MODEL_PLANE_MAX)
	{
		main_console->Print("Encountered collision model %s with plane count %d greater than plane buffer\n", c_model->title, c_model->planeCount);
		return;
	}

	cmodel_edgeCount[cid] = c_model->edgeCount;
	cmodel_edge[cid] = c_model->edge;

	//calculate verticies
	for (v = 0; v < cmodel_vtxCount[cid]; v++)
	{
		static vector_t *vtx;
		vtx = &cmodel_vtx[cid][v];
		//scale
		vtx->x = c_model->vtx[v].x * scale->x;
		vtx->y = c_model->vtx[v].y * scale->y;
		vtx->z = c_model->vtx[v].z * scale->z;
		//rotate
		QuatRotateVector(vtx, angle, vtx);
		//translate
		VectorAdd(vtx, position, vtx);
	}

	//calculate planes... (this is gonna be nasty)
	for (v = 0; v < cmodel_planeCount[cid]; v++)
	{
		static plane_t *plane;
		plane = &cmodel_plane[cid][v];

		//scale
		plane->normal.x = c_model->plane[v].normal.x * scale->x;
		plane->normal.y = c_model->plane[v].normal.y * scale->y;
		plane->normal.z = c_model->plane[v].normal.z * scale->z;
		scalar_t mag = VectorLength(&plane->normal);
		if (mag)
		{
			plane->dist_neg = c_model->plane[v].dist_neg * mag;
			mag = 1.f / mag;
			plane->normal.x *= mag;
			plane->normal.y *= mag;
			plane->normal.z *= mag;
		}
		else
		{
			plane->dist_neg = 0;
		}

		//rotate
		QuatRotateVector(&plane->normal, angle, &plane->normal);

		//translate
		plane->dist_neg -= VectorDot(&plane->normal, position);
	}

	//store position
	VectorCopy(position, &cmodel_position[cid]);


//render it all to be cool
	static int Collision_RenderCModel = 0;
	static CVarw *fdjklsdf = new CVarw(&Collision_RenderCModel, "Collision_RenderCModel");
	if (Collision_RenderCModel)
	{
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glDisable(GL_TEXTURE);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

		glColor3f(1,1,0);
		glBegin(GL_TRIANGLES);
		for (int i = 0; i < c_model->triCount*3; i++)
			glVertex3fv(cmodel_vtx[cid][c_model->tri[i]].v);
		glEnd();

		vector_t src;
		scalar_t len;
		glColor3f(1,0,1);
		glBegin(GL_LINES);
		for (i = 0; i < c_model->planeCount; i++)
		{
			VectorSub(&cmodel_vtx[cid][c_model->tri[3*i]], &cmodel_vtx[cid][c_model->tri[3*i+1]], &src);
			len = VectorLength(&src);

			VectorAdd(&cmodel_vtx[cid][c_model->tri[3*i]], &cmodel_vtx[cid][c_model->tri[3*i+1]], &src);
			VectorAdd(&cmodel_vtx[cid][c_model->tri[3*i+2]], &src, &src);
			VectorScale(&src, 1.f / 3.f, &src);

			glVertex3fv(src.v);

			VectorMultAdd(&src, &cmodel_plane[cid][i].normal, len, &src);

			glVertex3fv(src.v);
		}
		glEnd();

		glPopAttrib();
	}
}


int WitnessSeparates(const plane_t *witness)
{

	int v;
	int side;

	//store the first - we need a base case
	side = PlaneVectorDistance(witness, &cmodel_vtx[0][0]) > 0;
	
	//and test the rest of the vtxs
	for (v = 1; v < cmodel_vtxCount[0]; v++)
	{
		//if this vtx is on the opposite side as any previously calculated
		if (side != (PlaneVectorDistance(witness, &cmodel_vtx[0][v]) > 0))
			return 0;
	}

	for (v = 0; v < cmodel_vtxCount[1]; v++)
	{
		//if this vtx is on the same side as any of the first model's vertexes
		if (side == (PlaneVectorDistance(witness, &cmodel_vtx[1][v]) > 0))
			return 0;
	}

	//the witness worked
	return 1;
}




static scalar_t vertex_bestdepth[C_MODEL_VERTEX_MAX];
static int vertex_bestplane[C_MODEL_VERTEX_MAX];

int ObjectCollision(vector_t *res_point, vector_t *res_normal, plane_t *witness)
{
	plane_t *p0 =	cmodel_plane[C_MODEL_OBJ0];
	int p0count =	cmodel_planeCount[C_MODEL_OBJ0];
	vector_t *v0 =	cmodel_vtx[C_MODEL_OBJ0];
	int v0count =	cmodel_vtxCount[C_MODEL_OBJ0];
	edge_t *e0 =	cmodel_edge[C_MODEL_OBJ0];
	int e0count =	cmodel_edgeCount[C_MODEL_OBJ0];

	plane_t *p1 =	cmodel_plane[C_MODEL_OBJ1];
	int p1count =	cmodel_planeCount[C_MODEL_OBJ1];
	vector_t *v1 =	cmodel_vtx[C_MODEL_OBJ1];
	int v1count =	cmodel_vtxCount[C_MODEL_OBJ1];
	edge_t *e1 =	cmodel_edge[C_MODEL_OBJ1];
	int e1count =	cmodel_edgeCount[C_MODEL_OBJ1];

	static const scalar_t planeEpsilon = 1.f;

	int p,v;		//indexes
	scalar_t dist;
	bool allOutside;

	//best potential vertex information
		//have this just above zero so we can toss all non-colliding verticies on the plane test
	scalar_t best_depth = planeEpsilon;
	int bestv_index = -1,
		bestv_obj = -1,			//res_point = cmodel_vtx[bestv_obj][bestv_index]
		bestp_index = -1;		//res_normal = cmodel_plane[!bestv_obj][bestp_index]

	vector_t	o_delta;

	//whether we have collided or not
	bool collided = false;
	//whether it was a vertex/face collision or an edge/edge collision
	bool vertex_face;

	//calc delta from o1 to o0
	VectorSub(&cmodel_position[0], &cmodel_position[1], &o_delta);		//from o1 to o0

//init depth array
	for (v = 0; v < v1count; v++)
	{
		vertex_bestdepth[v] = -fInf;
		//dont worry about vertex_bestplane[]
		//it will be overwritten upon first iteration
	}

//test P0 to V1
	for (p = 0; p < p0count; p++)
	{
		allOutside = true;

		for (v = 0; v < v1count; v++)
		{
			dist = PlaneVectorDistance(&p0[p], &v1[v]);
			if (dist < 0) allOutside = false;
			if (dist > vertex_bestdepth[v])
			{
				vertex_bestdepth[v] = dist;
				vertex_bestplane[v] = p;
			}
		}
		//separating plane - BAM!
		if (allOutside)
		{
			//store 'p' as a witness
			PlaneCopy(&p0[p], witness);
			//and bail out
			return 0;
		}
	}
//store deepest vertex of V1
	for (v = 0; v < v1count; v++)
	{
		//test to best dist
		if (vertex_bestdepth[v] < best_depth)
		{
			best_depth = vertex_bestdepth[v];
			bestv_index = v;
			bestv_obj = 1;		//1 = V1
			bestp_index = vertex_bestplane[v];
			collided = true;
			vertex_face = true;
		}
		//and cleanup after we process this vertex - for the next round
		vertex_bestdepth[v] = -fInf;
	}

//test P1 to V0
	for (p = 0; p < p1count; p++)
	{
		allOutside = true;

		for (v = 0; v < v0count; v++)
		{
			dist = PlaneVectorDistance(&p1[p], &v0[v]);
			if (dist < 0) allOutside = false;
			if (dist > vertex_bestdepth[v])
			{
				vertex_bestdepth[v] = dist;
				vertex_bestplane[v] = p;
			}
		}
		//separating plane - BAM!
		if (allOutside)
		{
			//store 'p' as a witness
			PlaneCopy(&p1[p], witness);
			//and bail out
			return 0;
		}
	}
//store deepest vertex of V0
	for (v = 0; v < v0count; v++)
	{
		if (vertex_bestdepth[v] < best_depth)
		{
			best_depth = vertex_bestdepth[v];
			bestv_index = v;
			bestv_obj = 0;		//0 = V0
			bestp_index = p;
			collided = true;
			vertex_face = true;
		}
	}

#ifdef DONT_DO_EDGE_EDGE
	if (!collided) return 0;

	VectorCopy(&cmodel_vtx[bestv_obj][bestv_index], res_point);
	VectorCopy(&cmodel_plane[!bestv_obj][bestp_index].normal, res_normal);
	if (bestv_obj)		//therefore best plane obj == !bestv_obj == 0, and we need a flip
		VectorScale(res_normal, -1, res_normal);

	return 1;

#else

	int ea,eb;
	vector_t	best_point;
	vector_t	best_normal;

	//if a vertex/face collision occured
	if (collided)
	{
		VectorCopy(&cmodel_vtx[bestv_obj][bestv_index], &best_point);
		VectorCopy(&cmodel_plane[!bestv_obj][bestp_index].normal, &best_normal);
		if (bestv_obj)		//therefore best plane obj == !bestv_obj == 0, and we need a flip
			VectorScale(&best_normal, -1, &best_normal);
	}

	//test object 0 edge to object 1 edge
	for (ea = 0; ea < e0count; ea++)
	{
		for (eb = 0; eb < e1count; eb++)
		{
			line_t del0, del1;
			scalar_t param0, param1;
			vector_t vec0, vec1, cross;

			VectorCopy(&v0[e0[ea].v[0]], &del0.pos);
			VectorSub(&v0[e0[ea].v[1]], &v0[e0[ea].v[0]], &del0.dir);

			VectorCopy(&v1[e1[eb].v[0]], &del1.pos);
			VectorSub(&v1[e1[eb].v[1]], &v1[e1[eb].v[0]], &del1.dir);

			LineLineClosestPointParameter(&del0, &del1, &param0, &param1);

			if (param0 < 0 || param0 > 1 || param1 < 0 || param1 > 1)
				continue;

			VectorMultAdd(&del0.pos, &del0.dir, param0, &vec0);
			VectorMultAdd(&del1.pos, &del1.dir, param1, &vec1);

			VectorCross(&del0.dir, &del1.dir, &cross);
			if (VectorLengthSq(&cross) < 0.000000001) continue;	//assume parallel
			VectorUnit(&cross, &cross);

			scalar_t depth0 = fInf;
			scalar_t depth1 = fInf;

			//only need to test depth to the two planes that make up the edge
/*
int CModel_PointSurfaceDepth(int cid, const vector_t *vec, scalar_t *bestdist)
{
	//make sure its a valid collision model ID
	if (cid < 0 || cid >= C_MODEL_COUNT) return -1;

	*bestdist = -fInf;

	int bestplane = -1;

	for (int p = 0; p < cmodel_planeCount[cid]; p++)
	{
		scalar_t dist = PlaneVectorDistance(&cmodel_plane[cid][p], vec);
		if (dist > *bestdist)
		{
			*bestdist = dist;
			bestplane = p;
		}
	}

	return bestplane;
}
*/
			int plane0 = -1, plane1 = -1;
//			int plane0 = CModel_PointSurfaceDepth(C_MODEL_OBJ0, &vec1, &depth0);
			{
				depth0 = -fInf;

				for (int p = 0; p < cmodel_planeCount[C_MODEL_OBJ0]; p++)
				{
					scalar_t dist = PlaneVectorDistance(&cmodel_plane[C_MODEL_OBJ0][p], &vec1);
					if (dist > depth0)
					{
						depth0 = dist;
						plane0 = p;
					}
				}
			}
//			int plane1 = CModel_PointSurfaceDepth(C_MODEL_OBJ1, &vec0, &depth1);
			{
				depth1 = -fInf;

				for (p = 0; p < cmodel_planeCount[C_MODEL_OBJ1]; p++)
				{
					scalar_t dist = PlaneVectorDistance(&cmodel_plane[C_MODEL_OBJ1][p], &vec0);
					if (dist > depth0)
					{
						depth1 = dist;
						plane1 = p;
					}
				}
			}

			if (plane0 == -1 || plane1 == -1) continue;		//shouldnt ever happen

			//should this be comparing best_depth or only plane_epsilon?  no worries...
			if (depth0 < best_depth && depth1 < best_depth)
			{
				if (VectorDot(&cross, &o_delta) < 0)
					VectorScale(&cross, -1, &best_normal);
				else VectorCopy(&cross, &best_normal);
				//best_normal = ((cross % o_delta) < 0) ? -cross : cross;
				best_depth = min(depth0,depth1);
				VectorAdd(&vec0, &vec1, &best_point);
				VectorScale(&best_point, 0.5f, &best_point);
				//best_point = (vec0 + vec1) * 0.5f;
				collided = true;
				vertex_face = false;
			}
		}
	}

	if (!collided) return 0;

	VectorCopy(&best_point, res_point);
	VectorCopy(&best_normal, res_normal);

	return 1;
#endif
}

void CalcWorldInertiaInverse(const matrix_t *angle, const matrix_t *body, matrix_t *world)
{
	//calculate world inertia inverse matrix
	matrix_t angleInv;
	MatrixMul(angle, body, world);
	MatrixTranspose(angle, &angleInv);
	MatrixMul(world, &angleInv, world);
}

void ApplyForce(entity_t *e, const vector_t *point, const vector_t *force)
{
	vector_t reg;

//apply impulses - directly to velocity, as to not cause d/dt errors
	VectorScale(force, e->massInverse, &reg);
	VectorAdd(&e->velocity, &reg, &e->velocity);
	//e->velocity += force * e->massInverse;

//calc world inertia inverse
	matrix_t world_inertia;
	CalcWorldInertiaInverse(&e->matrix, &e->bodyInertiaInverse, &world_inertia);

//angular momentum influence
	VectorSub(point, &e->position, &reg);
	VectorCross(&reg, force, &reg);
	MatrixMul(&world_inertia, &reg, &reg);
	VectorAdd(&e->rotation, &reg, &e->rotation);
	//e->rotation += world_inertia * ((point - e->position) * force);
}

//#define CONSOLE_DEBUG_ON_YO_CHIN

void ApplyImpulseForces(const collision_info_t *info, collision_result_t *result)
{
	vector_t offset[2];
	vector_t vel_ab;
	scalar_t restitution;

	entity_t *e0 = info->obj[0];
	entity_t *e1 = info->obj[1];

	//temporary use vector
	vector_t reg;

#ifdef CONSOLE_DEBUG_ON_YO_CHIN

	particleSystem->CreateParticle(&info->point, &vectorZero, 10, 0xFFFFFFFF, 20, false);

	VectorMultAdd(&info->point, &info->normal, 5, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FFFF, 20, false);
	VectorMultAdd(&info->point, &info->normal, 10, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FFFF, 20, false);
	VectorMultAdd(&info->point, &info->normal, 15, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FFFF, 20, false);
	VectorMultAdd(&info->point, &info->normal, 20, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FFFF, 20, false);
	VectorMultAdd(&info->point, &info->normal, 25, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FFFF, 20, false);

	VectorMultAdd(&info->point, &info->normal, -5, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FF00, 20, false);
	VectorMultAdd(&info->point, &info->normal, -10, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FF00, 20, false);
	VectorMultAdd(&info->point, &info->normal, -15, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FF00, 20, false);
	VectorMultAdd(&info->point, &info->normal, -20, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FF00, 20, false);
	VectorMultAdd(&info->point, &info->normal, -25, &reg);
	particleSystem->CreateParticle(&reg, &vectorZero, 10, 0xFF00FF00, 20, false);

	main_console->Print("Collision Time %f: E0: %p  E1: %p\n", gameTime, e0, e1);
	if (e0)	main_console->Print("  e0->mass_inv: %.20f\n", info->obj[0]->massInverse);
	if (e1) main_console->Print("  e1->mass_inv: %.20f\n", info->obj[1]->massInverse);
	if (e0)	main_console->Print("  e0->scale:    %f %f %f\n", info->obj[0]->scale.x, info->obj[0]->scale.y, info->obj[0]->scale.z);
	if (e1)	main_console->Print("  e1->scale:    %f %f %f\n", info->obj[1]->scale.x, info->obj[1]->scale.y, info->obj[1]->scale.z);
	if (e0)	main_console->Print("  e0->volume:   %.20f\n", info->obj[0]->c_model->volume);
	if (e1)	main_console->Print("  e1->volume:   %.20f\n", info->obj[1]->c_model->volume);
	main_console->Print(" Pre:\n");
	if (e0)	main_console->Print("  e0->vel: %f %f %f\n", info->obj[0]->velocity.x, info->obj[0]->velocity.y, info->obj[0]->velocity.z);
	if (e1)	main_console->Print("  e1->vel: %f %f %f\n", info->obj[1]->velocity.x, info->obj[1]->velocity.y, info->obj[1]->velocity.z);
	main_console->Print("  info->normal:  %f %f %f\n", info->normal.x, info->normal.y, info->normal.z);
	main_console->Print("  info->point:   %f %f %f\n", info->point.x, info->point.y, info->point.z);
#endif

//calc collision offset
	if (e0)	VectorSub(&info->point, &info->pos[0], &offset[0]);
	if (e1)	VectorSub(&info->point, &info->pos[1], &offset[1]);

////grab dif in velocities (incorperate addition via angular velocity)
	VectorCopy(&vectorZero, &vel_ab);
	if (e0)
	{
		VectorAdd(&vel_ab, &e0->velocity, &vel_ab);
		VectorCross(&e0->rotation, &offset[0], &reg);
		VectorAdd(&vel_ab, &reg, &vel_ab);
		//vel_ab += e0->velocity + e0->rotation * offset[0];
	}
	if (e1)
	{
		VectorSub(&vel_ab, &e1->velocity, &vel_ab);
		VectorCross(&e1->rotation, &offset[1], &reg);
		VectorSub(&vel_ab, &reg, &vel_ab);
		//vel_ab -= e1->velocity + e1->rotation * offset[1];
	}

#ifdef CONSOLE_DEBUG_ON_YO_CHIN
	main_console->Print("  vel_ab:  %f %f %f\n", vel_ab.x, vel_ab.y, vel_ab.z);
#endif

	scalar_t vel_ab_dot_normal = VectorDot(&vel_ab, &info->normal);

	if (vel_ab_dot_normal > 0) return;

//calc world inertia inverse
	matrix_t i0, i1;
	if (e0)	CalcWorldInertiaInverse(&e0->matrix, &e0->bodyInertiaInverse, &i0);
	else	MatrixCopy(&matrixZero, &i0);
	if (e1)	CalcWorldInertiaInverse(&e1->matrix, &e1->bodyInertiaInverse, &i1);
	else	MatrixCopy(&matrixZero, &i1);

//	restitution = obj[0]->SurfaceRestitution * obj[1]->SurfaceRestitution;
	restitution = 1.f;

//calc impulse force
//             _      _    _   _                      _      _           _      _
//info->impulse_mag = -(1 + e) v_ab * n / (n * n * (1/m_a + 1/m_b) + (r_ap x n)^2/I_a + (r_bp x n)^2/I_b)
	scalar_t numerator = -(1.f + restitution) * vel_ab_dot_normal;

	scalar_t denominator = 0;
	if (e0)	denominator += e0->massInverse;
	if (e1)	denominator += e1->massInverse;		//scale this value by N*N if the collision normal is not a unit vector
//angular momentum influence
	if (e0)
	{
		VectorCross(&offset[0], &info->normal, &reg);
		MatrixMul(&i0, &reg, &reg);
		VectorCross(&reg, &offset[0], &reg);
		denominator += VectorDot(&info->normal, &reg);
//		denominator += info->normal % ((i0 * (offset[0] * info->normal)) * offset[0]);
	}
	if (e1)
	{
		VectorCross(&offset[1], &info->normal, &reg);
		MatrixMul(&i1, &reg, &reg);
		VectorCross(&reg, &offset[1], &reg);
		denominator += VectorDot(&info->normal, &reg);
//		denominator += info->normal % ((i1 * (offset[1] * info->normal)) * offset[1]);
	}

	result->impulse_mag = numerator / denominator;

//create force vector from force magnitude
	VectorScale(&info->normal, result->impulse_mag, &result->impulse);

#ifdef CONSOLE_DEBUG_ON_YO_CHIN
	main_console->Print("  impulse: %f %f %f\n", info->impulse.x, info->impulse.y, info->impulse.z);
#endif

	if (e0)
	{

//apply impulses - directly to velocity, as to not cause d/dt errors
		vector_t force_obj0;
		VectorScale(&result->impulse, e0->massInverse, &force_obj0);
		VectorAdd(&e0->velocity, &force_obj0, &e0->velocity);

#ifdef CONSOLE_DEBUG_ON_YO_CHIN
		main_console->Print("  force_0: %f %f %f\n", force_obj0.x, force_obj0.y, force_obj0.z);
#endif

//angular momentum influence
		VectorCross(&offset[0], &result->impulse, &reg);
		MatrixMul(&i0, &reg, &reg);
		VectorAdd(&e0->rotation, &reg, &e0->rotation);
		//e0->rotation += i0 * (offset[0] * info->impulse);
	}
	if (e1)
	{

//apply impulses - directly to velocity, as to not cause d/dt errors
		vector_t force_obj1;
		VectorScale(&result->impulse, -e1->massInverse, &force_obj1);
		VectorAdd(&e1->velocity, &force_obj1, &e1->velocity);

//		Vector force_obj1 = info->impulse * (-e1->massInverse);
//		e1->velocity += force_obj1;
#ifdef CONSOLE_DEBUG_ON_YO_CHIN
		main_console->Print("  force_1: %f %f %f\n", force_obj1.x, force_obj1.y, force_obj1.z);
#endif

//angular momentum influence
		VectorCross(&offset[1], &result->impulse, &reg);
		MatrixMul(&i1, &reg, &reg);
		VectorAdd(&e1->rotation, &reg, &e1->rotation);
		//e1->rotation -= i1 * (offset[1] * info->impulse);
	} 

//	e0->velocity += info->impulse * e0->massInverse;
//	e1->velocity -= info->impulse * e1->massInverse;


#ifdef CONSOLE_DEBUG_ON_YO_CHIN
	main_console->Print(" Pro:\n");
	if (e0) main_console->Print("  e0->vel: %f %f %f\n", info->obj[0]->velocity.x, info->obj[0]->velocity.y, info->obj[0]->velocity.z);
	if (e1) main_console->Print("  e1->vel: %f %f %f\n", info->obj[1]->velocity.x, info->obj[1]->velocity.y, info->obj[1]->velocity.z);
	main_console->Print("\n");
#endif
}
