/*

    File: Entity.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <math.h>

#include "../m_system3d/main.h"

#include "../m_foundation/main.h"
#include "../m_foundation/const.h"

#include "entity.h"

#include "main.h"
#include "entitytree.h"
#include "entityscript.h"
#include "collision.h"
#include "lighting.h"
#include "particle.h"		//for construction/destruction/update/display of the particle system

#include "menu.h"

//just sorta here - this is the texture for the client targetting
static Texture clientTargetTex;

//base array size
#define MAX_ENTS	5000

//base array of all entities
entity_t *entityarray = NULL;

//linked list used to hold live and dead entities
LinkedList entLiveChain, entDeadChain;

//the client entity - display, input, etc
entity_t *client_ent = NULL;

LinkedList cmodelChain;
LinkedList lodModelChain;
LinkedList skinChain;

double gameTime = 0;

//used for ammo - for script
static const char *ammo_name[AMMO_COUNT] = {
	"Gatling Ammo",
	"Rocket Ammo ",
	"Bomb Ammo   ",
};

// ******************** PUT ME SOMEWHERE ELSE ********************

Texture *GetTexture(LinkedList *p, char *title)
{
	for (Texture *t = (Texture *)(p->head); t; t = (Texture *)(t->chain_next))
		if (!_stricmp(title, t->title)) break;

	return t;
}

// ******************** USEFUL FUNCTIONS ******************** 

bool EntIsActive(entity_t *e)
{
	return (e && e->active);
}

void Ent_Relink(entity_t *e)
{
/*
		position, scale, model, c_model
					unlink and relink to entity tree
						relink based on position plus max(collision model size, render model size)
						this way the tree can be used for view occulsion + collision detection

	*/

	//revamp: only based upon physics, not render (doesnt help enuf)

	scalar_t size = 0;
	
//	if (e->model) size = e->model->GetModelByLevel(0)->radius;
//	if (e->c_model && size < e->c_model->max_radius) size = e->c_model->max_radius;
	if (e->c_model) size = e->c_model->max_radius;

	size *= VECTOR_MAX_ELEMENT(e->scale);

	for (int v = 0; v < 3; v++)
	{
		e->bbox.min.v[v] = e->position.v[v] - size;
		e->bbox.max.v[v] = e->position.v[v] + size;
	}

	EntityTree_Link(e);
}

void Ent_ModuloPosition(entity_t *e)
{
	world->Modulo(&e->position);
}


scalar_t Entity_CollisionMaxRadius(const entity_t *e)
{
	if (!e->c_model) return 0;
	return e->c_model->max_radius * VECTOR_MAX_ELEMENT(e->scale);
}

scalar_t Entity_CollisionMinRadius(const entity_t *e)
{
	if (!e->c_model) return 0;
	return e->c_model->min_radius * VECTOR_MAX_ELEMENT(e->scale);
}

// ********** CONSTRUCTION/DESTRUCTION ********** 

static const skin_t default_skin = {
	NULL,		//Texture		*tex;
	0xFFFFFFFF,	//color_t		color;
	0xFFFFFFFF,	//color_t		specular;
	128,		//unsigned char shininess;
	ST_NONE,	//skintype_t	type;
	0,			//blend_t		func;
	SF_LIGHTING	//skinflags_t	flags;
};

void Entity_Init(entity_t *e)
{
	memset(e, 0, sizeof(entity_t));

	//now set all nonzero defaults:
	e->active = true;

	//identity quat
	e->angle.w = 1;

	//identity matrix
	e->matrix.v[0][0] = 1;
	e->matrix.v[1][1] = 1;
	e->matrix.v[2][2] = 1;

	//identity scale
	VectorSet(&e->scale,	1,1,1);

	//bogus witness
	PlaneSet(&e->witness,	1,0,0,0);

	skin_t *s = e->skin;
	int i = 0;
	for (; i < ENT_MAX_SKINS; i++, s++) {
		*s = default_skin;
	}

	//and one skin
	e->skin_count = 1;		//if skin_count = 0 by default then it wont render the model

	//store the game's creation time
	e->creationTime = gameTime;
}

// ********** BLEND FUNCTION COMPRESSION ********** 

unsigned int sfactor_map[16] =
{
	GL_ZERO, GL_ONE, GL_DST_COLOR, GL_ONE_MINUS_DST_COLOR,
	GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA,
	GL_SRC_ALPHA_SATURATE, GL_ZERO, GL_ZERO, GL_ZERO,
	GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO
};

unsigned int dfactor_map[16] =
{
	GL_ZERO, GL_ONE, GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR,
	GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA,
	GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO,
	GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO
};

blend_t CompressBlend(unsigned int sfactor, unsigned int dfactor)
{
	int i;
	blend_t res = 0;

	for (i = 0; i < 16; i++)
	{
		if (sfactor_map[i] == sfactor)
		{
			res = i;
			break;
		}
	}

	for (i = 0; i < 16; i++)
	{
		if (dfactor_map[i] == dfactor)
		{
			res |= i << 4;
			break;
		}
	}

	return res;
}

unsigned int BlendSFactor(blend_t blend)
{
	return sfactor_map[blend&0x0F];
}

unsigned int BlendDFactor(blend_t blend)
{
	return dfactor_map[(blend & 0xF0) >> 4];
}

//********* ENTITY DISPLAY **********

static int Ent_Display = 1;
static int Ent_Display_BBox = 0;
static int Ent_Display_Count = 0;
static int Ent_Display_MultiTex = 1;
static int Ent_WorldLight = 1;
static int Client_DrawWire = 0;

static float modifier_LOD = 0;
//static float desired_FPS = 50;

static float cubemap_texture_matrix[16];

bool Entity_InitSkin(const skin_t *skin)
{
	//init tex
	unsigned int tex = 0;
	if (skin->tex) tex = skin->tex->texture_id;

	//init type
	switch (skin->type)
	{
	case ST_NONE:
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, tex);

		//what is up with multitex texcoords
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_TEXTURE_GEN_R);
		glDisable(GL_TEXTURE_GEN_Q);
		break;

	case ST_SPHEREMAP:
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, tex);

		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
		glEnable(GL_TEXTURE_GEN_S);
		glEnable(GL_TEXTURE_GEN_T);
		
		glMatrixMode(GL_TEXTURE);
		glScalef(1,-1,1);
		glMatrixMode(GL_MODELVIEW);
	
		break;

	case ST_CUBEMAP_REFLECTION:
	case ST_CUBEMAP_NORMAL:

		if (!found_GL_ARB_texture_cube_map || !menu_ent_cube_map)
			return false;

		//init cubemap reflection texture matrix
		glMatrixMode(GL_TEXTURE);
		glLoadMatrixf(cubemap_texture_matrix);
		glMatrixMode(GL_MODELVIEW);

		//init cube mapping
		glEnable(GL_TEXTURE_CUBE_MAP_ARB);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, tex);

		//init cubemap texcoord generators
		{
			int param = (skin->type == ST_CUBEMAP_NORMAL)
				? GL_NORMAL_MAP_ARB : GL_REFLECTION_MAP_ARB;
			glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, param);
			glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, param);
			glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, param);
		}

		glEnable(GL_TEXTURE_GEN_S);
		glEnable(GL_TEXTURE_GEN_T);
		glEnable(GL_TEXTURE_GEN_R);

		break;
	}

	//init flags/func
	if (skin->flags & SF_BLEND)
	{
		glEnable(GL_BLEND);
		glBlendFunc(BlendSFactor(skin->func),BlendDFactor(skin->func));
	}

	if (skin->flags & SF_LIGHTING && menu_ent_lighting)
		glEnable(GL_LIGHTING);

	if (skin->flags & SF_TWOSIDE)
		glDisable(GL_CULL_FACE);

	//init color
	glColor4ubv(skin->color.rgba);

	//specular color
	float specular[4] = 
	{
		(float)skin->specular.rgba[0] / 255.f,
		(float)skin->specular.rgba[1] / 255.f,
		(float)skin->specular.rgba[2] / 255.f,
		(float)skin->specular.rgba[3] / 255.f
	};
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);

	//shininess
	glMateriali(GL_FRONT, GL_SHININESS, skin->shininess);

	return true;
}

void Entity_ShutdownSkin(const skin_t *skin)
{
	//disable tex & type
	switch (skin->type)
	{
	case ST_NONE:
		glDisable(GL_TEXTURE_2D);
		break;

	case ST_SPHEREMAP:
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_TEXTURE_2D);

		glMatrixMode(GL_TEXTURE);
		glScalef(1,-1,1);
		glMatrixMode(GL_MODELVIEW);

		break;

	case ST_CUBEMAP_REFLECTION:
	case ST_CUBEMAP_NORMAL:

		if (!found_GL_ARB_texture_cube_map || !menu_ent_cube_map)
			return;

		//shutdown cubemap texcoord generators
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_TEXTURE_GEN_R);

		//shutdown cube mapping
		glDisable(GL_TEXTURE_CUBE_MAP_ARB);

		//shutdown cubemap reflection texture matrix
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);

		break;
	}

	//dont bother shutdown color...

	//shutdown flags
	if (skin->flags & SF_BLEND)
	{
		glDisable(GL_BLEND);
	}

	if (skin->flags & SF_LIGHTING && menu_ent_lighting)
		glDisable(GL_LIGHTING);

	if (skin->flags & SF_TWOSIDE)
	{
		glEnable(GL_CULL_FACE);
	}
}

void Entity_Display(const entity_t *e, const vector_t *pos)
{
	Ent_Display_Count++;

	int v;
	glPushMatrix();

	//debug display main bounding box
	if (Ent_Display_BBox)
	{
		glDisable(GL_TEXTURE);

		vector_t bbvtx[8];
		BoxCalcVtxs(&e->bbox, bbvtx);
		//debug - display maximum bounding box
		static int cube_lines[24] = {0,1,1,3,3,2,2,0,4,5,5,7,7,6,6,4,0,4,1,5,2,6,3,7};
		glBindTexture(GL_TEXTURE_2D, 0);
		glColor3f(e->touched,!e->touched,0);	//red if touched, green if not
		glBegin(GL_LINES);
		for (int v = 0; v < 24; v++) glVertex3fv(bbvtx[cube_lines[v]].v);
		glEnd();

		glEnable(GL_TEXTURE);
	}

	//offset the model
	glTranslatef(pos->x, pos->y, pos->z);

	if (Ent_Display_BBox)
	{
		glDisable(GL_TEXTURE);

		//debug - display axii
		//our debug world coordinates
		glPushMatrix();
		glScalef(2*e->c_model->max_radius,2*e->c_model->max_radius,2*e->c_model->max_radius);
		glBegin(GL_LINES);
		for (v = 0; v < 3; v++)
		{
			glColor3fv(vectorBaseAxis[v].v);
			glVertex3fv(vectorZero.v);
			glVertex3fv(e->matrix.vec[v].v);
		}
		glEnd();
		glPopMatrix();

		glEnable(GL_TEXTURE);
		glColor3f(1,1,1);
	}

	//client hack init
	if ((Client_DrawWire) && (e == client_ent))
	{
		glPushAttrib(GL_POLYGON_BIT);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}

	//find which LOD model it is that we should display...
	//right now this only looks to be used for the texture coord pointer of the model
	//something needs to be worked out...

	vector_t reg;

	VectorSub(pos, &main_camera->position, &reg);
	scalar_t dist = VectorDot(&main_camera->matrix.vec[0], &reg);
	if (dist < main_camera->znear * 0.001f) dist = main_camera->znear * 0.001f;
	scalar_t rad = e->model->GetModelByLevel(0)->radius * VECTOR_MAX_ELEMENT(e->scale);
	scalar_t tan_h2 = rad / dist;
	scalar_t tan_h1 = main_camera->halfWidth / main_camera->znear;
	scalar_t width = tan_h2 / tan_h1;

	//width = 1.f / (1.f / width + 1.f / (modifier_LOD + 1.f));
	width += modifier_LOD;
#define MINIMUM_WIDTH	0.00001f
	if (width < MINIMUM_WIDTH) width = MINIMUM_WIDTH;

	LODModel_Calculate(e->model, width);

	//now get to transforming the modelview matrix to work with
	//our to-be-displayed model

	quat_t q;

#ifdef DEBUG_TEST_FIXED_POINT_PRECISION
	//we can easily fit a unit quat into 64 bits without noticable loss of accuracy

	//convert angle to 64 bit
	short snap[4] =
	{
		(short)(e->angle.x * 32767.f),
		(short)(e->angle.y * 32767.f),
		(short)(e->angle.z * 32767.f),
		(short)(e->angle.w * 32767.f)
	};
	//convert 64 bit back to angle
	quat_t angle;
	angle.x = (float)snap[0] / 32767.f;
	angle.y = (float)snap[1] / 32767.f;
	angle.z = (float)snap[2] / 32767.f;
	angle.w = (float)snap[3] / 32767.f;
	//and convert...
	QuatToAngleAxis(&angle, &q);
#else
	QuatToAngleAxis(&e->angle, &q);
#endif

	glRotatef(q.w, q.x, q.y, q.z);	//rotate it
	glScalef(e->scale.x, e->scale.y, e->scale.z);

	//see if we are to hold off on rendering the depth image until later
	//suppressing depth rendering is extremely useful for rendering blended polys
	if (e->final_depth_pass)
	{
		//disable depth writing
		glDepthMask(GL_FALSE);
		//set the light model for two-sided lighting, just in case...
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	}

	//warning
	int skin_count = e->skin_count > ENT_MAX_SKINS ? ENT_MAX_SKINS : e->skin_count;

	//render all layers
	for (int i = 0; i < skin_count; i++)
	{
		const skin_t *skin = &e->skin[i];

		//init the skin
		if (Entity_InitSkin(skin))
		{

			//draw the model
			//e->model->DisplayInterpolate(width);
			LODModel_Display();
		}

		//shutdown the skin
		Entity_ShutdownSkin(skin);

		//prep for the second pass...
		//usually rendering algorthms set this to GL_EQUAL for only rendering
		//to the overlapping polyogns
		//I am using <= just in case e->final_depth_pass was set and writing to
		//the depth buffer was disabled
		glDepthFunc(GL_LEQUAL);
	}

	//and return to prev state
	glDepthFunc(GL_LESS);

	//if we were to suppress depth rendering until after rendering the model
	//then disable colors and write the image once again to the depth buffer
	if (e->final_depth_pass)
	{
		//set the light model for two-sided lighting, just in case...
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
		//re-enable depth writing
		glDepthMask(GL_TRUE);
		//disable color writing
		glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
		//draw the final image
		LODModel_Display();
		//and put colors back to how they were
		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
	}

	//client hack shutdown
	if (Client_DrawWire && e == client_ent)
	{
		glPopAttrib();
	}

	glPopMatrix();
}

//********* ENTITY FUNCTIONS **********  called by Engine

void EntityScript_Reset(void)
{
	Chain *link, *next;
	for (link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		entity_t *e = ENTITY_FROM_LINK(link);
		if (EntIsActive(e)) RemoveEntity(e);
	}
	
	InitEntityScript();
}

static void CFunc_EntityScript_Reset(int argc, char **argv) { EntityScript_Reset(); }

LODModel *Entity_LoadModel(char *filename)
{
	LODModel *model = GetLODModel(&lodModelChain, filename);
	if (model) return model;

	model = new LODModel();
	if (!model) return NULL;
	if (!model->Load(filename)) return NULL;

	model->Compile();

	lodModelChain.AddLast(model);
	return model;
}

CModel *Entity_LoadCModel(char *filename)
{
	CModel *cmodel = GetCModel(&cmodelChain, filename);
	if (cmodel) return cmodel;

	cmodel = new CModel();
	if (!cmodel) return NULL;
	if (!cmodel->Load(filename)) return NULL;

	cmodelChain.AddLast(cmodel);
	return cmodel;
}

Texture *Entity_LoadSkin(char *filename)
{
	Texture *skin = GetTexture(&skinChain, filename);
	if (skin) return skin;

	skin = new Texture();
	if (!skin) return NULL;
	if (!skin->Create(filename)) return NULL;

	skinChain.AddLast(skin);
	return skin;
}

static int Ent_Collision = 1;

bool InitEntities(void)
{
	clientTargetTex.Create("WorldData/Target.tga");

	//base init
	entityarray = (entity_t *)malloc(MAX_ENTS * sizeof(entity_t));
	if (!entityarray) return false;

	for (int a = 0; a < MAX_ENTS; a++)
	{
		memset(&entityarray[a], 0, sizeof(entity_t));			//zero it
		entDeadChain.AddLast(&entityarray[a].entLink);		//link it
	}

	//init cvars
	new CVarw(&Ent_Display,			"Ent_Display");
	new CVarw(&Ent_Collision,			"Ent_Collision");
	new CVarw(&Ent_Display_MultiTex,	"Ent_Display_MultiTex");
	new CVarw(&Ent_Display_Count,		"Ent_Display_Count");
	new CVarw(&Ent_Display_BBox,		"Ent_Display_BBox");
	new CVarw(&Ent_WorldLight,			"Ent_WorldLight");
	new CVarw(&Client_DrawWire,		"Client_DrawWire");

	new CVarw(0,						"Reset",	CFunc_EntityScript_Reset);
	new CVarw(&modifier_LOD,			"modLOD");
//	new CVarw(&desired_FPS,			"desired_FPS");

	if (!Init_EntityTree()) return false;
	if (!InitEntityScript()) return false;

	return true;
}

void ShutdownEntities(void)
{
	clientTargetTex.Destroy();

	Shutdown_EntityTree();

	entLiveChain.UnlinkAll();
	entDeadChain.UnlinkAll();

	if (entityarray) free(entityarray);
	entityarray = NULL;

	cmodelChain.DeleteAll();
	lodModelChain.DeleteAll();
	skinChain.DeleteAll();
}

entity_t *CreateEntity(void)
{
/*
	entity_t *e = (entity_t *)entDeadChain.tail;

	if (e)
	{
		e->Unlink();
		Entity_Init(e);
		entLiveChain.AddLast(e);
		return e;
	}

	return NULL;
*/
	Chain *l = entDeadChain.tail;

	if (l)
	{
		entity_t *e = ENTITY_FROM_LINK(l);
		e->entLink.Unlink();
		Entity_Init(e);
		entLiveChain.AddLast(&e->entLink);
		return e;
	}

	return NULL;
}

void RemoveEntity(entity_t *ent)
{
	//upon RemoveEntity, set the entity's 'active' flag to 'false'
	//after each state space update, cycle thru live chain 
	//and move all 'active == false' entities to the dead chain

	ent->active = false;

	if (ent == client_ent) client_ent = NULL;

	//unlink this entity from all other object pointers (including cameras, etc) *HERE*
	//not from the entity list, however. that will have been already done by the chain destructor
	Chain *link, *next;
	for (link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		entity_t *e = ENTITY_FROM_LINK(link);
		if (!EntIsActive(e)) continue;
		if (e->owner == ent) e->owner = NULL;
		if (e->hook_ent == ent) e->hook_ent = NULL;
	}
}


void Entity_Traceline(const vector_t *point, const vector_t *delta, entity_t *ignore)
{
	//inverse rotate, translate, and scale the traceline - and then traceline through the base collision model
}

void DrawClientTarget(entity_t *ent)
{
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_FOG);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	//push the matrix
	glPushMatrix();
	//rotate and translate to entity space
	quat_t q;
	glTranslatef(ent->position.x, ent->position.y, ent->position.z);
	QuatToAngleAxis(&ent->angle, &q);
	glRotatef(q.w, q.x, q.y, q.z);

	static float tex_offset = 0;
	tex_offset += 1.f / 60.f;

	clientTargetTex.Set();
	glColor3f(1,1,1);

	for (float dist = 300; dist < main_camera->zfar; dist += main_camera->zfar*0.55f)
	{
		glBegin(GL_QUADS);

		glTexCoord2f(0,1);		glVertex3f(dist, 75, -75);
		glTexCoord2f(1,1);		glVertex3f(dist, -75, -75);
		glTexCoord2f(1,0);		glVertex3f(dist, -75, 75);
		glTexCoord2f(0,0);		glVertex3f(dist, 75, 75);

		glEnd();

	}

	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_FOG);
	glDisable(GL_BLEND);
}

void DrawEntityHUD_DrawRing(entity_t *ent, scalar_t fraction, scalar_t startangle)
{
	const static scalar_t innerRadius = 0.55f;
	const static scalar_t outerRadius = 0.60f;
	scalar_t v_step = 0.025f;

	vector_t vtx[4];

	scalar_t costh = (scalar_t)cos(startangle);
	scalar_t sinth = (scalar_t)sin(startangle);
	scalar_t costh_delta = (scalar_t)cos(v_step * M_PI);
	scalar_t sinth_delta = (scalar_t)sin(v_step * M_PI);
	
	for (scalar_t v = 0; v < fraction; v += v_step)
	{
		scalar_t costh_next = costh * costh_delta - sinth * sinth_delta;
		scalar_t sinth_next = sinth * costh_delta + costh * sinth_delta;

		VectorSet(&vtx[0], 1, innerRadius * costh, innerRadius * sinth);
		VectorSet(&vtx[1], 1, outerRadius * costh, outerRadius * sinth);
		VectorSet(&vtx[2], 1, outerRadius * costh_next, outerRadius * sinth_next);
		VectorSet(&vtx[3], 1, innerRadius * costh_next, innerRadius * sinth_next);

		MatrixMul(&main_camera->matrix, &vtx[0], &vtx[0]);
		MatrixMul(&main_camera->matrix, &vtx[1], &vtx[1]);
		MatrixMul(&main_camera->matrix, &vtx[2], &vtx[2]);
		MatrixMul(&main_camera->matrix, &vtx[3], &vtx[3]);

		VectorMultAdd(&main_camera->position, &vtx[0], 10.f, &vtx[0]);
		VectorMultAdd(&main_camera->position, &vtx[1], 10.f, &vtx[1]);
		VectorMultAdd(&main_camera->position, &vtx[2], 10.f, &vtx[2]);
		VectorMultAdd(&main_camera->position, &vtx[3], 10.f, &vtx[3]);

		glBegin(GL_QUADS);
		glTexCoord2f(v+v_step,0);	glVertex3fv(vtx[3].v);
		glTexCoord2f(v+v_step,1);	glVertex3fv(vtx[2].v);
		glTexCoord2f(v,1);			glVertex3fv(vtx[1].v);
		glTexCoord2f(v,0);			glVertex3fv(vtx[0].v);
		glEnd();
		
		costh = costh_next;
		sinth = sinth_next;
	}

	if (v != fraction)
	{
		v -= v_step;
		v_step = fraction - v;
		costh_delta = (scalar_t)cos(v_step * M_PI);
		sinth_delta = (scalar_t)sin(v_step * M_PI);

		scalar_t costh_next = costh * costh_delta - sinth * sinth_delta;
		scalar_t sinth_next = sinth * costh_delta + costh * sinth_delta;

		VectorSet(&vtx[0], 1, innerRadius * costh, innerRadius * sinth);
		VectorSet(&vtx[1], 1, outerRadius * costh, outerRadius * sinth);
		VectorSet(&vtx[2], 1, outerRadius * costh_next, outerRadius * sinth_next);
		VectorSet(&vtx[3], 1, innerRadius * costh_next, innerRadius * sinth_next);

		MatrixMul(&main_camera->matrix, &vtx[0], &vtx[0]);
		MatrixMul(&main_camera->matrix, &vtx[1], &vtx[1]);
		MatrixMul(&main_camera->matrix, &vtx[2], &vtx[2]);
		MatrixMul(&main_camera->matrix, &vtx[3], &vtx[3]);

		VectorMultAdd(&main_camera->position, &vtx[0], 10.f, &vtx[0]);
		VectorMultAdd(&main_camera->position, &vtx[1], 10.f, &vtx[1]);
		VectorMultAdd(&main_camera->position, &vtx[2], 10.f, &vtx[2]);
		VectorMultAdd(&main_camera->position, &vtx[3], 10.f, &vtx[3]);

		glBegin(GL_QUADS);
		glTexCoord2f(v+v_step,0);	glVertex3fv(vtx[3].v);
		glTexCoord2f(v+v_step,1);	glVertex3fv(vtx[2].v);
		glTexCoord2f(v,1);			glVertex3fv(vtx[1].v);
		glTexCoord2f(v,0);			glVertex3fv(vtx[0].v);
		glEnd();
	}
}

void DrawEntityHUD_DrawString(char *p, float x, float y, float z)
{
	int tx,ty;
	vector_t vtx[4];

	VectorSet(&vtx[0], x, y, z);
	MatrixMul(&main_camera->matrix, &vtx[0], &vtx[0]);
	VectorAdd(&vtx[0], &main_camera->position, &vtx[0]);

	for (; *p; p++)
	{
		tx = (*(unsigned char *)p & 15);
		ty = (*(unsigned char *)p >> 4) - 2;

		VectorSub(&vtx[0], &main_camera->matrix.vec[2], &vtx[1]);
		VectorSub(&vtx[1], &main_camera->matrix.vec[1], &vtx[2]);
		VectorSub(&vtx[0], &main_camera->matrix.vec[1], &vtx[3]);

		glBegin(GL_QUADS);
		glTexCoord2f((float)tx/16.f, (float)(ty)/16.f);			glVertex3fv(vtx[0].v);
		glTexCoord2f((float)tx/16.f, (float)(ty+1)/16.f);		glVertex3fv(vtx[1].v);
		glTexCoord2f((float)(tx+1)/16.f, (float)(ty+1)/16.f);	glVertex3fv(vtx[2].v);
		glTexCoord2f((float)(tx+1)/16.f, (float)(ty)/16.f);		glVertex3fv(vtx[3].v);
		glEnd();

		VectorSub(&vtx[0], &main_camera->matrix.vec[1], &vtx[0]);
	}
}

#include <stdio.h>

//this func, right there, is why rendering the HUD should
//be a part of scripting:
void DrawEntityHUD(entity_t *ent)
{
	if (!EntIsActive(ent) ||
		!ent->max_health)

		return;

	if (ent->skin[0].tex) ent->skin[0].tex->Set();

	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f(1, 1, 1, 0.3f);

	scalar_t fraction = ent->health / ent->max_health;
	if (fraction < 0) fraction = 0;
	DrawEntityHUD_DrawRing(ent, fraction, 0);

	scalar_t speed = VectorLength(&ent->velocity);
	fraction = speed / 30.f;		//current max is 20
	DrawEntityHUD_DrawRing(ent, fraction, (scalar_t)M_PI);

	//display stats
	char buf[32];
	float y;
	if (main_console->font) main_console->font->Set();
	glBlendFunc(GL_ONE, GL_ONE);

	y = -8;
	sprintf(buf, "vel: %3.3f", speed);
	DrawEntityHUD_DrawString(buf, 30, 15, y);	y--;

	sprintf(buf, "pwr: %3.3f/%3.3f", ent->health, ent->max_health);
	DrawEntityHUD_DrawString(buf, 30, 15, y);	y--;

	int j;
	for (j = 0; j < AMMO_COUNT; j++) {
		sprintf(buf, "%s: %d", ammo_name[j], ent->ammo[j]);
		DrawEntityHUD_DrawString(buf, 30, 15, y);	y--;
	}

//	for (j = 0; j < 4; j++) {
//		sprintf(buf, "r%c: %.7f", 'w' + j, ent->angle.v[(j+3)&3]);
//		DrawEntityHUD_DrawString(buf, 30, 15, y);	y--;
//	}

	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
}

#include "lodmodel.h"

void DisplayEntities(void)
{
	static int ModuloSpace = 1;
	static CVarw *fdjklsfd = new CVarw(&ModuloSpace, "ModuloSpace");

	if (!Ent_Display) return;

	//lets try not assuming this is always on for a change...
	glDisable(GL_TEXTURE_2D);

	//set texenv to our assumed default
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glEnable(GL_NORMALIZE);

	//init cubemap texture matrix
/*
to calculate the correct texture matrix,
find the inverse of the modelview matrix
~ excluding the last entity rotation

1) rotate by camera quaternion (inverse camera rotation is the standard.)
2) rotate by inverse default matrix

to keep the texture matrix aligned to the model *regardless* of model orientation,
add a inverse model quaternion rotation to the front of the list
*/
	glMatrixMode(GL_TEXTURE);

//	glLoadIdentity();	//for now im assuming this is already true
	quat_t qc;
	QuatToAngleAxis(&main_camera->angle, &qc);
	glRotatef(qc.w, qc.x, qc.y, qc.z);
	glMultMatrixf(Camera_DefaultMatrix);
	glGetFloatv(GL_TEXTURE_MATRIX, cubemap_texture_matrix);		//store externally
	glLoadIdentity();
	
	glMatrixMode(GL_MODELVIEW);

	int min[3], max[3], iv[3];

	//init modulo space
	{
		box_t frustumRange;

		//construct main camera frustum bounding box
		VectorSet(&frustumRange.min, fInf, fInf, fInf);
		VectorSet(&frustumRange.max, -fInf, -fInf, -fInf);
		for (int i = 0; i < 8; i++)
			BoxStretch(&frustumRange, &main_camera->frustum_vtx[i]);

		for (i = 0; i < 3; i++)
		{
			frustumRange.min.v[i] += world->WorldSize;
			min[i] = (int)(frustumRange.min.v[i] * 0.5 / world->WorldSize);
			if (frustumRange.min.v[i] < 0) min[i]--;		//compensate for rounded-up negative numbers

			frustumRange.max.v[i] += world->WorldSize;
			max[i] = (int)floor(frustumRange.max.v[i] * 0.5 / world->WorldSize);
			if (frustumRange.max.v[i] < 0) max[i]--;		//compensate for rounded-up negative numbers
		}
	}

	Ent_Display_Count = 0;

	Chain *link, *next;
	for (link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		entity_t *e = ENTITY_FROM_LINK(link);

		if (!EntIsActive(e)) continue;
		if (!e->model) continue;

		if (!ModuloSpace)
		{
			Entity_Display(e, &e->position);
			continue;
		}

		for (iv[0] = min[0]; iv[0] <= max[0]; iv[0]++)
		{
			for (iv[1] = min[1]; iv[1] <= max[1]; iv[1]++)
			{
				for (iv[2] = min[2]; iv[2] <= max[2]; iv[2]++)
				{
					vector_t position;
					VectorSet(&position, (scalar_t)iv[0], (scalar_t)iv[1], (scalar_t)iv[2]);
					VectorScale(&position, world->WorldSize * 2.f, &position);
					VectorAdd(&position, &e->position, &position);

					bool inside = true;

					for (int i = 0; i < 6; i++)
					{
						if (PlaneVectorDistance(&main_camera->frustum_plane[i], &position)
//TODO - SPECIAL EXCEPTION TO ADJUST FOR LODMODEL RADIUS
							> e->model->GetModelByLevel(0)->radius * VECTOR_MAX_ELEMENT(e->scale))
						{
							inside = false;
							break;
						}
					}

					if (inside) Entity_Display(e, &position);
				}
			}
		}
	}

	glDisable(GL_NORMALIZE);

	//cleanup our assumption that texture_2d is always on
	glEnable(GL_TEXTURE_2D);

	if (EntIsActive(client_ent))
	{
		DrawClientTarget(client_ent);
		DrawEntityHUD(client_ent);
	}
}


/* SYSTEM FUNCTIONS */

void UpdateEntities(void)
{
	int counter;
	entity_t *e;
	Chain *link, *next;

	//add in a world light
	//	this little light o' mine
	//	im gonna let it shine
	//	this little light o' mine
	//	im gonna let it shine
	if (Ent_WorldLight)
	{
		lightinfo_t *light = Create_Light();
		light->position[0] = 2*world->WorldSize;
		light->position[1] = 0;
		light->position[2] = 0;
		light->position[3] = 0;				///directional lighting
		Register_Light(light);
	}

	counter = 0;
	//clear 'touched' flag - debug display only
	for (link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		e = ENTITY_FROM_LINK(link);
		counter++;
		e->touched = false;
	}
	static int display_e_count = 0, display_p_count = 0;
	static scalar_t deltaTime = 1.f;
	static CVarw *fdsj[] = {
		new CVarw(&display_e_count,	"EntCounter"),
		new CVarw(&display_p_count,	"ParCounter"),
		new CVarw(&deltaTime,			"Timestep"),
		NULL};
	if (display_e_count) main_console->Print("EntCount: %d\n", counter);
	if (display_p_count) main_console->Print("ParCount: %d\n", Particle_Counter());

	//update state space / think
	for (link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		e = ENTITY_FROM_LINK(link);
		if (!EntIsActive(e)) continue;

		//execute the think function
		if (e->Think) e->Think(e);
	}

	//update physics
	if (Ent_Collision) EntityTree_UpdatePhysics(deltaTime);

	//update frame counter (fix this mess)
	gameTime++;// = main_timer->GetTime();

	//update client input (fix this mess)
	UpdateClientInput();

	//cleanup	- move all dead entities to the live ent chain
	//			- and remove all 'spawn' flags
	for (link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		e = ENTITY_FROM_LINK(link);
		if (!e->active)
		{
			//unlink from live ent chain
			e->entLink.Unlink();

			//add to dead chain
			entDeadChain.AddLast(&e->entLink);

			//unlink from world BSP tree
			//do this here so we dont get muddled bsp trees of alive/dead entities
			e->treeLink.Unlink();
		}
	}

	//try to tweak the framerate
//	modifier_LOD += (system_currentFPS < desired_FPS) ? -0.01f : 0.01f;
}
