/*

    File: Entity.h

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#ifndef ENTITY_H
#define ENTITY_H

typedef struct entity_s entity_t;

#include "../m_math3d/geometry.h"
#include "../m_math3d/quat.h"
#include "../m_math3d/matrix.h"

#include "../m_foundation/chain.h"
#include "../m_foundation/texture.h"
#include "../m_foundation/types.h"

#include "lodmodel.h"
#include "cmodel.h"
#include "collision.h"

//very unorganized
//used for entity and entity friends
#define VECTOR_MAX_ELEMENT(v) ((v.x > v.y && v.x > v.z) ? v.x : ((v.y > v.z) ? v.y : v.z))
#define VECTOR_MIN_ELEMENT(v) ((v.x < v.y && v.x < v.z) ? v.x : ((v.y < v.z) ? v.y : v.z))

//used for ammo - for script
enum {
	AMMO_GATLING,
	AMMO_ROCKETS,
	AMMO_BOMBS,
	AMMO_COUNT
};

//used for texcoord generation
typedef unsigned char skintype_t;
enum {
	ST_NONE,
	ST_SPHEREMAP,
	ST_CUBEMAP_REFLECTION,		//calc some reflection vectors before mapping the cube to the obj
	ST_CUBEMAP_NORMAL,			//directly map the cube to the object based upon the normals
};

//used for enable/disable bits
typedef unsigned char skinflags_t;
#define SF_BLEND		1		//GL_BLENDFUNC (dont make use of BLEND_NONE anymore)
#define SF_LIGHTING		2		//GL_LIGHTING
#define SF_TWOSIDE		4		//GL_CULL_FACE inverse

//the blend type
typedef unsigned char blend_t;

//compressing the two uint pairs into one byte:
blend_t CompressBlend(unsigned int sfactor, unsigned int dfactor);

#define ENT_MAX_SKINS	8

typedef struct skin_s {
	//the texture - pointer to the struct with the goods
	Texture		*tex;
	//diffuse lighting
	color_t		color;
	//specular lighting
	color_t		specular;
	//specular highlight
	unsigned char shininess;		//even tho 128 is the highest, this allows 0 thru 255
	//texture coordinate generation method
	skintype_t	type;
	//blend function - compressed and decompressed
	blend_t		func;
	//whether to enable lighting for this layer
	skinflags_t	flags;
} skin_t;

#define	ENTITY_FROM_LINK(e) ((entity_t *)((unsigned char *)e - (int)&(((entity_t *)NULL)->entLink)))

struct entity_s {

// ********** ENTITY VARIABLES ********** 

// ***** system ***** 

	//read-only
	Chain			entLink;			//system-only - written in entity_t, read everywhere else - links to the live or dead chain
	Chain			treeLink;			//system-only - used by EntityTree - determines where in the BSP Tree this belongs
	bool			active;				//system-only - whether an obj is active or not
	double			creationTime;		//system-only - gameTime upon creation

	//called by state system per frame
	void			(*Think)(entity_t *ent);

// ***** physics ***** 

	bool			check_collision;	//whether or not collision detection will be attempted
	bool			bounces;			//whether or not collision responce will be attempted
	
	vector_t		velocity;			//physics system state variable
	vector_t		rotation;			//physics system state variable

	//called by physics system upon collision detection
	void			(*Touch)(entity_t *ent, entity_t *touch, collision_info_t *info);

	//called by physics system after collision forces are applied (PostTouch)
	void			(*Bounce)(entity_t *ent, entity_t *touch, collision_info_t *info, collision_result_t *result);

	//read-only.  use Entity_SetCModel to write
	CModel			*c_model;			//collision model - convex hull - polygonal surface 

	//read-only.  use Entity_SetDensityInverse to write.
	scalar_t		densityInverse;		//used by physics system to calculate mass and inertia inverse
	//read-only.  use Entity_CalcMassInertia to write. updated upon Entity_SetDensityInverse
	scalar_t		massInverse;		//used by physics system to calculate forces upon bodies with mass
	//read-only.  use Entity_CalcMassInertia to write. updated upon Entity_SetDensityInverse
	matrix_t		bodyInertiaInverse;	//used by physics system to calculate torques upon bodies with inertia

	//collision access only
	plane_t			witness;			//the last sucessful witness used - this could speed things up

// ***** physics/render ***** 

	//read-only.  use Entity_SetPosition to write
	vector_t		position;			//upon set modulate by world->WorldSize and relink to entity bsp tree
	//read-only.  use Entity_SetAngle to write
	quat_t			angle;				//upon set udpate matrix
	//read-only.  use Entity_SetScale to write
	vector_t		scale;				//object scale magnitude
	//read-only.  updated upon Entity_SetAngle
	matrix_t		matrix;				//orientation matrix - useful to have around

	//updated on position change.
	//based upon position, c_model size, model size, scale.
	//used for bsp tree linking

	//read-only.  updated upon Entity_SetPosition, Entity_SetCModel, Entity_SetScale
	box_t			bbox;

// ***** render ***** 

	//read-only.  use Entity_SetModel to write
	LODModel		*model;				//display model - triangle mesh - LOD burliness

	skin_t			skin[ENT_MAX_SKINS];
	unsigned char	skin_count;

	//set this flag if you are using any overlapping two_sided blended polys
	//in other words, set this if you ever use SF_BLEND on non-convex models or
	//(SF_BLEND | SF_TWO_SIDE) on any models and do not want self-overlapping
	//polys to be blocked out of the blending process
	bool			final_depth_pass;
										

// ***** debug ***** 

	//do not bother with.  written by physics, read & cleared by render
	bool			touched;			//debug - true if an object has been touched this frame

// ***** script ***** 
	char			type[64];			//64max byte null-terminated string depicting object type
	bool			takesdamage;		//inter-script - whether an object can take damage

	scalar_t		max_health;			//inter-script - used with 'takesdamage' flag
	scalar_t		health;				//inter-script - used with 'takesdamage' flag

	int				ammo[AMMO_COUNT];

	//used when damage is dealt
	void			(*Pain)(entity_t *ent);
	//used when damage is dealt and health is below zero
	void			(*Death)(entity_t *ent);

	double			timer;				//inter-script - varies per use
	double			lastRockTouch;		//inter-script - used by player to prevent multiple rock contacts occuring within milliseconds
	entity_t		*owner;				//inter-script.  used mainly for keepint track of projectile ownership

	entity_t		*hook_ent;			//inter-script - who the hook latches onto
	vector_t		hook_ofs;			//inter-script - where the hook latched onto something (in that ent's modelspace)
	scalar_t		hook_len;			//inter-script - how long the hook is
};

//suggestion - instead of adding craploads onto this struct like QC does, why dont we have a 256-byte template?
//nah... enforcing standard variable locations would be too tough...
//dare we superclass/subclass?

//used by main.cpp - for main construction/destruction
bool InitEntities(void);
void UpdateEntities(void);
void DisplayEntities(void);
void ShutdownEntities(void);

//used by EntityScript.cpp, EntityTree.cpp
extern LinkedList entLiveChain;
extern entity_t *client_ent;				//we're still working on what to do with this one

//used by our entity.cpp friend: entityscript.cpp
entity_t *CreateEntity(void);
void RemoveEntity(entity_t *e);
bool EntIsActive(entity_t *e);

//entityscript useful func
void Entity_Traceline(const vector_t *point, const vector_t *delta, entity_t *ignore);

//entityscript chain shell functions
LODModel *Entity_LoadModel(char *filename);
CModel *Entity_LoadCModel(char *filename);
Texture *Entity_LoadSkin(char *filename);

//interface to entitytree physics update system
void Ent_Relink(entity_t *e);
void Ent_UpdateAngle(entity_t *e);
void Ent_ModuloPosition(entity_t *e);

//entitytree and entityscript
//dependant upon e->scale and e->c_model
//returns the maximum radius for potential collision (furthest scaled vertex on cmodel)
scalar_t Entity_CollisionMaxRadius(const entity_t *e);
scalar_t Entity_CollisionMinRadius(const entity_t *e);

//turn this into a integer frame counter
extern double gameTime;

//communcation with main/menu - called upon new game
void EntityScript_Reset(void);

#endif