/*

    File: EntityScript.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <math.h>

#include "../m_system3d/main.h"

#include "../m_foundation/main.h"		//main_console
#include "../m_foundation/const.h"
#include "../m_foundation/random.h"

#include "entityscript.h"
#include "main.h"
#include "lighting.h"
#include "entity.h"

#include "collision.h"	//applyforce... 
#include "particle.h"

#include "menu.h"

//particle system textures
static Texture *particleSmokeTex = NULL;
static Texture *particleFlameTex = NULL;

//console variables
static int		Client_Invincible = 0;

static scalar_t	Client_Accel = 0.2f;
static scalar_t	Client_MaxVel = 20.f;

static scalar_t	Client_Rot = 6.f;		//only used for rolling

static int		Client_UseMouse = 1;
static scalar_t	Client_Sensitivity = 2.f;

/* CORE FUNCTIONS
 *   always use these to set vars as opposed to directly setting them
 */

void Entity_SetModel(entity_t *e, char *filename)
{
	if (filename)	e->model = Entity_LoadModel(filename);
	else			e->model = NULL;

	Ent_Relink(e);
}

void Ent_CalcMassInertia(entity_t *e)
{
	if (!e->c_model) return;

	e->massInverse = e->densityInverse / (e->c_model->volume * e->scale.x * e->scale.y * e->scale.z);

	vector_t inertia;
	e->c_model->CalcInertia(&e->scale, &inertia);

	//calc body inertia inverse
	MatrixSet(&e->bodyInertiaInverse,
		e->densityInverse / inertia.x, 0, 0,
		0, e->densityInverse / inertia.y, 0,
		0, 0, e->densityInverse / inertia.z);
}

void Entity_SetCModel(entity_t *e, char *filename)
{
	if (filename)	e->c_model = Entity_LoadCModel(filename);
	else			e->c_model = NULL;

	//update bodyInertiaInverse, massInverse		
	Ent_CalcMassInertia(e);

	//redo bbox
	Ent_Relink(e);
}

void Ent_UpdateAngle(entity_t *e)
{
	//normalize angle
	QuatUnit(&e->angle, &e->angle);

	//update axii
	QuatToMatrix(&e->angle, &e->matrix);
}

void Entity_SetPosition(entity_t *e, const vector_t *pos)
{
	VectorCopy(pos, &e->position);

	//modulate position - check the bounds to the world
	Ent_ModuloPosition(e);

	//redo bbox
	Ent_Relink(e);
}

void Entity_SetAngle(entity_t *e, const quat_t *a)
{
	QuatCopy(a, &e->angle);

	//update axii
	Ent_UpdateAngle(e);
}

void Entity_SetScale(entity_t *e, const vector_t *s)
{
	VectorCopy(s, &e->scale);

	//update bodyInertiaInverse, massInverse		
	Ent_CalcMassInertia(e);

	//redo bbox
	Ent_Relink(e);
}

void Entity_SetDensityInverse(entity_t *e, scalar_t dInv)
{
	e->densityInverse = dInv;

	//update bodyInertiaInverse, massInverse		
	Ent_CalcMassInertia(e);
}


/* MISC FUNCTIONS */

void Deal_Damage(entity_t *victim, scalar_t dmg)
{
	if (!victim->takesdamage) return;
// dmg*=1000;
	victim->health -= dmg;

	if (victim->health > 0)
	{
		if (victim->Pain) victim->Pain(victim);
	}
	else
	{
		if (victim->Death) victim->Death(victim);
	}
}

void Remove_Self(entity_t *e)
{
	RemoveEntity(e);
}

void Entity_RocketEffect(entity_t *e)
{
	//add some rocket particle effects
	Particle_SetTexture(particleFlameTex);

	for (int count = 0; count < 5; count++)
	{
		scalar_t dist = MFRand();
		scalar_t size = VectorLengthSq(&e->velocity) * 0.01f;	//approx range 0 thru 45

		vector_t pos;
		VectorMultAdd(&e->position, &e->matrix.vec[0], -(30 + dist * size), &pos);
		VectorMultAdd(&pos, &e->matrix.vec[2], 7, &pos);
		//pos = e->position - e->axis[0] * (30 + dist * size) + e->axis[2] * 7;

		size = (size * 0.25f + 10)*(1.f - dist * MFRand()) * 2;
		Particle_Create(&pos, &vectorZero, size, 0xFFFF7F7F, 2, true);

		//assume vel max is around 4500
	}

	lightinfo_t *light = Create_Light();
	light->position[0] = e->position.x - e->matrix.vec[0].x * 30 + e->matrix.vec[2].x * 7;
	light->position[1] = e->position.y - e->matrix.vec[0].y * 30 + e->matrix.vec[2].y * 7;
	light->position[2] = e->position.z - e->matrix.vec[0].z * 30 + e->matrix.vec[2].z * 7;
	light->position[3] = 1;
	light->attenuation[2] = light_default_quadratic_attenuation;
	Register_Light(light);
}

/* SHOT FUNCTIONS */

void Particle_Explosion(const vector_t *center, int count, scalar_t size)
{
	Particle_SetTexture(particleFlameTex);

	for (int v = 0; v < count; v++)
	{
		float rad = MFRand()*MFRand() + MFRand();

		vector_t vel;
		VectorSet(&vel, MCRand(),MCRand(),MCRand());
		VectorUnit(&vel, &vel);
		VectorScale(&vel, size * (rad - 0.75f), &vel);

		color_t c;
		c.rgba[0] = 0xFF;
		c.rgba[1] = MRand() & 0x7F;
		c.rgba[2] = c.rgba[1] >> 1;
		c.rgba[3] = 0xFF;

		Particle_Create(center, &vel, size * 10.f*(2.25f - rad), c.l, 50 + (MRand() % 50), true);
	}
}

void Particle_ExplosionAlt(const vector_t *center)
{
#if 1
#if 0			//oldschool

	particleSystem->SetTexture(particleFlameTex);

	for (int v = 0; v < 50; v++)
	{
		color_t color;

		float rad = MFRand();

		Vector vel(MCRand(),MCRand(),MCRand());
		vel.Normalize();
		vel *= 30.f * (1.0f - rad);

		color.rgba[0] = 0x3F;
		color.rgba[1] = (MRand() & 0x7F) + 0x3F;
		color.rgba[2] = 0xFF;
		color.rgba[3] = 0xFF;

		Scalar partRad = 350.f * (rad * rad * rad + 0.1f);
		int partTime = 150 + (int)(rad * 50);

		particleSystem->CreateParticle(center + vel, vel, partRad, color.l, partTime, true);
		particleSystem->CreateParticle(center + vel, vel, partRad, 0xFFFFFFFF, partTime, true);

	}

#else			//newschool

	Particle_SetTexture(particleFlameTex);

	for (int v = 0; v < 100; v++) {
		color_t color;
		vector_t pos, vel;
		scalar_t partRad;
		int partTime;

		VectorSet(&vel, MCRand(),MCRand(),MCRand());
		VectorUnit(&vel, &vel);

		if (MRand() & 3) {

			VectorScale(&vel, 20.f, &vel);

			color.rgba[0] = 0x3F;
			color.rgba[1] = (MRand() & 0x7F) + 0x3F;
			color.rgba[2] = 0xFF;
			color.rgba[3] = 0xFF;

			partRad = 50.f;
			partTime = 150 + (MRand() % 50);
		} else {

			VectorScale(&vel, 2.5f * MFRand(), &vel);

			color.rgba[0] = 0;
			color.rgba[1] = 0;
			color.rgba[2] = 0xFF;
			color.rgba[3] = 0xFF;

			partRad = 150.f;
			partTime = 200 + (MRand() % 100);
		}

		VectorAdd(center, &vel, &pos);

		Particle_Create(&pos, &vel, partRad, color.l, partTime, true);
		Particle_Create(&pos, &vel, partRad, 0xFFFFFFFF, partTime, true);

	}

#endif
#else

	Particle_SetTexture(particleFlameTex);

	for (int v = 0; v < 10; v++)
	{
		color_t c;

		vector_t vel, pos;

		VectorSet(&vel, MCRand(),MCRand(),MCRand());
		VectorUnit(&vel, &vel);
		VectorScale(&vel, 2.5f * MFRand(), &vel);

		VectorAdd(center, &vel, &pos);

		c.rgba[0] = 0;
		c.rgba[1] = 0;
		c.rgba[2] = 0xFF;
		c.rgba[3] = 0xFF;

		Particle_Create(&pos, &vel, 50.f, c.l, 100 + (MRand() % 100), true);
		Particle_Create(&pos, &vel, 50.f, 0xFFFFFFFF, 100 + (MRand() % 100), true);

	}
#endif
}

void Particle_Ring(const vector_t *center, const vector_t *normal)
{
	Particle_SetTexture(particleFlameTex);

	for (int v = 0; v < 25; v++)
	{
		color_t c;

		vector_t vel, reg, pos;

		VectorSet(&vel, MCRand(),MCRand(),MCRand());
		VectorUnit(&vel, &vel);
		VectorScale(&vel, 2.5f * MFRand(), &vel);

		VectorScale(normal, (VectorDot(&vel, normal) * (0.1f * MFRand() + 0.95f) ), &reg);
		VectorSub(&vel, &reg, &vel);
		//vel -= normal * ((vel % normal) * (0.1f * MFRand() + 0.95f) );

		VectorAdd(center, &vel, &pos);

		c.rgba[0] = 0;
		c.rgba[1] = 0;
		c.rgba[2] = 0xFF;
		c.rgba[3] = 0xFF;

		Particle_Create(&pos, &vel, 25.f, c.l, 100 + (MRand() % 100), true);
		Particle_Create(&pos, &vel, 25.f, 0xFFFFFFFF, 100 + (MRand() % 100), true);

	}
}

void StartMediumRock(entity_t *e, entity_t *rock);
void StartSmallRock(entity_t *e, entity_t *rock);
void StartTinyRock(entity_t *e, entity_t *rock);
void StartItem(entity_t *e);

void RockDeath(entity_t *rock)
{
	int v;

	if ((MRand() & 4) == 4)
	{
		entity_t *e = CreateEntity();
		if (!EntIsActive(e)) return;
		Entity_SetPosition(e, &rock->position);
		StartItem(e);
	}

	if (!strcmp(rock->type, "rock_l"))
	{
		rock->check_collision = false;
		for (v = 0; v < 4; v++)
		{
			entity_t *e = CreateEntity();
			if (!EntIsActive(e)) return ;
			StartMediumRock(e, rock);
		}
	}
	else if (!strcmp(rock->type, "rock_m"))
	{
		rock->check_collision = false;
		for (v = 0; v < 4; v++)
		{
			entity_t *e = CreateEntity();
			if (!EntIsActive(e)) return ;
			StartSmallRock(e, rock);
		}
	}
	else if (!strcmp(rock->type, "rock_s"))
	{
		rock->check_collision = false;
		for (int t = 0; t < 4; t++)
		{
			entity_t *e = CreateEntity();
			if (!EntIsActive(e)) return;
			StartTinyRock(e, rock);
		}
	}

	RemoveEntity(rock);
}

void RockThink(entity_t *rock)
{
	static scalar_t rockMaxVelocity = 10;
	static scalar_t rockMaxRotation = 5 * (scalar_t)DEGREE_TO_RADIAN;

	static CVarw *dfjlksfd[] = {
		new CVarw(&rockMaxVelocity,"rockMaxVel"),
		new CVarw(&rockMaxRotation,"rockMaxRot"),
		NULL};

	scalar_t speed = VectorLength(&rock->velocity);
	if (speed > rockMaxVelocity )
		VectorScale(&rock->velocity, rockMaxVelocity / speed, &rock->velocity);

	speed = VectorLength(&rock->rotation);
	if (speed > rockMaxRotation)
		VectorScale(&rock->rotation, rockMaxRotation / speed, &rock->rotation);
}

void Fade_Think(entity_t *fade)
{
	fade->timer--;

	double x = 1.0 - fade->timer / 60.0;

	//fade->color.w = 0.5f * (float)sqrt(1.f - x*x);
	//fade->color.rgba[3] = (unsigned char)(127.0 * sqrt(1.0 - x*x));
	fade->skin[0].color.rgba[3] = (unsigned char)(127.0 * sqrt(1.0 - x*x));

	if (fade->timer < 0) RemoveEntity(fade);
}

void Shot_Bright(entity_t *shot)
{
	float life = (float)shot->timer / 30.f;

	lightinfo_t *light = Create_Light();
	light->position[0] = shot->position.x;
	light->position[1] = shot->position.y;
	light->position[2] = shot->position.z;
	light->attenuation[0] = 0.1f + (1 - life) * 5;
	light->attenuation[1] = light_default_linear_attenuation;
	Register_Light(light);

	shot->timer--;
	if (shot->timer < 0) RemoveEntity(shot);
}

void Shot_Touch(entity_t *shot, entity_t *touch, collision_info_t *info)
{
	if (!EntIsActive(touch)) return ;
	if (touch == shot->owner) return;

	vector_t force;
	VectorScale(&shot->velocity, (200000), &force);

	//realign realworld position (thanks modulo space)
	vector_t position;
	world->Relative(&shot->position, &touch->position, &position);
	ApplyForce(touch, &position, &force);

	Deal_Damage(touch, 1.5);

	Particle_Explosion(&shot->position, 5, 15);

	shot->timer = 30;
	shot->Think = Shot_Bright;
	shot->Touch = NULL;
	Entity_SetModel(shot, NULL);
}

void Shot_Think(entity_t *shot)
{
	float lum = MFRand();

	lightinfo_t *light = Create_Light();
	light->position[0] = shot->position.x;
	light->position[1] = shot->position.y;
	light->position[2] = shot->position.z;
	light->attenuation[0] = lum * 0.2f + 0.9f;
	light->attenuation[2] = light_default_quadratic_attenuation;
	Register_Light(light);

	if ((int)(shot->timer) % 5 == 0)
	{
		//1 and 16 chance of emitting a particle
		//if (MRand() & 15 == 15)

		//1 and 64 chance of emitting a particle
		if (MRand() & 63 == 63)
		{
			vector_t vel;
			VectorSet(&vel, -MFRand(), MCRand(), MCRand());
			MatrixMul(&shot->matrix, &vel, &vel);

//			vel =	(shot->axis[0] * -MFRand()) +
//					(shot->axis[1] * MCRand()) +
//					(shot->axis[2] * MCRand());

			Particle_SetTexture(particleSmokeTex);
			Particle_SetBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

			Particle_Create(&shot->position, &vel, 50.f + 150.f * MFRand(), 0x7FFFFFFF, MRand()%150 + 150, true);
			
			Particle_SetBlend(GL_SRC_ALPHA, GL_ONE);
		}
	}

	//create the base particle
	{
		vector_t pos;

		Particle_SetTexture(particleFlameTex);
		
		pos.x = shot->position.x + MCRand();
		pos.y = shot->position.y + MCRand();
		pos.z = shot->position.z + MCRand();
		Particle_Create(&pos, &shot->velocity, 150.f * lum * MFRand(), 0xFF3FBFFF, 3, false);

		pos.x = shot->position.x + MCRand();
		pos.y = shot->position.y + MCRand();
		pos.z = shot->position.z + MCRand();
		Particle_Create(&shot->position, &shot->velocity, 150.f * lum * MFRand(), 0xFF3FBFFF, 3, false);
	
		pos.x = shot->position.x + MCRand();
		pos.y = shot->position.y + MCRand();
		pos.z = shot->position.z + MCRand();
		Particle_Create(&shot->position, &shot->velocity, 150.f * lum * MFRand(), 0xFF3FBFFF, 3, false);
	}

	shot->timer--;
	if (shot->timer < 0) RemoveEntity(shot);
}

/* SHOT ALT */

void ShockwaveEffectThink(entity_t *e)
{
	e->timer++;
	if (e->timer > 100)
	{
		RemoveEntity(e);
		return;
	}

	float x = (float)e->timer / 100.f;

	float alpha = 2 - 2 * x;
	if (alpha > 1) alpha = 1;
	e->skin[0].color.rgba[3] = (unsigned char)(255.f * alpha);
	e->skin[1].color.rgba[3] = e->skin[0].color.rgba[3];

	x *= 2;
	VectorSet(&e->scale, .125f*x,x,x);
	Entity_SetScale(e, &e->scale);
}

void ShotAlt_Explode(entity_t *shot, vector_t *normal)
{
	const static scalar_t shotAltRadius = 300;//1500;

	//do something important

	Chain *link, *next;
	for (link = entLiveChain.head; link; link = next) {
		next = link->chain_next;
		entity_t *e = ENTITY_FROM_LINK(link);
		if (!EntIsActive(e)) continue;

		//dont hit entities created this frame
		if (e->creationTime == gameTime) continue;

		//dont both inflict damage upon itself
		if (e == shot) continue;

		if (!e->check_collision) continue;

		//apply forces to all colliding objects in the area (should we?)
		{
			vector_t force;
			VectorSet(&force, MCRand(),MCRand(),MCRand());
			VectorAdd(&force, &shot->position, &force);
			world->Delta(&e->position, &force, &force);
			VectorScale(&force, 20000000.f / VectorLengthSq(&force), &force);

			//realign realworld position (thanks modulo space)
			vector_t position;
			world->Relative(&shot->position, &e->position, &position);
			ApplyForce(e, &position, &force);
		}
		
		if (e->takesdamage)
		{
			scalar_t del = world->DistSq(&e->position, &shot->position);

			scalar_t rad = shotAltRadius + Entity_CollisionMaxRadius(e) + Entity_CollisionMaxRadius(shot);
			if (del > rad * rad) continue;

			//break the rock and remove the self
			Particle_Explosion(&e->position, 3, 10);
			Deal_Damage(e, 10);
		}
	}

	//create a shockwave
/*
	for (int i = 0; i < 1; i++)
	{
		entity_t *e = CreateEntity();
		if (!EntIsActive(e)) continue;

		//set think
		e->Think = ShockwaveEffectThink;

		//set position
		Entity_SetPosition(e, &shot->position);

		//set angle
		QuatVectorDifference(&vectorBaseAxis[0], normal, 1.f, &e->angle);
		Entity_SetAngle(e, &e->angle);

		//set scale
		VectorSet(&e->scale, 0,0,0);
		Entity_SetScale(e, &e->scale);

		//set model
		Entity_SetModel(e, "WorldData/Models/disk.mdl");

		//set skin 0
		e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/shockwave_effect.tga");
		e->skin[0].color.l = 0xFF0000FF;
		e->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE);//GL_SRC_ALPHA, GL_ONE);//GL_SRC_ALPHA,GL_ONE);
		e->skin[0].flags = SF_BLEND;

		//set skin 1
		//e->skin[1].tex = Entity_LoadSkin("WorldData/Skins/shot.bmp");
		//e->skin[1].func = CompressBlend(GL_DST_COLOR, GL_ZERO);//GL_SRC_ALPHA,GL_ONE);
		//e->skin[1].flags = SF_BLEND;

		//set skin count
		e->skin_count = 1;

		//type
		strcpy(e->type, "shockwave_effect");
	}
*/
	Particle_ExplosionAlt(&shot->position);

	RemoveEntity(shot);
}

void ShotAlt_Touch(entity_t *shot, entity_t *touch, collision_info_t *info)
{
	if (touch == shot->owner) return;

	ShotAlt_Explode(shot, &info->normal);
}

//if it takes damage then make it go off the next think frame
void ShotAlt_Death(entity_t *shot)
{
	shot->takesdamage = false;
	VectorCopy(&vectorZero, &shot->velocity);
	shot->timer = 0;
	//already set:
//	shot->Think = ShotAlt_Explode;
}

void ShotAlt_Think(entity_t *shot)
{
	if ((int)shot->timer & 16)
	{
		lightinfo_t *light = Create_Light();
		light->position[0] = shot->position.x;
		light->position[1] = shot->position.y;
		light->position[2] = shot->position.z;
		light->diffuse[0] = 1.f;
		light->diffuse[1] = 0.f;
		light->diffuse[2] = 0.f;
		light->attenuation[0] = 0.8f;
		light->attenuation[2] = light_default_quadratic_attenuation;
		Register_Light(light);

		Particle_SetTexture(particleFlameTex);

		Particle_Create(&shot->position, &vectorZero, 150.f, 0xEF0000FF, 2, false);
		Particle_Create(&shot->position, &vectorZero, 25.f, 0xFFFFFFFF, 2, false);
	}

	shot->timer--;

	if (shot->timer < 0)
	{
		vector_t vrand;
		VectorSet(&vrand, MFRand(), MFRand(), MFRand());
		VectorUnit(&vrand,&vrand);
		ShotAlt_Explode(shot, &vrand);
	}
}

void TempShot_Think(entity_t *shot)
{
	shot->timer--;
	if (shot->timer < 0) RemoveEntity(shot);
}

/* PLAYER FUNCTIONS */

void Player_Fire_Temp(entity_t *owner)
{
	if (owner->timer > gameTime) return;

	owner->timer = gameTime + 10.0;

	entity_t *e = CreateEntity();
	if (!EntIsActive(e)) return ;

	Entity_SetModel(e, "WorldData/Models/c_rock.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_rock.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/shot.bmp");

	Entity_SetAngle(e, &owner->angle );

	vector_t newpos;
	VectorMultAdd(&owner->position, &owner->matrix.vec[0], 100.f, &newpos);
	VectorAdd(&newpos, &owner->velocity, &newpos);
	Entity_SetPosition(e, &newpos);

	VectorScale(&owner->matrix.vec[0], 5, &e->velocity);
	
	e->check_collision = true;
	e->bounces = true;

//	VectorSet(&e->scale, 25, 25, 25);
//	Entity_SetScale(e,  &e->scale );
	Entity_SetDensityInverse(e, 1);

	e->timer = 1000;
	e->Think = TempShot_Think;
	e->owner = owner;

	strcpy(e->type, "temp");
}


void Player_Fire_Rocket(entity_t *owner)
{
	if (owner->timer > gameTime) return;

	if (owner->ammo[AMMO_ROCKETS] <= 0) return;
	owner->ammo[AMMO_ROCKETS]--;

	owner->timer = gameTime + 30.0;

	entity_t *e = CreateEntity();
	if (!EntIsActive(e)) return ;

	Entity_SetModel(e, "WorldData/Models/shot_64.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_shot.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/shot.bmp");

	Entity_SetAngle(e, &owner->angle );

	vector_t newpos;
	VectorMultAdd(&owner->position, &owner->matrix.vec[0], 100.f, &newpos);
	VectorAdd(&newpos, &owner->velocity, &newpos);
	Entity_SetPosition(e, &newpos);
//	Entity_SetPosition(e, owner->axis[0] * 100 + owner->position + owner->velocity);

	VectorMultAdd(&owner->velocity, &e->matrix.vec[0], 20, &e->velocity);
	VectorScale(&e->matrix.vec[0], 5 * (scalar_t)DEGREE_TO_RADIAN, &e->rotation);
	
	e->check_collision = true;
	e->Touch = Shot_Touch;

	e->timer = 100;
	e->Think = Shot_Think;
	e->owner = owner;

	strcpy(e->type, "shot");
}


void Player_Fire_Bomb(entity_t *owner)
{
	if (owner->timer > gameTime) return;

	if (owner->ammo[AMMO_BOMBS] <= 0) return;
	owner->ammo[AMMO_BOMBS]--;

	owner->timer = gameTime + 10.0;

	entity_t *e = CreateEntity();
	if (!EntIsActive(e)) return ;

	Entity_SetModel(e, "WorldData/Models/shot_64.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_shot.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/ship.bmp");

	Entity_SetAngle(e, &owner->angle );

	vector_t newpos;
	VectorMultAdd(&owner->position, &owner->matrix.vec[2], -10, &newpos);
	Entity_SetPosition(e, &newpos);

	VectorScale(&owner->matrix.vec[1], 5 * (scalar_t)DEGREE_TO_RADIAN, &e->rotation);

	VectorMultAdd(&owner->velocity, &owner->matrix.vec[0], 5.f, &e->velocity);
	VectorMultAdd(&e->velocity, &owner->matrix.vec[2], -0.2f, &e->velocity);

	e->check_collision = true;

	e->takesdamage = true;
	e->health = e->max_health = 1;
	e->Death = ShotAlt_Death;

	e->timer = 160;
	e->Think = ShotAlt_Think;
	e->Touch = ShotAlt_Touch;

	e->owner = owner;
	strcpy(e->type, "shot_alt");
}


void TempShotAlt2_Touch(entity_t *shot, entity_t *touch, collision_info_t *info)
{
	if (!EntIsActive(touch)) return ;
	if (touch == shot->owner) return;

	vector_t force;
	VectorScale(&shot->velocity, 1000.f, &force);

	//realign realworld position (thanks modulo space)
	vector_t position;
	world->Relative(&shot->position, &touch->position, &position);

	ApplyForce(touch, &position, &force);

	Deal_Damage(touch, 0.05f);

	Particle_Explosion(&shot->position, 3, 2);

	shot->timer = 30;
	shot->Think = Shot_Bright;
	shot->Touch = NULL;
	Entity_SetModel(shot, NULL);
}

void TempShotAlt2_Think(entity_t *shot)
{
	shot->timer--;
	if (shot->timer < 0) RemoveEntity(shot);
}

void Player_Fire_Gatling(entity_t *owner)
{

	if (owner->timer > gameTime) return;

	if (owner->ammo[AMMO_GATLING] <= 0) return;
	owner->ammo[AMMO_GATLING]--;

	owner->timer = gameTime + 1.0;

	scalar_t side = ((int)gameTime & 2) ? -1.f : 1.f;

	vector_t source;
	VectorMultAdd(&owner->position, &owner->matrix.vec[0], 20, &source);
	VectorMultAdd(&source, &owner->matrix.vec[1], 20 * side, &source);

	lightinfo_t *light = Create_Light();
	light->position[0] = source.x;
	light->position[1] = source.y;
	light->position[2] = source.z;
	light->diffuse[0] = 1.f;
	light->diffuse[1] = 0.f;
	light->diffuse[2] = 0.f;
	light->attenuation[0] = 0.8f;
	light->attenuation[2] = light_default_quadratic_attenuation;
	Register_Light(light);

	vector_t vel;
	VectorSet(&vel, MFRand() * 0.1f, MCRand() * 0.1f, MCRand() * 0.1f);
	MatrixMul(&owner->matrix, &vel, &vel);
//	vel = (owner->axis[0] * MFRand()) + (owner->axis[1] * MCRand()) + (owner->axis[2] * MCRand()) * 0.1f;

	vector_t pos;
	VectorMultAdd(&source, &owner->matrix.vec[0], 5, &pos);

	//smoke
	Particle_SetTexture(particleSmokeTex);
	Particle_SetBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	Particle_Create(&pos, &vel, 20.f + MFRand() * 10.f, 0x7FFFFFFF, 60 + MRand() & 63, true);
	Particle_SetBlend(GL_SRC_ALPHA, GL_ONE);

	//fire
	Particle_Create(&source, &vectorZero, 50.f, 0xFF3FBFFF, 2, false);
	Particle_Create(&source, &vectorZero, 40.f, 0xFF3FBFFF, 2, false);
	Particle_Create(&source, &vectorZero, 30.f, 0xFF3FBFFF, 2, false);

	//temp
	entity_t *e = CreateEntity();
	if (!EntIsActive(e)) return;

	Entity_SetModel(e, "WorldData/Models/shot.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_shot.mdl");
	e->skin[0].color.l = 0xFF4CB3FF;

	Entity_SetAngle(e, &owner->angle );

	VectorMultAdd(&source, &owner->matrix.vec[0], 100, &pos);
	Entity_SetPosition(e, &pos);

	VectorMultAdd(&owner->velocity, &e->matrix.vec[0], 100.f, &e->velocity);

	e->check_collision = true;
	
	e->timer = 20;
	e->Think = TempShotAlt2_Think;
	e->Touch = TempShotAlt2_Touch;

	e->owner = owner;

	strcpy(e->type, "shotalt2");
}

void GrapplingHook_Think(entity_t *shot)
{
	shot->timer--;
	if (shot->timer < 0)
	{
		RemoveEntity(shot);
		return;
	}
	if (!EntIsActive(shot->owner))
	{
		RemoveEntity(shot);
		return;
	}
	if (shot->owner->hook_ent != shot)
	{
		RemoveEntity(shot);
		return;
	}

	Particle_Create(&shot->position, &vectorZero, 25.f*(scalar_t)sin(shot->timer*0.2f), 0xFF0FFF0F, 50, true);
}

//#define MAKE_THE_HOOK_BROKEN

void GrapplingHook_Hooked(entity_t *shot)
{
	shot->timer--;
	if (shot->timer < 0)
	{
		RemoveEntity(shot);
		return;
	}
	if (!EntIsActive(shot->owner))
	{
		RemoveEntity(shot);
		return;
	}
	if (shot->owner->hook_ent != shot)
	{
		RemoveEntity(shot);
		return;
	}
	if (!EntIsActive(shot->hook_ent))
	{
		RemoveEntity(shot);
		return;
	}

	static scalar_t hookSpringScalar = 1000.f;
	static scalar_t hookSpringDecay = 0.99f;
	static scalar_t hookSpringMaxForce = 5000000;	//whats a good # for this anyways?
	static CVarw *fhldsfd[] = {
		new CVarw(&hookSpringScalar,	"hs"),
		new CVarw(&hookSpringDecay,	"hd"),
		new CVarw(&hookSpringMaxForce,	"hm"),
		NULL};

	vector_t hookpos, force_lin, force_ang, force_sum, vel;
	scalar_t force_len, force_del;

//apply force based upon spring const
	//calc the offset
	MatrixMul(&shot->hook_ent->matrix, &shot->hook_ofs, &hookpos);
	//and calc the true pos
	VectorAdd(&shot->hook_ent->position, &hookpos, &hookpos);

	if (((int)shot->timer & 15) == 0)
		Particle_Create(&hookpos, &vectorZero, 100.f, 0xFF0FFF0F, 50, true);

	world->Delta(&shot->owner->position, &hookpos, &force_lin);
	force_len = VectorLength(&force_lin);

//	`->Print("force_len: %f\n", force_len);
//	do not allow springs greater than 1000.f. if so then snap the rope.
#if !defined MAKE_THE_HOOK_BROKEN
	if (force_len > 1000.f)
	{
		RemoveEntity(shot);
		return;
	}
#endif

	force_del = force_len - shot->hook_len;
//	shot->hook_len += force_del * 0.1f;					//adjust hook_len - make the spring stretch (will this be bad?)
	VectorScale(&force_lin, force_del * hookSpringScalar / force_len, &force_lin);

//apply force based upon spring torque
	world->Delta(&hookpos, &shot->owner->position, &vel);	//vel holds vector from owner->pos to hookpos
//	force = axis cross arm
//	axis = arm cross dest
//	force = (arm cross dest) cross arm = dest |arm|^2 - arm (dest dot arm)
	VectorScale(&shot->owner->matrix.vec[0], VectorLengthSq(&vel), &force_ang);
	VectorMultAdd(&force_ang, &vel, -VectorDot(&shot->owner->matrix.vec[0], &vel), &force_ang);

//finally, apply the force
	VectorAdd(&force_lin, &force_ang, &force_sum);
	force_len = VectorLength(&force_sum);
//	main_console->Print("hook force: %f\n", force_len);
#if !defined MAKE_THE_HOOK_BROKEN
	if (force_len > hookSpringMaxForce)
		VectorScale(&force_sum, hookSpringMaxForce / force_len, &force_sum);
#endif
	ApplyForce(shot->hook_ent, &hookpos, &force_sum);
	VectorScale(&shot->hook_ent->velocity, hookSpringDecay, &shot->hook_ent->velocity);
	VectorScale(&shot->hook_ent->rotation, hookSpringDecay, &shot->hook_ent->rotation);

#if !defined MAKE_THE_HOOK_BROKEN
//apply linear force to the player - negative and scale it down a lot (generous to the person using the hook)
	//no more generousity
	//VectorScale(&force_lin, -0.1f, &force_lin);
	VectorScale(&force_lin, -1, &force_lin);
	world->Relative(&hookpos, &shot->owner->position, &hookpos);
	ApplyForce(shot->owner, &hookpos, &force_lin);
#endif

//	//flip force for other ent
//	VectorScale(&force, -1, &force);
//	//realign realworld position (thanks modulo space)
//	world->Relative(&realworld, &shot->owner->position, &realworld);
//	ApplyForce(shot->owner, &realworld, &force);
//	VectorSet(&vel, MCRand(), MCRand(), MCRand());
//	VectorUnit(&vel, &vel);

	//no more retraction
//	shot->hook_len -= 2.f;
//	if (shot->hook_len < 0) shot->hook_len = 0;
}

void GrapplingHook_Touch(entity_t *shot, entity_t *touch, collision_info_t *info)
{
	//dont grab a non-collider
	if (!touch->check_collision) return;
	//dont grab the owner
	if (touch == shot->owner) return;

	shot->hook_ent = touch;
	shot->hook_len = 500.f;
//	shot->hook_len = 0.5f * (scalar_t)sqrt(world->DistSq(&shot->owner->position, &touch->position));

	world->Delta(&shot->position, &touch->position, &shot->hook_ofs);
	//transpose mul - because the transpose of an orthogonal matrix is its inverse
	//we want to inverse mul so we can find the coordinates of the hook offset in the touch's modelspace
	MatrixTMul(&touch->matrix, &shot->hook_ofs, &shot->hook_ofs);

	//initial hook resting length starts at 500
	//we want complete decay within a timeframe
	//decay of 1 len/tick, for 500 ticks = len of 500.
	//2 len/tick is approx 60 len/second,          _
	//therefore total decay occurs within approx 8.3 seconds
	//we will set a max life of 15 seconds, just in case
	shot->timer = 15*30;
	shot->Think = GrapplingHook_Hooked;

	shot->check_collision = false;
}

void Player_Fire_Hook(entity_t *owner)
{
	if (owner->timer > gameTime) return;
	owner->timer = gameTime + 20.0;

	if (EntIsActive(owner->hook_ent))
	{
		RemoveEntity(owner->hook_ent);
		return;
	}

	vector_t source;
	VectorMultAdd(&owner->position, &owner->matrix.vec[0], 20, &source);

	lightinfo_t *light = Create_Light();
	light->position[0] = source.x;
	light->position[1] = source.y;
	light->position[2] = source.z;
	light->diffuse[0] = 1.f;
	light->diffuse[1] = 0.f;
	light->diffuse[2] = 0.f;
	light->attenuation[0] = 0.8f;
	light->attenuation[2] = light_default_quadratic_attenuation;
	Register_Light(light);

	//temp
	entity_t *e = CreateEntity();
	if (!EntIsActive(e)) return;

	Entity_SetModel(e, "WorldData/Models/shot.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_shot.mdl");
	e->skin[0].color.l = 0xFF7FFF7F;

	Entity_SetAngle(e, &owner->angle );

	VectorMultAdd(&source, &owner->matrix.vec[0], 100, &source);
	Entity_SetPosition(e, &source);

	VectorMultAdd(&owner->velocity, &e->matrix.vec[0], 20.f, &e->velocity);

	e->check_collision = true;
	
	e->timer = 40;		//40 ticks * 20 units/tick = 800 max units.  just shy of the 1000 snapping distance
	e->Think = GrapplingHook_Think;
	e->Touch = GrapplingHook_Touch;

	//keep this linked to owner
	owner->hook_ent = e;
	e->owner = owner;

	strcpy(e->type, "graphook");
}

void DrunkMissile_Touch(entity_t *shot, entity_t *touch, collision_info_t *info)
{
	Shot_Touch(shot,touch,info);
}

void DrunkMissile_Think(entity_t *shot)
{
	/*
	local	float 	points;
	local	entity	head;
	local   vector  org, dir;
	local	float	len;

	head = findradius(self.origin, 300);
	
	while (head)
	{
		if (head.classname == "scram_eggs")
		{
			org = head.origin + (head.mins + head.maxs)*0.5;
			points = 0.5*vlen (self.origin - org);
			if (points < 0) points = 0;
			points = 300 - points;
			dir = head.origin - self.origin;
			dir = normalize(dir);
			head.flags = head.flags - (head.flags & FL_ONGROUND);
			points = points * 0.2;
			head.velocity = head.velocity - dir*points;
		}
		head = head.chain;
	}

	self.velocity = normalize(self.velocity);
	self.velocity = self.velocity * 1000;

	self.angles = vectoangles(self.velocity);
	*/
}

void Player_FireDrunkMissiles(entity_t *owner)
{
	if (owner->timer > gameTime) return;
	owner->timer = gameTime + 1.0;

	vector_t source;
	VectorMultAdd(&owner->position, &owner->matrix.vec[0], 20, &source);

	//light
	lightinfo_t *light = Create_Light();
	light->position[0] = source.x;
	light->position[1] = source.y;
	light->position[2] = source.z;
	light->diffuse[0] = 1.f;
	light->diffuse[1] = 0.f;
	light->diffuse[2] = 0.f;
	light->attenuation[0] = 0.8f;
	light->attenuation[2] = light_default_quadratic_attenuation;
	Register_Light(light);

	vector_t vel;
	VectorSet(&vel, MFRand() * 0.1f, MCRand() * 0.1f, MCRand() * 0.1f);
	MatrixMul(&owner->matrix, &vel, &vel);

	vector_t pos;
	VectorMultAdd(&source, &owner->matrix.vec[0], 5, &pos);

	//smoke
	Particle_SetTexture(particleSmokeTex);
	Particle_SetBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	Particle_Create(&pos, &vel, 20.f + MFRand() * 10.f, 0x7FFFFFFF, 60 + MRand() & 63, true);
	Particle_SetBlend(GL_SRC_ALPHA, GL_ONE);

	//fire
	Particle_Create(&source, &vectorZero, 50.f, 0xFF3FBFFF, 2, false);
	Particle_Create(&source, &vectorZero, 40.f, 0xFF3FBFFF, 2, false);
	Particle_Create(&source, &vectorZero, 30.f, 0xFF3FBFFF, 2, false);

	for (int i = 0; i < 5; i++)
	{
		//temp
		entity_t *e = CreateEntity();
		if (!EntIsActive(e)) return;

		Entity_SetModel(e, "WorldData/Models/shot.mdl");
		Entity_SetCModel(e, "WorldData/Models/c_shot.mdl");
		e->skin[0].color.l = 0xFF4FB3FF;

		Entity_SetAngle(e, &owner->angle );

		VectorMultAdd(&source, &owner->matrix.vec[0], 100, &pos);
		Entity_SetPosition(e, &pos);

		VectorMultAdd(&owner->velocity, &e->matrix.vec[0], 100.f, &e->velocity);

		e->check_collision = true;
		
		e->timer = 20;
		e->Think = DrunkMissile_Think;
		e->Touch = DrunkMissile_Touch;

		e->owner = owner;

		strcpy(e->type, "shotalt2");
	}
}


void Particle_DrawField(const vector_t *v)
{
	static scalar_t	particleFieldSize = 500;
	static scalar_t	particleFieldRad = 5;
	static int		particleFieldLife = 2;
	static int		particleFieldCount = 10;

	static CVarw *fdsjkfjlkfasdl[] = {
		new CVarw(&particleFieldSize,		"parFieldSize"),
		new CVarw(&particleFieldRad,		"parFieldRad"),
		new CVarw(&particleFieldLife,		"parFieldLife"),
		new CVarw(&particleFieldCount,		"parFieldCount"),
		NULL};

	if (!menu_starfield_enable) return;

	int fx = (int)(v->x / particleFieldSize);
	int fy = (int)(v->y / particleFieldSize);
	int fz = (int)(v->z / particleFieldSize);

	Particle_SetTexture(particleFlameTex);

	for (int i = -1; i <= 1; i++)
	{
		for (int j = -1; j <= 1; j++)
		{
			for (int k = -1; k <= 1; k++)
			{
				int seed = (fx + i) + (fy + j) * 499 + (fz + k) * 4999;
				MSRand(seed);

				for (int t = 0; t < particleFieldCount; t++)
				{
					vector_t pos;
					VectorSet(&pos,
						(scalar_t)(fx + i) + MCRand(),
						(scalar_t)(fy + j) + MCRand(),
						(scalar_t)(fz + k) + MCRand());
					VectorScale(&pos, particleFieldSize, &pos);

					Particle_Create(&pos, &vectorZero, particleFieldRad, 0xFFFFFFFF, particleFieldLife, false);
				}
			}
		}
	}

	//re-randomize
	MSRand((int)(main_timer->GetTime() * 1000.0));
}

void PlayerDust_Think(entity_t *dust)
{
	dust->timer--;

	//1 and 16 chance of emitting a particle
	if (MRand() & 15 == 15)
	{
		color_t c;
		c.rgba[0] = 0xFF;
		c.rgba[1] = MRand() & 0x7F;
		c.rgba[2] = c.rgba[1] >> 1;
		c.rgba[3] = 0xFF;

		Particle_SetTexture(particleSmokeTex);

		Particle_Create(&dust->position, &vectorZero, 30*MCRand()+1, c.l, MRand()%50, true);
	}
	
	if (dust->timer < 0)
	{
		dust->Think = Fade_Think;
		dust->timer = 60;
	}
}

void StartItem_Ammo(entity_t *e);

void Player_Death(entity_t *player)
{
	if (Client_Invincible) return;

	// DIE!!!

	for (int v = 0; v < 20; v++)
	{
		entity_t *e = CreateEntity();
		if (!EntIsActive(e)) return ;

		Entity_SetModel(e, "WorldData/Models/rock.mdl");
		Entity_SetCModel(e, "WorldData/Models/c_rock.mdl");

		e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/ship.bmp");
		e->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		e->skin[0].color.l = 0x7FFFFFFF;
		e->skin[0].flags = SF_BLEND | SF_LIGHTING;

		e->skin[1].tex = world->GetSkyTex();
		e->skin[1].type = ST_CUBEMAP_REFLECTION;
		e->skin[1].func = CompressBlend(GL_ONE, GL_ONE);
		e->skin[1].flags = SF_BLEND;

		e->skin_count = 2;

		quat_t reg;
		QuatSet(&reg, MCRand(),MCRand(),MCRand(),MCRand());
		Entity_SetAngle(e, &reg );

		VectorSet(&e->rotation, MCRand(),MCRand(),MCRand());
		VectorScale(&e->rotation, (scalar_t)DEGREE_TO_RADIAN, &e->rotation);
		//e->rotation = Vector(MCRand(), MCRand(), MCRand()) * (scalar_t)DEGREE_TO_RADIAN;

		strcpy(e->type, "dust");

		VectorSet(&reg.vec, MCRand(),MCRand(),MCRand());
		VectorAdd(&player->velocity, &reg.vec, &e->velocity);
		e->timer = 100;
		e->Think = PlayerDust_Think;

		e->scale.x = 0.1f + 0.1f * MFRand();
		e->scale.y = e->scale.z = e->scale.x;
		Entity_SetScale(e, &e->scale );

		e->check_collision = false;

		VectorSet(&reg.vec, MCRand(),MCRand(),MCRand());
		VectorUnit(&reg.vec, &reg.vec);
		VectorScale(&reg.vec, Entity_CollisionMaxRadius(e), &reg.vec);
		VectorAdd(&reg.vec, &player->position, &reg.vec);
		Entity_SetPosition(e, &reg.vec);
	}

	//make some random weapons
	static int ammomax[AMMO_COUNT] = {
		2500,
		25,
		5
	};
	for (int ammotype = 0; ammotype < AMMO_COUNT; ammotype++) {
		while (player->ammo[ammotype]) {
			entity_t *e = CreateEntity();
			if (EntIsActive(e)) {
				Entity_SetPosition(e, &player->position);
				e->velocity.x = 10.f * MCRand();
				e->velocity.y = 10.f * MCRand();
				e->velocity.z = 10.f * MCRand();
				StartItem_Ammo(e);
				e->ammo[ammotype] = player->ammo[ammotype];
				if (e->ammo[ammotype] > ammomax[ammotype]) {
					e->ammo[ammotype] = ammomax[ammotype];
				}
				player->ammo[ammotype] -= e->ammo[ammotype];
			}
		}
	}

	RemoveEntity(player);
}

void Player_Think(entity_t *player)
{
	vector_t pos;
	VectorAdd(&main_camera->position, &player->velocity, &pos);
	Particle_DrawField(&pos);

	//decay
	VectorScale(&player->velocity, (1 - Client_Accel / Client_MaxVel), &player->velocity);//0.9975f;
	VectorScale(&player->rotation, 0.75f, &player->rotation);
}

void PlayerSetSkin(entity_t *player)
{
	player->skin[0].tex = Entity_LoadSkin("WorldData/Skins/ship.bmp");
	player->skin[0].color.l = 0xFFFFFFFF;
	player->skin[0].specular.l = 0xFFFFFFFF;
	player->skin[0].shininess = 128;
	player->skin[0].type = ST_NONE;
	player->skin[0].func = 0;
	player->skin[0].flags = SF_LIGHTING;

	player->skin[1].tex = world->GetSkyTex();
	player->skin[1].color.l = 0xFFFFFFFF;
	player->skin[1].specular.l = 0xFFFFFFFF;
	player->skin[1].shininess = 128;
	player->skin[1].type = ST_CUBEMAP_REFLECTION;
	player->skin[1].func = CompressBlend(GL_ONE, GL_ONE);
	player->skin[1].flags = SF_BLEND;

	player->skin_count = 2;

	player->final_depth_pass = false;
}

void PlayerSetSkinInvis(entity_t *player)
{
/*
	player->skin_count = 0;

	player->final_depth_pass = true;
*/
	player->skin[0].tex = world->GetSkyTex();
	player->skin[0].color.l = 0x7FFFFFFF;
	player->skin[0].specular.l = 0;
	player->skin[0].shininess = 0;
	player->skin[0].type = ST_CUBEMAP_REFLECTION;
	player->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE);
	player->skin[0].flags = SF_BLEND;

	player->skin_count = 1;

	player->final_depth_pass = true;
}

void StartPlayer(entity_t *player)
{

	if (player == client_ent)
	{
//tweak the camera
		main_camera->trackPosition = &player->position;
		main_camera->trackAngle = &player->angle;
		main_camera->trackMethod = CAM_TRACK_CHASE;
		main_camera->trackLerpC = 1.f;
		main_camera->trackSlerpC = 0.1f;
		VectorSet(&main_camera->chaseOffsetPos, -50, 0, 20);
		QuatCopy(&quatIdentity, &main_camera->chaseOffsetAngle);
	}

	Entity_SetModel(player, "WorldData/Models/ship_224.mdl");
	Entity_SetCModel(player, "WorldData/Models/c_ship.mdl");

	PlayerSetSkin(player);

	player->takesdamage = true;
	player->check_collision = true;
	player->bounces = true;

	player->max_health = player->health = 25;
	player->ammo[AMMO_BOMBS] = 2;

	player->Death = Player_Death;

	player->Think = Player_Think;

	Entity_SetPosition(player, &vectorZero);

	strcpy(player->type, "player");

	player->lastRockTouch = 0;

	Entity_SetDensityInverse(player, 0.1f);
}

/* ROCK FUNCTIONS */

void SmallRock_Think(entity_t *rock)
{
	rock->timer--;

	//1 and 16 chance of emitting a particle
	if (MRand() & 15 == 15)
	{
		float size = MCRand();

		Particle_SetTexture(particleSmokeTex);

		vector_t vel;
		VectorSet(&vel, MCRand(),MCRand(),MCRand());
		VectorScale(&vel, 3.f*(1-size), &vel);
		Particle_Create(&rock->position, &vel, 30*size+1, 0x7F0000CF, MRand()%50, true);
	}
	
	if (rock->timer < 0)
	{
		rock->Think = Fade_Think;
		rock->timer = 60;
	}
}

void RockBounce(entity_t *rock, entity_t *touch, collision_info_t *info, collision_result_t *result)
{
	if (!EntIsActive(touch)) return;

	//if (!strcmp(touch->type, "player")) 
	if (touch->takesdamage)	// && memcmp(touch->type, "rock", 4))
	{
		const static double timeThreshold = 1.0;
		if ((gameTime - touch->lastRockTouch) < timeThreshold) return;
		touch->lastRockTouch = gameTime;

//		scalar_t damage = collisioninfo_impulse_mag;
		scalar_t damage = result->impulse_mag;

//		damage /= 500000.f;		//damage is one millionth the force - time to work on conversions?
		damage /= 5000000.f;
//		main_console->Print("rock collision impulse of %f\n", result->impulse_mag);

		Deal_Damage(touch, damage);
	}
}

void StartLargeRock(entity_t *e)
{
	Entity_SetModel(e, "WorldData/Models/rock_512.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_rock.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/rock.bmp");
	e->skin[0].color.l = 0xFFFFFFFF;//0xFF3F3F3F;

	vector_t reg;

	VectorSet(&e->rotation, MCRand(), MCRand(), MCRand());
	VectorScale(&e->rotation, (scalar_t)DEGREE_TO_RADIAN * 0.1f, &e->rotation);

	reg.x = 6.f+6.f*MFRand();
	reg.y = reg.z = reg.x;
	Entity_SetScale(e, &reg );

	e->velocity.x = MCRand() * (4.f / e->scale.x);
	e->velocity.y = MCRand() * (4.f / e->scale.y);
	e->velocity.z = MCRand() * (4.f / e->scale.z);

	e->check_collision = true;
	e->bounces = true;
	e->takesdamage = true;

	e->max_health = e->health = 10;

	e->Think = RockThink;
	e->Bounce = RockBounce;
	e->Death = RockDeath;

	strcpy(e->type, "rock_l");

	VectorSet(&reg, MCRand(), MCRand(), MCRand());
	VectorScale(&reg, world->WorldSize, &reg);
	Entity_SetPosition(e, &reg);
	Entity_SetDensityInverse(e, 1.f);
}

void StartMediumRock(entity_t *e, entity_t *rock)
{
	Entity_SetModel(e, "WorldData/Models/rock_512.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_rock.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/rock_med.bmp");
	e->skin[0].color.l = 0xFFFFFFFF;//0xFF7F7F7F;

	VectorSet(&e->velocity, MCRand()* 0.2f, MCRand()* 0.2f, MCRand()* 0.2f);
	VectorSet(&e->rotation,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN);

	e->check_collision = true;
	e->bounces = true;
	e->takesdamage = true;

	e->max_health = e->health = 5;

	e->Think = RockThink;
	e->Bounce = RockBounce;
	e->Death = RockDeath;

	strcpy(e->type, "rock_m");

	vector_t reg;
	VectorScale(&rock->scale, (0.3f + 0.2f * MFRand()), &reg);

	Entity_SetScale(e, &reg );

	VectorScale(&e->velocity, (2 + MFRand()), &e->velocity);

	VectorSet(&reg, MCRand(), MCRand(), MCRand());
	VectorUnit(&reg, &reg);
	VectorScale(&reg, Entity_CollisionMaxRadius(e), &reg);
	VectorAdd(&reg, &rock->position, &reg);

	Entity_SetPosition(e, &reg);
	Entity_SetDensityInverse(e, 1.f);
}

void StartSmallRock(entity_t *e, entity_t *rock)
{
	Entity_SetModel(e, "WorldData/Models/rock_512.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_rock.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/rock_small.bmp");
	e->skin[0].color.l = 0xFFFFFFFF;//0xFF003F7F;

	VectorSet(&e->velocity,MCRand() * 0.2f, MCRand() * 0.2f, MCRand() * 0.2f);
	VectorSet(&e->rotation,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN);

	e->check_collision = true;
	e->bounces = true;
	e->takesdamage = true;

	e->max_health = e->health = 2;

	e->Think = RockThink;
	e->Bounce = RockBounce;
	e->Death = RockDeath;

	strcpy(e->type, "rock_s");

	vector_t reg;
	VectorScale(&rock->scale, 0.5f, &reg);
	Entity_SetScale(e, &reg );

	VectorScale(&e->velocity, 10.f, &e->velocity);

	VectorSet(&reg, MCRand(), MCRand(), MCRand());
	VectorUnit(&reg, &reg);
	VectorScale(&reg, Entity_CollisionMaxRadius(e), &reg);
	VectorAdd(&reg, &rock->position, &reg);
	Entity_SetPosition(e, &reg);

	Entity_SetDensityInverse(e,1.f);
}

void StartTinyRock(entity_t *e, entity_t *rock)
{
	Entity_SetModel(e, "WorldData/Models/rock.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_rock.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/rock_tiny.bmp");
	e->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	e->skin[0].color.l = 0xFFFFFFFF;//0x7F0000CF;
	e->skin[0].flags = SF_BLEND | SF_LIGHTING;

	quat_t reg;
	QuatSet(&reg,MCRand(),MCRand(),MCRand(),MCRand());
	Entity_SetAngle(e, &reg );

	VectorSet(&e->rotation,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN);

	strcpy(e->type, "dust");

	e->timer = 100 + MRand() % 100;
	e->Think = SmallRock_Think;

	float rad = MFRand();

	VectorScale(&rock->scale, 0.7f * (1.25f - rad), &reg.vec);

	Entity_SetScale(e, &reg.vec );

	vector_t ptemp;
	for (int v = 0; v < 3; v++) ptemp.v[v] = MCRand();
	VectorUnit(&ptemp, &ptemp);
	for (v = 0; v < 3; v++)  e->velocity.v[v] = ptemp.v[v] + rock->velocity.v[v] * 0.25f;
	VectorUnit(&e->velocity, &e->velocity);
	for (v = 0; v < 3; v++) ptemp.v[v] = ptemp.v[v] * (Entity_CollisionMaxRadius(rock) * 0.5f * rad) + rock->position.v[v];
	for (v = 0; v < 3; v++)	e->velocity.v[v] = (e->velocity.v[v] * 4.f + rock->velocity.v[v] * .75f) * (.25f + rad);

	Entity_SetPosition(e,&ptemp);
}

/* ATTEMPTED ITEM */

void ItemEffect_Think(entity_t *e)
{
	e->timer++;
	if (e->timer > 30)
	{
		RemoveEntity(e);
		return;
	}

	float frac = (float)e->timer / 30.f;

	float wave = (float)sin(M_PI * frac);

	e->skin[0].color.rgba[3] = (unsigned char)(255.f * wave);

	wave *= 0.5f;
	VectorSet(&e->scale, 0, wave,wave);
	Entity_SetScale(e, &e->scale);

	if (EntIsActive(e->owner))
	{
		VectorLerp(&e->position, &e->owner->position, 0.65f, &e->position);
		Entity_SetPosition(e, &e->position);
	}
}

void Create_ItemEffect(entity_t *item, entity_t *touch)
{
	/*	newschool item effect
	for (int i = 0; i < 3; i++)
	{
		entity_t *e = CreateEntity();
		if (!EntIsActive(e)) continue;

		//set think
		e->Think = ItemEffect_Think;

		//set position
		Entity_SetPosition(e, &item->position);

		//set angle
		quat_t mul;
		VectorCopy(&item->matrix.vec[i], &mul.vec);
		mul.w = 90;
		AngleAxisToQuat(&mul,&mul);
		QuatMul(&mul, &item->angle, &mul);
		Entity_SetAngle(e, &mul);

		//set scale
		Entity_SetScale(e, &vectorZero);

		//set model
		Entity_SetModel(e, "WorldData/Models/disk.mdl");

		//set skin
		e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/item_effect.tga");
		e->skin[0].color.l = 0x003FFFFF;
		e->skin[0].func = CompressBlend(GL_SRC_ALPHA,GL_ONE);//GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		e->skin[0].flags = SF_BLEND;

		//type
		strcpy(e->type, "item_effect");

		//owner
		e->owner = touch;
	}
	*/
	//oldschool item effect

	Particle_SetTexture(particleFlameTex);

	for (int i = 1; i < 6; i++) {
		for (int j = 0; j < 3; j++) {
			for (int k = 0; k < 2; k++) {
				vector_t v;
				VectorScale(&item->matrix.vec[j], (scalar_t)i * (k ? -1 : 1), &v);
				Particle_Create(&item->position,  &v, 30.f*(scalar_t)i, 0xFFFF7F7F, 200, true);
			}
		}
	}

}

// ***** ITEM HEALTH

void ItemHealth_Touch(entity_t *item, entity_t *touch, collision_info_t *info)
{
	if (!touch->takesdamage) return;
	if (strcmp(touch->type, "player")) return;

	touch->health = touch->max_health;

	Create_ItemEffect(item, touch);

	RemoveEntity(item);
}

void StartItem_Health(entity_t *e)
{
	Entity_SetModel(e, "WorldData/Models/cube.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_cube.mdl");

	e->skin[0].tex = world->GetSkyTex();//Entity_LoadSkin("WorldData/Skins/bluefog.bmp");
	e->skin[0].type = ST_CUBEMAP_REFLECTION;//ST_SPHEREMAP;
//	e->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//	e->skin[0].flags = SF_BLEND | SF_TWOSIDE;
	e->skin[0].specular.l = 0;

	e->skin[1].tex = Entity_LoadSkin("WorldData/Skins/health_t2.tga");
	e->skin[1].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	e->skin[1].flags = SF_BLEND | SF_LIGHTING | SF_TWOSIDE;
	e->skin[1].shininess = 0;

	e->skin_count = 2;

	//special case - because this object is so beefed up on graphics
	e->final_depth_pass = true;

	VectorSet(&e->velocity, MCRand()* 0.02f, MCRand()* 0.02f, MCRand()* 0.02f);
	VectorSet(&e->rotation,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN);

	e->check_collision = true;
	e->bounces = true;

	vector_t newscale;
	VectorSet(&newscale, 20,20,20);
	Entity_SetScale(e, &newscale );

	e->Touch = ItemHealth_Touch;

	strcpy(e->type, "item_health");

	Entity_SetDensityInverse(e,0.005f);
}

// ***** ITEM INVIS

void ItemInvis_Track(entity_t *item)
{
	entity_t *player = item->owner;

	if (!EntIsActive(player))
	{
		RemoveEntity(item);
		return;
	}

	if (item->timer < 150)	//the first 150 frames fading out
	{
		double frac = item->timer / 150.0;
		player->skin[0].color.rgba[3] = (unsigned char)(255.0 * (1.0 - frac));
		player->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		player->skin[0].flags = SF_BLEND | SF_TWOSIDE | SF_LIGHTING;
	}
	else if (item->timer < 1950) //the next 1800 as invisible
	{
		player->skin[0].color.rgba[3] = 0;
		player->skin[0].func = CompressBlend(GL_ZERO, GL_ONE);
		player->skin[0].flags = SF_BLEND | SF_TWOSIDE | SF_LIGHTING;
	}
	else if (item->timer < 2100) //the final 150 fading in
	{
		double frac = (item->timer - 1950) / 150.0;
		player->skin[0].color.rgba[3] = (unsigned char)(255.0 * frac);
		player->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		player->skin[0].flags = SF_BLEND | SF_TWOSIDE | SF_LIGHTING;
	}
	else //timer is up
	{
		player->skin[0].color.rgba[3] = 0xFF;
		player->skin[0].func = 0;
		player->skin[0].flags = SF_LIGHTING;

		RemoveEntity(item);
	}

	item->timer++;
}

void ItemInvis_Touch(entity_t *item, entity_t *touch, collision_info_t *info)
{
	if (!touch->takesdamage) return;
	if (strcmp(touch->type, "player")) return;

	Create_ItemEffect(item, touch);

	//remove this ent from all physics/rendering
	//attach it to the player
	//and set its Think func to something that makes the player skins go invis
	//WARNING - this might be very player-specific due to the skin modifying

	//disable rendering
	Entity_SetModel(item, NULL);
	item->skin_count = 0;
	item->final_depth_pass = false;

	//disable physics
	Entity_SetCModel(item, NULL);

	item->check_collision = false;
	item->bounces = false;

	VectorSet(&item->velocity, 0,0,0);
	VectorSet(&item->rotation, 0,0,0);

	item->Touch = NULL;

	//and tell it to track the player
	item->timer = 0;
	item->owner = touch;
	item->Think = ItemInvis_Track;
}

void StartItem_Invis(entity_t *e)
{
	Entity_SetModel(e, "WorldData/Models/cube.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_cube.mdl");

	e->skin[0].tex = world->GetSkyTex();
	e->skin[0].type = ST_CUBEMAP_REFLECTION;
	e->skin[0].func = CompressBlend(GL_ONE, GL_ONE);
	e->skin[0].flags = SF_BLEND | SF_TWOSIDE | SF_LIGHTING;

	//special case - due to SF_BLEND | SF_TWOSIDE
	e->final_depth_pass = true;

	VectorSet(&e->velocity, MCRand()* 0.02f, MCRand()* 0.02f, MCRand()* 0.02f);
	VectorSet(&e->rotation,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN);

	e->check_collision = true;
	e->bounces = true;

	vector_t newscale;
	VectorSet(&newscale, 20,20,20);
	Entity_SetScale(e, &newscale );

	e->Touch = ItemInvis_Touch;

	strcpy(e->type, "item_invis");

	Entity_SetDensityInverse(e,0.005f);
}

// ***** ITEM AMMO


void Item_Ammo_Touch(entity_t *item, entity_t *touch, collision_info_t *info)
{
	if (!touch->takesdamage) return;
	if (strcmp(touch->type, "player")) return;

	for (int i = 0; i < AMMO_COUNT; i++) {
		touch->ammo[i] += item->ammo[i];
	}

	Create_ItemEffect(item, touch);

	RemoveEntity(item);
}

void StartItem_Ammo(entity_t *e)
{
	Entity_SetModel(e, "WorldData/Models/cube.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_cube.mdl");

	e->skin[0].tex = world->GetSkyTex();//Entity_LoadSkin("WorldData/Skins/bluefog.bmp");
	e->skin[0].type = ST_CUBEMAP_REFLECTION;//ST_SPHEREMAP;
//	e->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//	e->skin[0].flags = SF_BLEND | SF_TWOSIDE;
	e->skin[0].specular.l = 0;

	e->skin[1].tex = Entity_LoadSkin("WorldData/Skins/ammo.tga");
	e->skin[1].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	e->skin[1].flags = SF_BLEND | SF_LIGHTING;// | SF_TWOSIDE;
	e->skin[1].shininess = 0;

	e->skin_count = 2;

	//special case - because this object is so beefed up on graphics
//	e->final_depth_pass = true;

	VectorSet(&e->velocity, MCRand()* 0.02f, MCRand()* 0.02f, MCRand()* 0.02f);
	VectorSet(&e->rotation,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN,
		MCRand() * (scalar_t)DEGREE_TO_RADIAN);

	e->check_collision = true;
	e->bounces = true;

	VectorSet(&e->scale, 20,20,20);
	Entity_SetScale(e, &e->scale);

	e->Touch = Item_Ammo_Touch;

	strcpy(e->type, "item_ammo");

	Entity_SetDensityInverse(e,0.005f);
}

// ***** ITEMS

void StartItem(entity_t *e)
{
	float f = MFRand() * 100;

	if (f < 20) {
		StartItem_Health(e);
	} else if (f < 95) {

		StartItem_Ammo(e);

		switch (MRand() % 3) {
		case 0:	e->ammo[AMMO_GATLING] = ((MRand()%5)+1) * 500;	break;
		case 1:	e->ammo[AMMO_ROCKETS] = ((MRand()%5)+1) * 5;	break;
		case 2:	e->ammo[AMMO_BOMBS] = ((MRand()%5)+1);			break;
		//set e->skin[1].tex = Entity_LoadSkin("WorldData/Skins/ammo_***.tga");
		}

	} else {
		StartItem_Invis(e);
	}
}

/* ATTEMPTED ENEMY */

void EnemyDeath(entity_t *enemy)
{
	for (int v = 0; v < 20; v++)
	{
		entity_t *e = CreateEntity();
		if (!EntIsActive(e)) return ;

		Entity_SetModel(e, "WorldData/Models/rock.mdl");
		Entity_SetCModel(e, "WorldData/Models/c_rock.mdl");

		e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/ship.bmp");
		e->skin[0].func = CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		e->skin[0].color.l = 0x7FFFFFFF;
		e->skin[0].flags = SF_BLEND | SF_LIGHTING;

		e->skin[1].tex = world->GetSkyTex();
		e->skin[1].type = ST_CUBEMAP_REFLECTION;
		e->skin[1].func = CompressBlend(GL_ONE, GL_ONE);
		e->skin[1].flags = SF_BLEND;

		e->skin_count = 2;

		quat_t reg;
		QuatSet(&reg, MCRand(),MCRand(),MCRand(),MCRand());
		Entity_SetAngle(e, &reg );

		VectorSet(&e->rotation, MCRand(),MCRand(),MCRand());
		VectorScale(&e->rotation, (scalar_t)DEGREE_TO_RADIAN, &e->rotation);

		strcpy(e->type, "dust");

		VectorSet(&reg.vec, MCRand(),MCRand(),MCRand());
		VectorAdd(&enemy->velocity, &reg.vec, &e->velocity);
		e->timer = 100;
		e->Think = PlayerDust_Think;

		e->scale.x = 0.1f + 0.1f * MFRand();
		e->scale.y = e->scale.z = e->scale.x;
		Entity_SetScale(e, &e->scale );

		e->check_collision = false;

		VectorSet(&reg.vec, MCRand(),MCRand(),MCRand());
		VectorUnit(&reg.vec, &reg.vec);
		VectorScale(&reg.vec, Entity_CollisionMaxRadius(e), &reg.vec);
		VectorAdd(&reg.vec, &enemy->position, &reg.vec);
		Entity_SetPosition(e,&reg.vec);
	}

	RemoveEntity(enemy);
}

void EnemyThink(entity_t *enemy)
{
	static int enemy_think = 1;
	static CVarw *dfsjl = new CVarw(&enemy_think, "EnemyThink");
	if (!enemy_think) return;

	//eval target
	entity_t *target = client_ent;

//	for (entity_t *e = (entity_t *)entLiveChain.tail;
//		e != (entity_t *)entLiveChain.head; e = (entity_t *)e->chain_prev)

/*	for (entity_t *e = (entity_t *)entLiveChain.head; e; e = (entity_t *)e->chain_next)
	{
		if (!EntIsActive(e)) continue;
		if (e == enemy) continue;
		if (e == client_ent) continue;
		if (!e->takesdamage) continue;

		target = e;
		break;
	}*/

	if (!EntIsActive(target)) return;

//	Vector fwd = ProjectXAxis(enemy->angle);

	//aim
	vector_t del;
	world->Delta(&target->position, &enemy->position, &del);
	//Vector del = world->Delta(target->position, enemy->position);
#if 1
	scalar_t est_time = VectorLength(&del) / /* est shot velocity */ 20.f;
	VectorMultAdd(&target->position, &target->velocity, est_time, &del);
	world->Delta(&del, &enemy->position, &del);
	//del = world->Delta(target->position + target->velocity * est_time, enemy->position);
#endif
	scalar_t dist = VectorLength(&del);
	if (dist) VectorScale(&del, 1.f / dist, &del);
//	del.Normalize();

	quat_t axis;
	axis.w = 0;
	VectorCross(&enemy->matrix.vec[0], &del, &axis.vec);
	scalar_t axis_length = VectorLength(&axis.vec);
	if (axis_length < 0.04) Player_Fire_Gatling(enemy);

	//rotate
	VectorMultAdd(&axis.vec, &enemy->matrix.vec[0], 0.05f, &axis.vec);

	VectorScale(&axis.vec, 15.f * (scalar_t)DEGREE_TO_RADIAN * 0.5f, &axis.vec);
	QuatMul(&axis, &enemy->angle, &axis);
	QuatAdd(&enemy->angle, &axis, &enemy->angle);
	Entity_SetAngle(enemy, &enemy->angle );

//	Entity_SetAngle(enemy,
//		enemy->angle + (Quat(axis * 15.f * (scalar_t)DEGREE_TO_RADIAN) * enemy->angle) * 0.5f
//	);

	//move towards target
	if (dist > 500)
		VectorMultAdd(&enemy->velocity, &enemy->matrix.vec[0], 0.05f, &enemy->velocity);	//ship accel


	Entity_RocketEffect(enemy);

	//implement decay per-object? cap speed limit per object? overall?
	VectorScale(&enemy->velocity, 0.9975f, &enemy->velocity);
	VectorScale(&enemy->rotation, 0.9f, &enemy->rotation);
}

void StartEnemy(entity_t *e)
{
	Entity_SetModel(e, "WorldData/Models/ship_224.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_ship.mdl");

	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/ship.bmp");
	e->skin[1].tex = world->GetSkyTex();
	e->skin[1].type = ST_CUBEMAP_REFLECTION;
	e->skin[1].func = CompressBlend(GL_ONE, GL_ONE);
	e->skin[1].flags = SF_BLEND;
	e->skin_count = 2;

	//cheat what the hey
	e->ammo[AMMO_ROCKETS] = 9999999;
	e->ammo[AMMO_BOMBS] = 9999999;
	e->ammo[AMMO_GATLING] = 9999999;

	e->check_collision = true;
	e->takesdamage = true;
	e->bounces = true;

	e->max_health = e->health = 20;
	e->Think = EnemyThink;
	e->Death = EnemyDeath;

	strcpy(e->type, "enemy");

	VectorSet(&e->position, 
		MCRand() * world->WorldSize,
		MCRand() * world->WorldSize,
		MCRand() * world->WorldSize);

	Entity_SetPosition(e, &e->position );
	Entity_SetDensityInverse(e, 0.1f);
}

void CFunc_SpawnEnemy(int argc, char **argv)
{
	entity_t *e = CreateEntity();
	if (EntIsActive(e)) StartEnemy(e);
}

void CFunc_Respawn(int argc, char **argv)
{
	if (client_ent) RemoveEntity(client_ent);

	client_ent = CreateEntity();
	if (!EntIsActive(client_ent)) return;
	StartPlayer(client_ent);
}

void StartBigBlock(entity_t *e, int vtx)
{

	Entity_SetModel(e, "WorldData/Models/cube.mdl");
	Entity_SetCModel(e, "WorldData/Models/c_cube.mdl");

	e->skin[0].tex = Entity_LoadSkin("WorldData/Skins/ship.bmp");
	e->skin[1].tex = world->GetSkyTex();
	e->skin[1].type = ST_CUBEMAP_REFLECTION;
	e->skin[1].func = CompressBlend(GL_ONE, GL_ONE);
	e->skin[1].flags = SF_BLEND;
	e->skin_count = 2;

	e->check_collision = true;
	e->bounces = true;

	VectorSet(&e->scale, world->WorldSize * 0.25f, world->WorldSize * 0.25f, world->WorldSize * 0.25f);
	Entity_SetScale(e, &e->scale );

	strcpy(e->type, "big_block");

	vector_t origin;
	origin.x = world->WorldSize * 0.5f * ((vtx & 1) ? -1 : 1);
	origin.y = world->WorldSize * 0.5f * ((vtx & 2) ? -1 : 1);
	origin.z = world->WorldSize * 0.5f * ((vtx & 4) ? -1 : 1);
	Entity_SetPosition(e, &origin);

//	Entity_SetDensityInverse(e, 10.f);
}


/* SYSTEM FUNCTIONS */

void InitRace(void);

#include <time.h>
bool InitEntityScript(void)
{
	static int game_LargeRockCount = 10;
	static int game_EnemyCount = 0;

	static CVarw *fdjlkdsf[] = {
		new CVarw(&game_LargeRockCount,	"game_LargeRockCount"),
		new CVarw(&game_EnemyCount,		"game_EnemyCount"),
		NULL};

	int v;

	MSRand(time(NULL));

	particleSmokeTex = Entity_LoadSkin("WorldData/Particles/smoke.tga");
	particleFlameTex = Entity_LoadSkin("WorldData/Particles/flame.tga");

	Particle_SetTexture(particleSmokeTex);

	entity_t *e;

//create client ent
	client_ent = CreateEntity();
	if (!EntIsActive(client_ent)) return false;
	StartPlayer(client_ent);

//construct other objects

//	InitRace();

	for (v = 0; v < game_LargeRockCount; v++)
	{
		e = CreateEntity();
		if (!EntIsActive(e)) return false;
		StartLargeRock(e);
	}

/*	for (v = 0; v < 8; v++)
	{
		e = CreateEntity();
		if (!EntIsActive(e)) return false;
		StartBigBlock(e, v);
	}*/

	//construct enemy
	for (v = 0; v < game_EnemyCount; v++)
	{
		e = CreateEntity();
		if (!EntIsActive(e)) return false;
		StartEnemy(e);
	}

#if 0
#define DEBUG_SCALE		25
#define DEBUG_MODEL		"WorldData/Models/c_cube.mdl"
//COLLISION TEST OBJECTS

	quat_t r1, r2;
	AngleAxisToQuat(0,0,1, 45, &r1);
	AngleAxisToQuat(1,0,0, 45, &r2);
	QuatMul(&r2, &r1, &r1);
	
	entity_t *body;

	body = CreateEntity();
	VectorSet(&body->scale, DEBUG_SCALE,DEBUG_SCALE,DEBUG_SCALE);
	Entity_SetScale(body, &body->scale);
	VectorSet(&body->position, 100,-100,0);
	Entity_SetPosition(body, &body->position);
	Entity_SetAngle(body, &r1);
	Entity_SetModel(body, DEBUG_MODEL);
	Entity_SetCModel(body, DEBUG_MODEL);
	body->skin[0].tex = Entity_LoadSkin("WorldData/Skins/shot.bmp");
	body->check_collision = true;
	body->bounces = true;
	Entity_SetDensityInverse(body, 1);
	VectorSet(&body->velocity, 0,0.1f,0);

	body = CreateEntity();
	VectorSet(&body->scale, DEBUG_SCALE,DEBUG_SCALE,DEBUG_SCALE);
	Entity_SetScale(body, &body->scale);
	//determine constants - MassInv, BodyInertiaInv
	VectorSet(&body->position, 100,0,0);
	Entity_SetPosition(body, &body->position);
	Entity_SetModel(body, DEBUG_MODEL);
	Entity_SetCModel(body, DEBUG_MODEL);
	body->skin[0].tex = Entity_LoadSkin("WorldData/Skins/shot.bmp");
	//main_physics->AddRigidBody(body, .00000001f);
	body->check_collision = true;
	body->bounces = true;
	Entity_SetDensityInverse(body, 1);

	body = CreateEntity();
	VectorSet(&body->scale, DEBUG_SCALE,DEBUG_SCALE,DEBUG_SCALE);
	Entity_SetScale(body, &body->scale);
	//determine constants - MassInv, BodyInertiaInv
	VectorSet(&body->position, 100,100,0);
	Entity_SetAngle(body, &r1);
	Entity_SetModel(body, DEBUG_MODEL);
	Entity_SetCModel(body, DEBUG_MODEL);
	body->skin[0].tex = Entity_LoadSkin("WorldData/Skins/shot.bmp");
	body->check_collision = true;
	body->bounces = true;
	Entity_SetDensityInverse(body, 1);
#endif

	static bool init_d = false;
	if (!init_d)
	{
		init_d = true;
		new CVarw(&Client_UseMouse,	"Client_UseMouse"),
		new CVarw(&Client_Sensitivity,	"Client_Sensitivity"),
		new CVarw(&Client_Accel,		"Client_Accel"),
		new CVarw(&Client_MaxVel,		"Client_MaxVel"),
		new CVarw(&Client_Rot,			"Client_Rot"),
		new CVarw(&Client_Invincible,	"Client_Invincible");
		new CVarw(0,					"SpawnEnemy",			CFunc_SpawnEnemy);
		new CVarw(0,					"Respawn",				CFunc_Respawn);
	}

	return true;
}

/* CLIENT INPUT */

void UpdateClientInput(void)
{
	if (overrideCameraKeys) return;

	char *keystate = main_keyboard->keystate;

	if (keystate[VkKeyScan('n')] == KBH_PRESS)
		CFunc_SpawnEnemy(0,0);
	if (keystate[VkKeyScan('m')] == KBH_PRESS)
		CFunc_Respawn(0,0);

	//from here on out they're only client controls

	if (!EntIsActive(client_ent)) return;

	scalar_t movement = 0;
	if (keystate[VkKeyScan('w')])	movement += Client_Accel;	//fwd			
	if (keystate[VkKeyScan('s')])	movement -= Client_Accel;	//back			
	VectorMultAdd(&client_ent->velocity, &client_ent->matrix.vec[0], movement, &client_ent->velocity);

	//interpret rotation keys
	vector_t rot;
	VectorCopy(&vectorZero, &rot);
	if (keystate[VkKeyScan('a')]) rot.x -= Client_Rot;	//roll left
	if (keystate[VkKeyScan('d')]) rot.x += Client_Rot;	//roll right

	//get mouse input
	scalar_t ofx = (main_mouse->mouseSY - 0.5f) * 2.f;
	scalar_t ofy = (main_mouse->mouseSX - 0.5f) * 2.f;

////threshold
//	if (ofx > 0.05f || ofx < -0.05f || ofy > 0.05f || ofy < -0.05f)
	{
		rot.y -= ofx * Client_Sensitivity;
		rot.z -= ofy * Client_Sensitivity;
	}

	if (main_mouse->buttonPress[0]) Player_Fire_Rocket(client_ent);
	if (main_mouse->buttonPress[1]) Player_Fire_Gatling(client_ent);
	if (main_mouse->buttonPress[2]) Player_Fire_Bomb(client_ent);
	if (keystate[VkKeyScan(' ')]) Player_Fire_Hook(client_ent);
	if (keystate[VkKeyScan('f')]) Player_Fire_Temp(client_ent);
	if (keystate[VkKeyScan('y')]) Player_FireDrunkMissiles(client_ent);

#if 0		//oldschool - pure setangle - works a little smoother
	quat_t rotation;	//construct to identity - this is the rotation made due to the player input
	QuatCopy(&quatIdentity, &rotation);

	if (rot.x || rot.y || rot.z)
	{
		quat_t reg;

		if (rot.x)
		{
			AngleAxisToQuat(1,0,0,rot.x, &reg);
			QuatMul(&rotation, &reg, &rotation);
		}
		if (rot.y)
		{
			AngleAxisToQuat(0,1,0,rot.y, &reg);
			QuatMul(&rotation, &reg, &rotation);
		}
		if (rot.z)
		{
			AngleAxisToQuat(0,0,1,rot.z, &reg);
			QuatMul(&rotation, &reg, &rotation);
		}
#if 0
		//Entity_SetAngle(client_ent, client_ent->angle * rotation);
		//how to orient a rotation to the space of another quaternion
		QuatConj(&client_ent->angle, &reg);
		QuatMul(&client_ent->angle, &rotation, &rotation);
		QuatMul(&rotation, &reg, &rotation);
		//rotation = client_ent->angle * rotation * Conjugate(client_ent->angle);
		//now rotation can be interpolated from the identity axis for the sake of rotating the velocity via turn sharpness
		QuatMul(&rotation, &client_ent->angle, &client_ent->angle);
		Entity_SetAngle(client_ent, &client_ent->angle );
		//Entity_SetAngle(client_ent, rotation * client_ent->angle);
#else
		//orientate our rotation in the entity matrix
		//R' = M * R * M^t			<- R' = R relative to M's basis
		//calculate the new entity matrix based upon its rotation
		//M' = R' * M				<- M' = M concatenated with the R' rotation
		//M' = (M * R * M^t) * M	//substitute...
		//M' = M * R * (M^t * M)	//simplify...
		//M' = M * R * I
		//M' = M * R				//and in the end we're saved a few quaternion multiplications
		//but doesnt this transpose the rotation when cat'd to M?
		//the reason we're not using this more efficient method
		//is because we want 'rotation' to be relative to 'client_ent->angle' for later calculations
//		QuatConj(&rotation, &rotation);
		QuatMul(&client_ent->angle, &rotation, &client_ent->angle);
		Entity_SetAngle(client_ent, &client_ent->angle );
#endif
	}

	//adjusting velocity =)
	static scalar_t Player_TurnSharpness = 1;
	static CVarw *dsfjklafsd = new CVarw(&Player_TurnSharpness, "Player_TurnSharpness");
	if (Player_TurnSharpness && rotation.w != 1)		//if we have a turn sharpness and a rotation
	{
		//rotation = SlerpQuat(quatIdentity, rotation, Player_TurnSharpness);
		//client_ent->velocity = RotateVector(client_ent->velocity, rotation);
		QuatSlerp(&quatIdentity, &rotation, Player_TurnSharpness, &rotation);
		QuatRotateVector(&client_ent->velocity, &rotation, &client_ent->velocity);
	}

#else	//newschool - add rotation - integrate during physics frame (works better with collision)
	if (rot.x || rot.y || rot.z)
	{
		//align rotation to current modelspace
		QuatRotateVector(&rot, &client_ent->angle, &rot);
		//scale by degrees to radians and add to current rotation (very sucky) - and decay a lil because of continuous summation
		VectorMultAdd(&client_ent->rotation, &rot, (scalar_t)DEGREE_TO_RADIAN * 0.5f, &client_ent->rotation);
	}
#endif

	//change view key
	static int player_view = -1;
	if (keystate[VkKeyScan('x')] == KBH_PRESS || player_view == -1)
	{
		player_view = (player_view + 1) % 4;
		switch(player_view)
		{
		case 0:
			VectorSet(&main_camera->chaseOffsetPos, -150, 0, 25);
			break;
		case 1:
			VectorSet(&main_camera->chaseOffsetPos, -500, 0, 0);
			break;
		case 2:
			VectorSet(&main_camera->chaseOffsetPos, -10, 0, 20);
			break;
		case 3:
			VectorSet(&main_camera->chaseOffsetPos, -50, 0, 20);
			break;
		}
	}

	//update the chasecam
	main_camera->Update();

	//draw a rocket tail
	if (movement > 0) Entity_RocketEffect(client_ent);
}




void Marker_Think(entity_t *marker)
{
	if (!EntIsActive(marker->owner)) return;

//	if ((int)marker->timer & 16)
	{
		vector_t position;
		VectorLerp(&marker->position, &marker->owner->position, (scalar_t)marker->timer / 256.f, &position);

		Particle_SetTexture(particleFlameTex);
		
		Particle_Create(&position, &vectorZero, 150.f, 0x7F00FF00, 6, true);
		Particle_Create(&position, &vectorZero, 25.f, 0xFFFFFFFF, 6, true);

		Particle_Create(&marker->position, &vectorZero, 150.f, 0x7F00FFFF, 6, true);
		Particle_Create(&marker->position, &vectorZero, 25.f, 0xFFFFFFFF, 6, true);
	}

	marker->timer++;

	if (marker->timer >= 256.f)
		marker->timer -= 256.f;
}

void StartMarker(entity_t *e, entity_t *next)
{
	quat_t reg;
	
	strcpy(e->type, "marker");

	VectorSet(&reg.vec,
		MCRand() * world->WorldSize,
		MCRand() * world->WorldSize,
		MCRand() * world->WorldSize);
	Entity_SetPosition(e, &reg.vec );

	e->Think = Marker_Think;

	//flag for next marker
	e->owner = next;

	//create the marker
	Entity_SetModel(e, "WorldData/models/shot.mdl");
	Entity_SetCModel(e, "WorldData/models/c_shot.mdl");
	e->skin[0].tex = Entity_LoadSkin("WorldData/skins/shot.bmp");
	e->skin[1].tex = world->GetSkyTex();
	e->skin[1].type = ST_CUBEMAP_REFLECTION;
	e->skin[1].func = CompressBlend(GL_ONE, GL_ONE);
	e->skin[1].flags = SF_BLEND;
	e->skin_count = 2;
}

void InitRace(void)
{
	entity_t *first, *e, *next;

	e = CreateEntity();
	if (!EntIsActive(e)) return;

	first = e;

	for (int counter = 0; counter < 5; counter++)
	{
		next = CreateEntity();
		if (!EntIsActive(next)) return;

		StartMarker(e, next);

		e = next;
	}

	StartMarker(e, first);
}

