/*

    File: EntityTree.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <math.h>
#include <windows.h>	//malloc, NULL, etc, etc

#include "../m_system3d/error.h"

#include "../m_math3d/geometry.h"
#include "../m_foundation/const.h"
#include "../m_foundation/main.h"

#include "entitytree.h"

#include "main.h"
#include "entity.h"
#include "world.h"
#include "collision.h"

static scalar_t defaultIntegrationTimestep = 1.f;

void Entity_Integrate(entity_t *e, scalar_t time)
{
	VectorMultAdd(&e->position, &e->velocity, time, &e->position);
	QuatIntegrate(&e->angle, &e->rotation, time, &e->angle);
}

// ****************************
//			ENTITY TREE 
// ****************************

typedef struct collision_test_s {
	entity_t	*ent;
	entity_t	*collided;
	box_t		move_bbox;
	scalar_t	move_inttime;
	bool		model_loaded;
} collision_test_t;

#define ENT_TREE_DEPTH	4
#define ENT_TREE_COUNT	31		//((1<<(ENT_TREE_DEPTH+1))-1)

class EntTreeNode {
public:
	int			plane_axis;			//dividing plane axis (pointing towards
	scalar_t	plane_dist;			//distance along that axis
	bool		leaf;				//whether this is a leaf or not

	vector_t	center;				//used for ent_display_bsptree

	LinkedList	list;
	EntTreeNode	*child[2];

	int			ChildContains(const box_t *box);
	void		AddEnt(entity_t *e);
	void		TestCollision(collision_test_t *test, collision_info_t *info);
};

static EntTreeNode *entTree = NULL;

void Init_EntTreeNode(int i, box_t bbox)
{
	EntTreeNode *node = &entTree[i];

	//clear list ptrs
	node->list.head = NULL;
	node->list.tail = NULL;

	BoxCenter(&bbox, &node->center);

	int child = (((i + 1) << 1) - 1);
	if (child >= ENT_TREE_COUNT)
	{
		node->leaf = true;

		node->plane_axis = 0;
		node->plane_dist = 0;

		node->child[0] = node->child[1] = NULL;
	}
	else
	{
		node->leaf = false;

		//calculate plane division - only need this if we do have children
		//set up node division
		vector_t size;
		VectorSub(&bbox.max, &bbox.min, &size);

		node->plane_axis = 2;

		if (size.x > size.y && size.x > size.z)
			node->plane_axis = 0;
		else if (size.y > size.z)
			node->plane_axis = 1;

		node->plane_dist = bbox.min.v[node->plane_axis] + size.v[node->plane_axis] * 0.5f;

		node->child[0] = &entTree[child];
		node->child[1] = &entTree[child+1];

//			node->bbox.max.v[node->plane_axis] = node->plane_dist;
		scalar_t oldmax = bbox.max.v[node->plane_axis];
		bbox.max.v[node->plane_axis] = node->plane_dist;
		Init_EntTreeNode(child, bbox);
		bbox.max.v[node->plane_axis] = oldmax;
		bbox.min.v[node->plane_axis] = node->plane_dist;
//			node->bbox.min.v[node->plane_axis] = node->plane_dist;
		Init_EntTreeNode(child+1, bbox);
	}
}

bool Init_EntityTree(void)
{
	entTree = (EntTreeNode *)malloc(sizeof(EntTreeNode) * ENT_TREE_COUNT);
	if (!entTree) return false;

	box_t bbox;

	VectorSet(&bbox.min, -world->WorldSize, -world->WorldSize, -world->WorldSize);
	VectorSet(&bbox.max, world->WorldSize, world->WorldSize, world->WorldSize);;

	Init_EntTreeNode(0, bbox);
	
	return true;
}

void Shutdown_EntityTree(void)
{
	if (entTree) free(entTree);
	entTree = NULL;
}

void EntityTree_Link(entity_t *e)
{
	entTree[0].AddEnt(e);
}

int EntTreeNode::ChildContains(const box_t *box)
{
	if (box->min.v[plane_axis] > plane_dist) return 2;
	if (box->max.v[plane_axis] > plane_dist) return 1;
	return 0;
}

void EntTreeNode::AddEnt(entity_t *e)
{
	//ie root node
	if (this == entTree)
	{
		//if the entire world doesnt contain the entity bbox
		//ie the entity bbox lies on a world dividing line
		//then link to the root node

		if (!(	e->bbox.min.x >= -world->WorldSize &&
				e->bbox.max.x <= world->WorldSize &&
				e->bbox.min.y >= -world->WorldSize &&
				e->bbox.max.y <= world->WorldSize &&
				e->bbox.min.z >= -world->WorldSize &&
				e->bbox.max.z <= world->WorldSize))
		{
			list.AddLast(&e->treeLink);
			return;
		}
	}

	//dont bother test to the dividing plane
	if (leaf)
	{
		list.AddLast(&e->treeLink);
		return;
	}

	int flags = ChildContains(&e->bbox);

	if (flags == 0)	//both min and max are below the plane line
		child[0]->AddEnt(e);
	else if (flags == 2) //both min and max are above the plane line
		child[1]->AddEnt(e);
	else		//crossing that line - attach to this node
		list.AddLast(&e->treeLink);
}

// ************************************************
//					UPDATE PHYSICS
// ************************************************

/*
//the ent to be used
entity_t		*coltest_ent;
//the last recorded collided ent
entity_t		*coltest_collided;
//the ent's movement bbox
box_t			coltest_move_bbox;
//the ent's integration time
scalar_t		coltest_move_inttime;
//whether the ent's collision model has been loaded
bool			coltest_model_loaded;
*/

/*
EntityTree_TestCollision - Tests entity 'coltest_ent' with entity 'e'

  reads:
	e =						the entity to be tested against 'coltest_ent'
	e_position =			'e's position, closest to 'coltest_ent'
	e_delta =				world delta vector from 'coltest_ent' to 'e''s position
	test =					current collision test data
  writes:
	test =					new collision test data
	info =					collision information - to be sent to the collision executor

  reads:
	test->ent =				gets lots of info from it
	test->move_inttime =	reads the current upper bound on integration time
	test->model_loaded =	reads to see if the model needs to be loaded
  writes:
	test->collided =		sets this to the tested entity
	test->move_inttime =	sets the new upper bound on integration time
	test->model_loaded =	sets this if we loaded the model
*/
void EntityTree_TestCollision(
	entity_t *e,
	const vector_t *e_position,
	const vector_t *e_delta,
	collision_test_t *test,
	collision_info_t *info)
{
	//dont bother test self to self
	if (e == test->ent) return;

	//dirty hack - before testing movement, see if the dist between 'e' and 'test->ent'
	//is below the sum of their mincollision radii.  if so then the models are stuck
	{
		scalar_t mindist;
		//calc min dist
		mindist = Entity_CollisionMinRadius(e) + Entity_CollisionMinRadius(test->ent);
		if (VectorLengthSq(e_delta) < mindist * mindist - 1)	//add in epsilon
		{
				//stuck models - collision inevitable - calc goods and leave
			test->collided = e;
			test->move_inttime = 0;

			info->obj[0] = test->ent;
			info->obj[1] = e;

			VectorCopy(&test->ent->position, &info->pos[0]);
			VectorCopy(e_position, &info->pos[1]);
			//VectorSub(&test->ent->position, e_position, &info->normal);
			VectorScale(e_delta, -1, &info->normal);
			//VectorCopy(e_delta, &info->normal);
			
			VectorMultAdd(
				e_position,
				&info->normal, 
				Entity_CollisionMaxRadius(e) / (Entity_CollisionMaxRadius(test->ent) + Entity_CollisionMaxRadius(e)),
				&info->point);
	//		info->point = info->normal * (
	//			e->CollisionMaxRadius() / (test->ent->CollisionMaxRadius() + e->CollisionMaxRadius())
	//			) + e_position;

			VectorUnit(&info->normal, &info->normal);
	//		info->normal.Normalize();
		}
	}

	//test 'test->ent' to 'o'

	scalar_t collision_time_range[2];

	int spheretrace_result;
	
	spheretrace_result = Collision_SphereTrace(
		test->move_inttime,								//timestep
		collision_time_range,								//
		&test->ent->position,								//position at time zero of sphere 1
		&test->ent->velocity,								//velocity of sphere 1
		Entity_CollisionMaxRadius(test->ent),	//radius of sphere 1
		e_position,										//position of sphere 2
		Entity_CollisionMaxRadius(e) + 0.2f);				//radius of sphere 2 (with 0.2 epsilon)

	if (spheretrace_result == CST_FREE)
		return;

//						(optional) use recursive time subdivision to find the
//							exact collision on the rigid bodies
//						integration_time = the first intersection time of the collision
//						test->collided = 'b'

	//And the true to life Convex Hull test
	static int e_TestHull = 1;
	static CVarw *fdldfslkj = new CVarw(&e_TestHull, "TestHull");
	if (e_TestHull)
	{
		//REPLACE THIS WITH RECURSIVE TIME SUBDIVISION COLLISION DETECTION
		//between collision_time_range[0] and collision_time_range[1]

		vector_t old_position;
		quat_t old_angle;
		vector_t res_point, res_normal;
		
		VectorCopy(&test->ent->position, &old_position);
		QuatCopy(&test->ent->angle, &old_angle);

		Entity_Integrate(test->ent, collision_time_range[0]);

		if (!test->model_loaded)
		{
			LoadCollisionModel(
				test->ent->c_model,
				&test->ent->position,
				&test->ent->angle,
				&test->ent->scale,
				C_MODEL_OBJ0);
			test->model_loaded = true;
		}

		LoadCollisionModel(
			e->c_model,
			e_position,
			&e->angle,
			&e->scale,
			C_MODEL_OBJ1);

		VectorCopy(&old_position, &test->ent->position);
		QuatCopy(&old_angle, &test->ent->angle);

		//first - test if the witness of the test->ent separates the two collision models
		if (!WitnessSeparates(&test->ent->witness) &&
		//if the witness fails - test for a collision between the two collision models
			ObjectCollision(&res_point, &res_normal, &test->ent->witness))
		//if collision occurs then please fill out this form...
		{
#ifndef OLDSCHOOL
			test->collided = e;
			test->move_inttime = collision_time_range[0];
#else
		if (test->move_inttime != defaultIntegrationTimestep)
		{
			if (test->move_inttime)
				main_console->Print("integration time overriden: %f\n", test->move_inttime);
		}
#endif
			info->obj[0] = test->ent;
			info->obj[1] = e;

			VectorCopy(&test->ent->position, &info->pos[0]);
			VectorCopy(e_position, &info->pos[1]);

//instead we have
#ifndef OLDSCHOOL
			VectorCopy(&res_point, &info->point);
			VectorCopy(&res_normal, &info->normal);
#else
//comment these out if we are allowing info->position / normal calculation within ObjectCollide
			VectorSub(&test->ent->position, e_position, &info->normal);
			VectorMultAdd(e_position, &info->normal, Entity_CollisionMaxRadius(e) / (Entity_CollisionMaxRadius(test->ent) + Entity_CollisionMaxRadius(e)), &info->point);
			VectorUnit(&info->normal, &info->normal);

			//info->point = info->normal * (
			//	e->CollisionMaxRadius() / (test->ent->CollisionMaxRadius() + e->CollisionMaxRadius())
			//	) + e_position;
			//info->normal.Normalize();
#endif
		}
	}
	else
	//if no hull testing, then just assume collision and get on with it
	{
		test->collided = e;
		test->move_inttime = collision_time_range[0];


		info->obj[0] = test->ent;
		info->obj[1] = e;

		VectorCopy(&test->ent->position, &info->pos[0]);
		VectorCopy(e_position, &info->pos[1]);
		//VectorSub(&test->ent->position, e_position, &info->normal);
		VectorScale(e_delta, -1, &info->normal);
		//VectorCopy(e_delta, &info->normal);
		
		VectorMultAdd(
			e_position,
			&info->normal, 
			Entity_CollisionMaxRadius(e) / (Entity_CollisionMaxRadius(test->ent) + Entity_CollisionMaxRadius(e)),
			&info->point);
//		info->point = info->normal * (
//			e->CollisionMaxRadius() / (test->ent->CollisionMaxRadius() + e->CollisionMaxRadius())
//			) + e_position;

		VectorUnit(&info->normal, &info->normal);
//		info->normal.Normalize();
	}
}

//#define	STRUCT_FROM_LINK(l,t,m) ((t *)((byte *)l - (int)&(((t *)0)->m)))
//#define	ENTITY_FROM_TREE(l) STRUCT_FROM_LINK(l,entity_t,treeLink)
#define	ENTITY_FROM_TREE(e) ((entity_t *)((unsigned char *)e - (int)&(((entity_t *)NULL)->treeLink)))

/* 
  reads:
	coltest_move_inttime	to see if a collision has already been found
	coltest_ent				reads position to calc delta vector to tested ent
	coltest_move_bbox		to see what children we have to test
  writes:
    all the stuff in TestCollision
*/
void EntTreeNode::TestCollision(collision_test_t *test, collision_info_t *info)
{
	//assert(bbox.Intersects(&test_main_ent_bbox));

	//we have already found a collision at 0 seconds - we dont need to test anymore
	if (test->move_inttime == 0) return;

	if (list.head)
	{
		Chain *link, *next;

		//test all entities within this node to 'e'
		for (link = list.head; link; link = next)
		{
			next = link->chain_next;
			entity_t *e = ENTITY_FROM_TREE(link);

			//continue if the ent is inactive (ie it died)
			if (!EntIsActive(e)) continue;
			//continue if the ent doesnt collide with other entities
			if (!e->check_collision || !e->c_model) continue;

			vector_t delta, pos;
			world->Delta(&e->position, &test->ent->position, &delta);
			VectorAdd(&test->ent->position, &delta, &pos);

			EntityTree_TestCollision(e, &pos, &delta, test, info);

			//we have already found a collision at 0 seconds - we dont need to test anymore
			if (test->move_inttime == 0) return;
		}
	}

	//test children

	if (!leaf)
	{
// plane separation
		int flags = ChildContains(&test->move_bbox);

		if (flags == 0 || flags == 1)
		{
			child[0]->TestCollision(test, info);
		}
		else if (flags == 2 || flags == 1)
		{
			child[1]->TestCollision(test, info);
		}
	}
}

/* uses
coltest_ent
coltest_collided

  reads:
	coltest_ent
	coltest_collided
  writes:
    everything in coltest_ent & coltest_collided's Touch and Bounce functions
	everything in ApplyImpulseForces
*/
void EntityTree_ExecuteCollision(const collision_test_t *test, collision_info_t *info)
{
	collision_result_t result;

//main_console->Print("collision init %p : %p\n", test->ent, test->collided);

	// ******************** PRE-TOUCH ******************** 
	
	//if there is no dividing plane for either of the loaded collision models
	//ie the objects *have* touched
	test->ent->touched = true;
	test->collided->touched = true;

	if (test->ent->Touch) test->ent->Touch(test->ent, test->collided, info);

	//test->ent died - stop checking for more collisions
	if (!EntIsActive(test->ent)) return;
	//test->collided died - dont do the other check
	if (!EntIsActive(test->collided)) return;

	if (test->collided->Touch) test->collided->Touch(test->collided, test->ent, info);

	// ******************** COLLIDE ******************** 

	if (EntIsActive(info->obj[0]) &&
		EntIsActive(info->obj[1]) &&
		info->obj[0]->check_collision && 
		info->obj[1]->check_collision && 
		info->obj[0]->bounces &&
		info->obj[1]->bounces)
	{
//		main_console->Print("collision forces applied: %f %f %f\n", info->point.x, info->point.y, info->point.z);
		ApplyImpulseForces(info, &result);
	}

	// ******************** POST-TOUCH ******************** 

	if (test->ent->Bounce) test->ent->Bounce(test->ent, test->collided, info, &result);

	//test->ent died - stop checking for more collisions
	if (!EntIsActive(test->ent)) return;
	//test->collided died - dont do the other check
	if (!EntIsActive(test->collided)) return;

	if (test->collided->Bounce) test->collided->Bounce(test->collided, test->ent, info, &result);
}

#include "../m_system3d/main.h"

/* uses
coltest_ent
coltest_collided
coltest_move_inttime
coltest_model_loaded

  reads:
  writes:
	coltest_ent
	coltest_move_inttime
	coltest_collided
	coltest_model_load
	(coltest_move_bbox is not mod'd because it is not used in the linear physics update)
*/
void EntityTree_UpdatePhysics_Linear(void)
{
	collision_info_t info;
	collision_test_t test;

#ifdef DEBUG_OUTPUT
	double starttime = main_timer->GetTime();
#endif

	Chain *link, *next;
	for (link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		test.ent = ENTITY_FROM_LINK(link);

		if (!EntIsActive(test.ent)) continue;

		//if the current ent does not collide then integrate it and move on
		if (!test.ent->check_collision || !test.ent->c_model)
		{
			Entity_Integrate(test.ent, defaultIntegrationTimestep);
			continue;
		}

		//else continue and test for collision

	//init collision stat vars

		test.move_inttime = defaultIntegrationTimestep;
		test.collided = NULL;
		test.model_loaded = false;

	//for (entity_t *e = (entity_t *)entLiveChain.head; e; e = (entity_t *)e->chain_next)
		for (Chain *link = entLiveChain.head; link; link = link->chain_next)
		{
			entity_t *e = ENTITY_FROM_LINK(link);
			if (!EntIsActive(e)) continue;
			if (!e->check_collision || !e->c_model) continue;

			vector_t delta, pos;
			world->Delta(&e->position, &test.ent->position, &delta);
			VectorAdd(&test.ent->position, &delta, &pos);
			EntityTree_TestCollision(e, &pos, &delta, &test, &info);

			//we have already found a collision at 0 seconds - we dont need to test anymore
			if (test.move_inttime == 0) break;
		}

//		if (test.move_inttime != defaultIntegrationTimestep)
//		{
//			if (test.move_inttime)
//				main_console->Print("integration time overriden: %f\n", test.move_inttime);
//		}
//override to temporarily prevent stuck objects
test.move_inttime = defaultIntegrationTimestep;

	//make use of collision stat vars

		//and integrate it
		Entity_Integrate(test.ent, test.move_inttime);

		//if we found a collider
		if (EntIsActive(test.collided))
			EntityTree_ExecuteCollision(&test, &info);

		if (EntIsActive(test.ent))
		{
			//and relink and modulo position
			Ent_ModuloPosition(test.ent);
			Ent_UpdateAngle(test.ent);
			Ent_Relink(test.ent);
		}
	}

#ifdef DEBUG_OUTPUT
	main_console->Print("physics time: %f\n", main_timer->GetTime() - starttime);
#endif
}

/* uses

  reads:
  writes:
	coltest_ent
	coltest_move_inttime
	coltest_collided
	coltest_model_load
	coltest_move_bbox
*/
void EntityTree_UpdatePhysics_BSP(void)
{
	collision_test_t test;
	collision_info_t info;

#ifdef DEBUG_OUTPUT
	double starttime = main_timer->GetTime();
#endif

	Chain *next;
	for (Chain *link = entLiveChain.head; link; link = next)
	{
		next = link->chain_next;
		test.ent = ENTITY_FROM_LINK(link);

		if (!EntIsActive(test.ent)) continue;

		//if the current ent does not collide then integrate it and move on
		if (!test.ent->check_collision || !test.ent->c_model)
		{
			Entity_Integrate(test.ent, defaultIntegrationTimestep);
			continue;
		}

		//else continue and test for collision

		test.ent->treeLink.Unlink();

	//init collision stat vars

		test.move_inttime = defaultIntegrationTimestep;
		test.collided = NULL;
		test.model_loaded = false;

		//calc bbox around current and dest pos - based upon pos, c_model->radius, and scale
		vector_t destpos;
		scalar_t rad;

		rad = Entity_CollisionMaxRadius(test.ent);

		VectorMultAdd(&test.ent->position, &test.ent->velocity, defaultIntegrationTimestep, &destpos);
		//destpos = test.ent->position + test.ent->velocity * defaultIntegrationTimestep;

		for (int v = 0; v < 3; v++)
		{
			test.move_bbox.min.v[v] = (test.ent->position.v[v] < destpos.v[v]) ? test.ent->position.v[v] - rad : destpos.v[v] - rad;
			test.move_bbox.max.v[v] = (test.ent->position.v[v] > destpos.v[v]) ? test.ent->position.v[v] + rad : destpos.v[v] + rad;
		}

		//test tree to 'test.ent' with the tree
		entTree->TestCollision(&test, &info);

//override to temporarily prevent stuck objects
test.move_inttime = defaultIntegrationTimestep;

	//make use of collision stat vars

		//and integrate it
		Entity_Integrate(test.ent, test.move_inttime);

		//if we found a collider
		if (EntIsActive(test.collided))
			EntityTree_ExecuteCollision(&test, &info);

		if (EntIsActive(test.ent))
		{
			//and relink and modulo position
			Ent_ModuloPosition(test.ent);
			Ent_UpdateAngle(test.ent);
			Ent_Relink(test.ent);
		}
	}

#ifdef DEBUG_OUTPUT
	main_console->Print("physics time: %f\n", main_timer->GetTime() - starttime);
#endif
}


//completely debug
void EntityTree_DebugDisplay(void)
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_TEXTURE);
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	for (int v = 0; v < ENT_TREE_COUNT; v++)
	{
		EntTreeNode *node = &entTree[v];

		vector_t thisPos, childPos;

		VectorCopy(&node->center, &thisPos);

		//draw children
		for (int c = 0; c < 2; c++)
		{
			if (node->child[c])
			{
				VectorCopy(&node->child[c]->center, &childPos);
				glColor3f(1,1,0);
				glVertex3fv(thisPos.v);
				glVertex3fv(childPos.v);
			}
		}

		//draw node entities
		if (node->list.head)
		{
			//test all entities within this node to 'e'
			Chain *link;
			for (link = node->list.head; link; link = link->chain_next)
			{
				entity_t *e = ENTITY_FROM_TREE(link);
				if (!EntIsActive(e)) continue;

				glColor3f(0,1,0);
				glVertex3fv(thisPos.v);
				glVertex3fv(e->position.v);
			}
		}
	}
	glEnd();
	glPopAttrib();
}

void EntityTree_UpdatePhysics(scalar_t deltaTime)
{
//debug - display bsp tree
	static int EntDisplayBSPTree = 0;
	static int UpdatePhysics_UseBSPTree = 1;
	static CVarw *static_cvars[] = {
		new CVarw(&EntDisplayBSPTree,			"Ent_Display_BSPTree"),
		new CVarw(&UpdatePhysics_UseBSPTree,	"UseBSPTree"),
		NULL};

	//store it in our EntityTree.cpp-local variable for access to all in this file.
	defaultIntegrationTimestep = deltaTime;

	if (EntDisplayBSPTree)
		EntityTree_DebugDisplay();

//before update - cap max speed - make sure no one is speeding
//that can cause nasty stalls/lock-ups in our modulo space calculations
	{
		static scalar_t worldSpeedLimit = 100;
		static scalar_t worldRotLimit = 0.5f;
		static CVarw *fdsjlds[] = {
			new CVarw(&worldSpeedLimit,	"worldSpeedLimit"),
			new CVarw(&worldRotLimit,		"worldRotLimit"),
			NULL};

		scalar_t v_sq, limit_lin_sq, limit_ang_sq;

		limit_lin_sq = worldSpeedLimit * worldSpeedLimit;
		limit_ang_sq = worldRotLimit * worldRotLimit;

#define EPSILON		0.01

		for (Chain *link = entLiveChain.head; link; link = link->chain_next)
		{
			entity_t *e = ENTITY_FROM_LINK(link);

			v_sq = VectorLengthSq(&e->velocity);
			if (v_sq > limit_lin_sq + EPSILON)
				VectorScale(&e->velocity, worldSpeedLimit / (scalar_t)sqrt(v_sq), &e->velocity);

			v_sq = VectorLengthSq(&e->rotation);
			if (v_sq > limit_ang_sq + EPSILON)
			{
				v_sq = (scalar_t)sqrt(v_sq);
				main_console->Print("exceeding rot limit: %f\n", v_sq);

				VectorScale(&e->rotation, worldRotLimit / v_sq, &e->rotation);
			}
		}
	}

//update the physics state without making use of the BSP tree.  a safe bet, though should be slower
	if (!UpdatePhysics_UseBSPTree)
		EntityTree_UpdatePhysics_Linear();
//update the physics state based upon the BSP tree.  should be O(n*log(n)) rather than O(n^2)
	else
		EntityTree_UpdatePhysics_BSP();
}
