/*

    File: LODModel.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <stdio.h>
#include <math.h>

#include "../m_foundation/const.h"
#include "../m_foundation/main.h"

#include "LODModel.h"

#include "menu.h"

bool LODModel::Load(char *filename)
{
	//main_console->Print("LODModel::Load(%s)\n", filename);

	Unload();

	//set model 0
	model[0] = new Model();
	if (!model[0])
	{
		main_console->Print("LODModel::Load() - failed - unable to allocate base object\n");
		return false;
	}
	if (!model[0]->Load(filename))
	{
		main_console->Print("LODModel::Load() - failed - unable to load base model\n");
		delete model[0];
		return false;
	}
	//set depth 0
	invWidth[0] = 0;		//this is never really used anyways
	//set count to 1
	count = 1;

	int index = 1;
	
	char lodname[256];
	
	Model *m;
	
	char *ext_ptr = strrchr(filename, '.');
	
	int ext_off =  ext_ptr ? (int)(ext_ptr - filename) : strlen(filename);
	
	memcpy(lodname, filename, ext_off);
	lodname[ext_off] = 0;
	strcat(lodname, "_l00");
	strcat(lodname, &filename[ext_off]);

	int index_off = ext_off + 2;

	do {
		//find a new file

		lodname[index_off] = (index / 10) + '0';
		lodname[index_off+1] = (index % 10) + '0';

		//verify existance
		FILE *file = fopen(lodname, "r");
		if (!file) break;
		fclose(file);

		m = new Model();
		if (!m) break;

		//main_console->Print("LODModel::Load() - loading LOD model file %s\n", lodname);

		//load it
		if (!m->Load(lodname))
		{
			delete m;
			break;
		}

		//and be on our merry way
		LoadLOD(m);

		index++;

	} while (1);

	//main_console->Print("LODModel::Load() success - loaded %d LODs\n", count);

	return true;
}

void LODModel::Unload(void)
{
	for (int i = 0; i < count; i++)
		delete model[i];
	
	Reset();
}

void LODModel::Reset(void)
{
	count = 0;
}

Model *LODModel::GetModelByWidth(scalar_t w)
{
//	Model *m = model[0];
//
//	for (int i = 0; i < count; i++)
//		if (w < width[i])
//			m = model[i];
//
//	return m;

//	for (int i = 1; i < count && w < 1.f/invWidth[i]; i++);

	scalar_t invW = 1.f / w;
	for (int i = 1; i < count && invW > invWidth[i]; i++);
	return model[i-1];
}

Model *LODModel::GetModelByLevel(int level)
{
	return model[level >= 0 ? (level < count ? level : 0) : 0];
}

void LODModel::Display(scalar_t w)
{
	GetModelByWidth(w)->Display();
}

void LODModel::Display(int level)
{
	GetModelByLevel(level)->Display();
}

//262144 bytes
//65536 floats * 4 bytes/float
//8192 vertexes * 8 floats/vertex * 4 bytes/float
static float tempvertexbuffer[8192*8];

void LODModel::DisplayInterpolate(scalar_t w)
{
	int i;

	if (count == 1)
	{
		model[0]->Display();
		return;
	}

	scalar_t invW = 1.f / w;
	
	for (i = 1; i < count && invW > invWidth[i]; i++);
	
	Model *m = model[i-1];
	//width[0] = fInf	->	0.0
	//		 w = fInf	->	0.0		-> 0%
	//		 w = 0.75	->	1.333	-> 67%
	//		 w = 0.5	->	2.0		-> 100%
	//width[1] = 0.5	->	2.0

	if (!menu_lod_interpolate)
	{
		m->Display();
		return;
	}


	if (m->vtxCount > 8192)
	{
		main_console->Print("we got a model with too many verticies: %d\n", m->vtxCount);
		return;
	}

	scalar_t prevwidth = invWidth[i-1];
	scalar_t nextwidth = invWidth[i];
	scalar_t frac = (invW - prevwidth) / (nextwidth - prevwidth);
	scalar_t ifrac = 1.f - frac;

	//first - interpolate the verticies (save on computing time)
	float *tvb = tempvertexbuffer;
	int di;
	for (i = 0; i < m->vtxCount; i++)
	{
		di = m->collapse[i];
		tvb[0] = m->vtx[i].x * ifrac + m->vtx[di].x * frac;
		tvb[1] = m->vtx[i].y * ifrac + m->vtx[di].y * frac;
		tvb[2] = m->vtx[i].z * ifrac + m->vtx[di].z * frac;
		tvb[3] = m->normal[i].x * ifrac + m->normal[di].x * frac;
		tvb[4] = m->normal[i].y * ifrac + m->normal[di].y * frac;
		tvb[5] = m->normal[i].z * ifrac + m->normal[di].z * frac;
		tvb[6] = m->texCoord[i<<1] * ifrac + m->texCoord[di<<1] * frac;
		tvb[7] = m->texCoord[(i<<1)+1] * ifrac + m->texCoord[(di<<1)+1] * frac;
		tvb += 8;
	}

	di = 3*m->triCount;
	glBegin(GL_TRIANGLES);
	for (i = 0; i < di; i++)
	{
		tvb = tempvertexbuffer + (m->tri[i] << 3);
		glTexCoord2fv(&tvb[6]);
		if (found_GL_ARB_multitexture) glMultiTexCoord2fvARB(GL_TEXTURE1_ARB, &tvb[6]);
		glNormal3fv(&tvb[3]);
		glVertex3fv(tvb);
	}
	glEnd();
}

bool LODModel::LoadLOD(Model *m)
{
	if (count == LODMODEL_MAX_DEPTH)
	{
		main_console->Print("LODModel::LoadLOD(): LOD pointer array full\n");
		return false;
	}

	if (count == 0)
	{
		main_console->Print("LODModel::LoadLOD(): must load base model before loading LOD models\n");
		return false;
	}

	scalar_t w = (scalar_t)m->triCount / (scalar_t)model[0]->triCount;

	if (w < 0 || w > 1)
	{
		main_console->Print("LODModel::LoadLOD(): LOD triCount exceeds base model triCount\n");
		return false;
	}

//	w = 0.5f * (scalar_t)sqrt(w);		//conversion from tri ratio to screen size ratio
	w = 0.5f * w;	//looks better =)

	scalar_t invW = 1.f / w;

	for (int i = 0; i < count; i++)
	{
		if (invWidth[i] > invW)
		{
			//shift everything up one unit
			for (int j = count-1; j >= i; j--)
			{
				model[j+1] = model[j];
				invWidth[j+1] = invWidth[j];
			}

			//and break out to loading the new model in its place
			break;
		}
	}

	//main_console->Print("LODModel::LoadLOD(): successfully loaded as LOD %d\n", i);

	//store, inc counter, and return 'good to go'
	model[i] = m;
	invWidth[i] = invW;
	
	count++;

	return true;
}


void LODModel::Compile(void)
{
	for (int i = 0; i < count; i++)
		model[i]->Compile();
}

LODModel *GetLODModel(LinkedList *p, char *title)
{
	for (LODModel *m = (LODModel *)(p->head); m; m = (LODModel *)(m->chain_next))
	{
		if (!m->count) continue;
		if (!_stricmp(title, m->model[0]->title)) break;
	}

	return m;
}


static Model *dest_model = NULL;
static bool use_dest_model = false;

void LODModel_Calculate(LODModel *m, float w)
{
	int i;

	use_dest_model = false;

	if (m->count == 1)
	{
		use_dest_model = true;
		dest_model = m->model[0];
		return;
	}

	scalar_t invW = 1.f / w;
	
	for (i = 1; i < m->count && invW > m->invWidth[i]; i++);
	
	dest_model = m->model[i-1];
	//width[0] = fInf	->	0.0
	//		 w = fInf	->	0.0		-> 0%
	//		 w = 0.75	->	1.333	-> 67%
	//		 w = 0.5	->	2.0		-> 100%
	//width[1] = 0.5	->	2.0

	if (!menu_lod_interpolate)
	{
		use_dest_model = true;
		return;
	}

	if (dest_model->vtxCount > 8192)
	{
		main_console->Print("we got a model with too many verticies: %d\n", dest_model->vtxCount);
		return;
	}

	scalar_t prevwidth = m->invWidth[i-1];
	scalar_t nextwidth = m->invWidth[i];
	scalar_t frac = (invW - prevwidth) / (nextwidth - prevwidth);
	scalar_t ifrac = 1.f - frac;

	//first - interpolate the verticies (save on computing time)
	float *tvb = tempvertexbuffer;
	int di;
	for (i = 0; i < dest_model->vtxCount; i++)
	{
		di = dest_model->collapse[i];
		tvb[0] = dest_model->vtx[i].x * ifrac + dest_model->vtx[di].x * frac;
		tvb[1] = dest_model->vtx[i].y * ifrac + dest_model->vtx[di].y * frac;
		tvb[2] = dest_model->vtx[i].z * ifrac + dest_model->vtx[di].z * frac;
		tvb[3] = dest_model->normal[i].x * ifrac + dest_model->normal[di].x * frac;
		tvb[4] = dest_model->normal[i].y * ifrac + dest_model->normal[di].y * frac;
		tvb[5] = dest_model->normal[i].z * ifrac + dest_model->normal[di].z * frac;
		tvb[6] = dest_model->texCoord[i<<1] * ifrac + dest_model->texCoord[di<<1] * frac;
		tvb[7] = dest_model->texCoord[(i<<1)+1] * ifrac + dest_model->texCoord[(di<<1)+1] * frac;
		tvb += 8;
	}
}

void LODModel_Display(void)
{
	if (use_dest_model)
	{
		dest_model->Display();
		return;
	}

	float *tvb = tempvertexbuffer;
	int di = 3*dest_model->triCount;
	glBegin(GL_TRIANGLES);
	for (int i = 0; i < di; i++)
	{
		tvb = tempvertexbuffer + (dest_model->tri[i] << 3);
		glTexCoord2fv(&tvb[6]);
		glNormal3fv(&tvb[3]);
		glVertex3fv(tvb);
	}
	glEnd();
}
