/*

    File: Lighting.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>

#include "../m_math3d/quat.h"
#include "../m_foundation/main.h"
#include "../m_foundation/random.h"

#include "lighting.h"

static int max_lights_at_once = 0;		//max lights to display in the BSP tree
static int max_lights = 8;				//size of array

typedef struct light_s {
	//dirty hack - make sure 'info' is 
	lightinfo_t		info;
	bool			active;
	Chain			*treeLink;
} light_t;

static light_t *light = NULL;

static quat_t light_ambient_level;

void CFunc_LightAmbientLevel(int, char**)
{
	glMaterialfv(GL_FRONT, GL_AMBIENT, light_ambient_level.v);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_ambient_level.v);
}

static int light_separate_specular = 1;
void CFunc_LightSeparateSpecular(int,char**)
{
	//this extension is so simple and useful that all cards should have it...
	//...but they dont
	if (found_GL_EXT_separate_specular_color)
	{
		if (light_separate_specular)
			glLightModelf(GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SEPARATE_SPECULAR_COLOR_EXT);
		else
			glLightModelf(GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SINGLE_COLOR_EXT);
	}
}

static lightinfo_t default_lightinfo;
static int light_frame_counter;

bool Init_Lighting(void)
{
	static float fv0001[4] = {0,0,0,1};
	static float fv1111[4] = {1,1,1,1};

	QuatSet(&light_ambient_level, 0.05f, 0.05f, 0.05f, 1.0f);

	//set up the default light structure
	default_lightinfo.position[0] = 0;
	default_lightinfo.position[1] = 0;
	default_lightinfo.position[2] = 0;
	default_lightinfo.position[3] = 1;
	default_lightinfo.diffuse[0] = 1;
	default_lightinfo.diffuse[1] = 1;
	default_lightinfo.diffuse[2] = 1;
	default_lightinfo.diffuse[3] = 1;
	default_lightinfo.attenuation[0] = 1;
	default_lightinfo.attenuation[1] = 0;
	default_lightinfo.attenuation[2] = 0;
	default_lightinfo.distance = -1;

	//get the max lights
	glGetIntegerv(GL_MAX_LIGHTS, &max_lights_at_once);
	main_console->Print("Init_Lighting(): max_lights_at_once %d\n", max_lights_at_once);

	max_lights = max_lights_at_once;

	//allocate the lights
	light = (light_t *)malloc(sizeof(light_t) * max_lights);
	if (!light) return false;

	Reset_Lighting();

	//init opengl lighting *here*

	//light model init
	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_ambient_level.v);
	CFunc_LightSeparateSpecular(0,0);

	//material init
//	glMaterialf(GL_FRONT,  GL_SHININESS, 128);						//updated per material
	glMaterialfv(GL_FRONT, GL_AMBIENT, light_ambient_level.v);		//keep me close to zero
//	glMaterialfv(GL_FRONT, GL_DIFFUSE, fv1111);						//updated per material - doesnt matter
//	glMaterialfv(GL_FRONT, GL_SPECULAR, fv1111);					//updated per material - doesnt matter
	glMaterialfv(GL_FRONT, GL_EMISSION, fv0001);					//keep me at zero (ambient level?)

	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	//light specific init
	for (int i = 0; i < max_lights; i++)
	{
		int gl_light = GL_LIGHT0 + i;
		glLightfv(gl_light, GL_AMBIENT, fv1111);
		glLightfv(gl_light, GL_SPECULAR, fv1111);
	}

	new CVarw(&light_ambient_level,		"LightAmbientLevel",		CFunc_LightAmbientLevel);
	new CVarw(&light_separate_specular,	"LightSeparateSpecular",	CFunc_LightSeparateSpecular);

	return true;
}

void Shutdown_Lighting(void)
{
	if (light) free(light);
	light = NULL;
}

void Reset_Lighting(void)
{
	for (int i = 0; i < max_lights; i++)
		memset(light, 0, sizeof(light_t) * max_lights);
}

void Update_Lighting(void)
{
	static bool useLightCounter = false;
	static CVarw *fdjsdsfl = new CVarw(&useLightCounter, "lightCounter");
	if (useLightCounter)
		main_console->Print("current lights this frame: %d\n", light_frame_counter);
	light_frame_counter = 0;

	for (int i = 0; i < max_lights; i++)
	{
		int gl_light = GL_LIGHT0 + i;

		if (!light[i].active)
		{
			glDisable(gl_light);
		}
		else
		{
			glEnable (gl_light);
			glLightf (gl_light, GL_CONSTANT_ATTENUATION,	light[i].info.attenuation[0]);
			glLightf (gl_light, GL_LINEAR_ATTENUATION,		light[i].info.attenuation[1]);
			glLightf (gl_light, GL_QUADRATIC_ATTENUATION,	light[i].info.attenuation[2]);
			glLightfv(gl_light, GL_DIFFUSE,					light[i].info.diffuse);
			glLightfv(gl_light, GL_POSITION,				light[i].info.position);
		}
	}
}

lightinfo_t *Create_Light(void)
{
	light_frame_counter++;

	light_t *dest = NULL;

	for (int i = 0; i < max_lights; i++)
	{
		//find an inactive light
		if (!light[i].active)
		{
			//activate the light
			light[i].active = true;
			dest = &light[i];
			break;
		}
	}

	//if no inactive lights were found then just use an old one
	if (!dest) dest = &light[MRand() % max_lights];

	//default the lightinfo contents and return them
	lightinfo_t *info = &dest->info;
	memcpy(info, &default_lightinfo, sizeof(lightinfo_t));
	return info;
}

void Register_Light(lightinfo_t *li)
{
	//get the light using ID's dirty C pointer hack
	light_t *l = (light_t *)li;

	//link it to the BSP tree... whenever you feel like it...
}