/*

    File: Menu.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include "main.h"
#include "../m_system3d/main.h"
#include "../m_foundation/random.h"
#include "../m_foundation/main.h"

//all the menu variables, all lined up:
bool menu_sky_enable = true;
bool menu_sky_cube_map = true;
bool menu_ent_cube_map = true;
bool menu_lod_interpolate = true;
bool menu_ent_lighting = true;
bool menu_particle_enable = true;
bool menu_starfield_enable = true;


#define MENU_TEXT_WIDTH		50
#define MENU_TEXT_HEIGHT	25

void Open_Menu(void);
void Close_Menu(void);

void Menu_DrawString(char *p, int x, int y)
{
	int a,tx,ty;
	for (a = 0; *p; a++, p++)
	{
		tx = (*(unsigned char *)p & 15);
		ty = (*(unsigned char *)p >> 4) - 2;

		glBegin(GL_QUADS);
		glTexCoord2f((float)tx/16.f, (float)(ty)/16.f);			glVertex2i(x+a,		y);
		glTexCoord2f((float)tx/16.f, (float)(ty+1)/16.f);		glVertex2i(x+a,		y+1);
		glTexCoord2f((float)(tx+1)/16.f, (float)(ty+1)/16.f);	glVertex2i(x+a+1,	y+1);
		glTexCoord2f((float)(tx+1)/16.f, (float)(ty)/16.f);		glVertex2i(x+a+1,	y);
		glEnd();
	}
}

enum menuType {TYPE_STATIC, TYPE_PARENT, TYPE_FUNCTION, TYPE_CHECK};

#define FLAG_OPEN		1		//if a parent - whether the children can be seen

static int menu_X = 1;
static int menu_Y = 1;

class M_Menu {
public:

	M_Menu(char *t, void *d, menuType i)
	{
		title = t;
		title_length = strlen(t);
		data = d;
		type = i;
		flags = 0;
	}

	void SetOpen(bool open)
	{
		flags = open ? FLAG_OPEN : 0;
	}

	void ToggleOpen(void)
	{
		SetOpen(!(flags & FLAG_OPEN));
	}

	void Display(void)
	{
		//store menu coords for input - calc'd after rendering
		//i know i know, this is a messed up system -
		// - i should precalc and not make input dependant upon rendering
		position[0] = (int)menu_X;
		position[1] = (int)menu_Y;

		//draw the menu
		Menu_DrawString(title, menu_X, menu_Y);
		menu_Y++;

		switch (type)
		{
		case TYPE_CHECK:
			if (data)
			{
				if (*(bool *)data)	Menu_DrawString("X", 45, menu_Y-1);
				else				Menu_DrawString("_", 45, menu_Y-1);
			}
			break;

		case TYPE_PARENT:
		
			//offset and draw the children
			if ((flags & FLAG_OPEN) && data)
			{
				menu_X++;
				for (M_Menu **m = (M_Menu **)data; *m; m++)
					(*m)->Display();
				menu_X--;
			}
			break;
		}
	}

	void CheckInput(int x, int y)
	{
		if (x >= position[0] &&
			x < position[0] + title_length &&
			y == position[1])
		{
			//do action
			//	either open the menu
			if (type == TYPE_PARENT)
				ToggleOpen();
			//or execute the function
			else if (type == TYPE_FUNCTION)
			{
				//what a statement...
				((void (*)(void))data)();
			}
			else if (type == TYPE_CHECK)
			{
				if (data) *(bool *)data = !(*(bool *)data);
			}
		}

		//check children
		if ((type == TYPE_PARENT) &&
			(flags & FLAG_OPEN) &&
			(data != NULL))
		{
			for (M_Menu **m = (M_Menu **)data; *m; m++)
				(*m)->CheckInput(x,y);
		}
	}

private:
	char		*title;
	int			title_length;
	void		*data;
	menuType	type;
	int			flags;			//so far only used for 
	int			position[2];
};

//options -> graphics
static M_Menu *menu_Graphics_data[] =
{
//world sky
	new M_Menu("Enable Sky Map", &menu_sky_enable, TYPE_CHECK),
	new M_Menu("-Use Cube Mapping", &menu_sky_cube_map, TYPE_CHECK),

//single textures on all the models
	new M_Menu("Maximum Texture Layers", NULL, TYPE_STATIC),
	new M_Menu("-Use Cube Mapping", &menu_ent_cube_map, TYPE_CHECK),

//LOD tweaking
	new M_Menu("Interpolate LOD Models", &menu_lod_interpolate, TYPE_CHECK),
	new M_Menu("Adjust LOD Detail", NULL, TYPE_STATIC),

//lighting
	new M_Menu("Enable Lighting", &menu_ent_lighting, TYPE_CHECK),
	new M_Menu("Max Lights", NULL, TYPE_STATIC),

//particle system
	new M_Menu("Enable Particle System", &menu_particle_enable, TYPE_CHECK),
	new M_Menu("-Maximum Particles", NULL, TYPE_STATIC),
	new M_Menu("-Particle Density", NULL, TYPE_STATIC),

//starfield particle effect
	new M_Menu("Enable Starfield", &menu_starfield_enable, TYPE_CHECK),
	new M_Menu("-Starfield Density", NULL, TYPE_STATIC),
	new M_Menu("-Starfield Distance", NULL, TYPE_STATIC),
	NULL
};

//options -> game play
//	modulo space - max range?  or should this be level-specific?

//new game
//instead of a gamestate variable, this should probably just send a 'restart' msg somewhere
static void MenuFunc_NewGame(void)
{
	Close_Menu();
	game_State = GAMESTATE_NEWGAME;
}

//new game
//instead of a gamestate variable, this should probably just send a 'restart' msg somewhere
static void MenuFunc_ResumeGame(void)
{
	Close_Menu();
	game_State = GAMESTATE_BRAWL;
}

//options
static M_Menu *menu_Options_data[] =
{
	new M_Menu("Graphics", menu_Graphics_data, TYPE_PARENT),
	new M_Menu("Game Play", NULL, TYPE_STATIC),
	NULL
};

//exit
static void MenuFunc_Exit(void) {RequestExit();}

//main
static M_Menu *menu_Main_data[] =
{
	new M_Menu("New Game", MenuFunc_NewGame, TYPE_FUNCTION),
	new M_Menu("Resume Game", MenuFunc_ResumeGame, TYPE_FUNCTION),
	new M_Menu("Options", menu_Options_data, TYPE_PARENT),
	new M_Menu("Exit", MenuFunc_Exit, TYPE_FUNCTION),
	NULL
};

static M_Menu menu_Main("Main", menu_Main_data, TYPE_PARENT);

bool Init_Menu(void)
{
	Open_Menu();

	return true;
}

void Shutdown_Menu(void)
{
}

//and this is just a mess...
static bool menu_Active = true;

void Open_Menu(void)
{
	menu_Active = true;
}

void Close_Menu(void)
{
	menu_Active = false;
}

void Toggle_Menu(void)
{
	if (!menu_Active)	Open_Menu();
	else				Close_Menu();
}

void Update_Menu(void)
{
	if (menu_Active)
	{
		glDisable(GL_COLOR_MATERIAL);
		glDisable(GL_DEPTH_TEST);

		//init menu OGL transform
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, MENU_TEXT_WIDTH, MENU_TEXT_HEIGHT,0,-1000,1000);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		//tell the world this is the menu: draw a big black box over it allg

		glColor4f(0,0,0,0.65f);
		glDisable(GL_TEXTURE_2D);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glBegin(GL_QUADS);
		glVertex2i(0,0);
		glVertex2i(0,MENU_TEXT_HEIGHT);
		glVertex2i(MENU_TEXT_WIDTH,MENU_TEXT_HEIGHT);
		glVertex2i(MENU_TEXT_WIDTH,0);
		glEnd();
		glDisable(GL_BLEND);
		glEnable(GL_TEXTURE_2D);

		glColor3f(1,1,1);
		if (main_console->font) main_console->font->Set();

		//(recursively) draw the menus displayed here
		menu_X = 1;
		menu_Y = 1;
		menu_Main.Display();

		//kill menu OGL transform
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);

		//trap whatever input you want
		if (main_mouse->buttonState[0] == MH_RELEASE)
		{
			menu_Main.CheckInput(
				(int)(main_mouse->mouseSX * MENU_TEXT_WIDTH),
				(int)((1.f - main_mouse->mouseSY) * MENU_TEXT_HEIGHT));
		}

		glEnable(GL_COLOR_MATERIAL);
		glEnable(GL_DEPTH_TEST);
	}

}
