/*

    File: Model.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <stdio.h>

#include "../m_foundation/main.h"
#include "../m_foundation/const.h"
#include "../m_foundation/opengl.h"

#include "model.h"

void Model::Reset(void)
{
	triCount = 0;
	vtxCount = 0;
	
	tri = NULL;
	vtx = NULL;
	texCoord = NULL;
	normal = NULL;
	collapse = NULL;

	compiled = false;
}

bool Model::Load(char *filename)
{
	//CUSTOM MDL FILE FORMAT
	Unload();

	int t,v;

	strcpy(title, filename);

	char sig[4];

	FILE *File = fopen(filename, "rb");
	if (File == NULL)
	{
		main_console->Print("%s: unable to open model file\n", filename);
		return false;
	}

	fread(sig, 1, 4, File);
	if (memcmp(sig, "MMDL", 4))
	{
		main_console->Print("%s: bad signature on file\n", filename);
		return false;
	}
	fread(&vtxCount, sizeof(int), 1, File);
	fread(&triCount, sizeof(int), 1, File);

	if (vtxCount < 0)
	{
		main_console->Print("%s: model has bad vertex count: %d\n", filename, vtxCount);
		fclose(File);
		return false;
	}
	if (triCount < 0)
	{
		main_console->Print("%s: model has bad triangle count: %d\n", filename, triCount);
		fclose(File);
		return false;
	}

	if ((vtx = (vector_t *)malloc(sizeof(vector_t) * vtxCount))==NULL)
	{
		main_console->Print("%s: unable to allocate vertex array\n", filename);
		fclose(File);
		return false;
	}
	if ((texCoord = (float *)malloc(sizeof(float) * 2 * vtxCount))==NULL)
	{
		main_console->Print("%s: unable to allocate texture coordinate array\n", filename);
		fclose(File);
		return false;
	}
	if ((normal = (vector_t *)malloc(sizeof(vector_t) * vtxCount))==NULL)
	{
		main_console->Print("%s: unable to allocate normal array\n", filename);
		fclose(File);
		return false;
	}
	if ((tri = (int *)malloc(sizeof(int) * 3 * triCount))==NULL)
	{
		main_console->Print("%s: unable to allocate triangle array\n", filename);
		fclose(File);
		return false;
	}

	if (fread(vtx, sizeof(vector_t), vtxCount, File) != (size_t)(vtxCount))
	{
		main_console->Print("%s: unable to read vertex array\n", filename);
		fclose(File);
		return false;
	}
	for (v = 0; v < vtxCount; v++)
	{
		if (fread(&texCoord[v*2], sizeof(float), 2, File) != 2)
		{
			main_console->Print("%s: unable to read texture coordinate array\n", filename);
			fclose(File);
			return false;
		}
	}
	if (fread(tri, sizeof(int), triCount*3, File) != (size_t)(triCount*3))
	{
		main_console->Print("%s: unable to read triangle array\n", filename);
		fclose(File);
		return false;
	}

	if ((collapse = (int *)malloc(sizeof(int) * vtxCount))==NULL)
	{
		main_console->Print("%s: unable to allocate collapse table\n", filename);
		fclose(File);
		return false;
	}
	for (v = 0; v < vtxCount; v++)
	{
		collapse[v] = v;
	}
	//if it doesnt read correctly - we already have the default loaded
	fread(collapse, sizeof(int), vtxCount, File);

	fclose(File);

	//flip tris
	for (t = 0; t < triCount*3; t+=3)
	{
		int tmp = tri[t];
		tri[t] = tri[t+2];
		tri[t+2] = tmp;
	}

	CalcNormals();
	CalcRange();

	return true;
}

void Model::CalcNormals(void)
{
	//calc normals
	for (int v = 0; v < vtxCount; v++)
		VectorSet(&normal[v], 0,0,0);

	//flip tris
	for (int t = 0; t < triCount*3; t+=3)
	{
		vector_t n;

		VectorPlaneUnitNormal(&vtx[tri[t+2]], &vtx[tri[t+1]], &vtx[tri[t]], &n);

		for (int i = 0; i < 3; i++)
			VectorAdd(&normal[tri[t+i]], &n, &normal[tri[t+i]]);
	}

	//normalize normals
	for (v = 0; v < vtxCount; v++)
		VectorUnit(&normal[v], &normal[v]);
}

void Model::CalcRange(void)
{
	VectorSet(&bbox.min, fInf,fInf,fInf);
	VectorSet(&bbox.max, -fInf,-fInf,-fInf);
	radius = -1;

	for (int v = 0; v < vtxCount; v++)
	{
		//and stretch bbox
		BoxStretch(&bbox, &vtx[v]);

		float l = VectorLength(&vtx[v]);
		if (l > radius) radius = l;
	}
}

void Model::Unload(void)
{
	if (tri) free(tri);
	if (vtx) free(vtx);
	if (texCoord) free(texCoord);
	if (normal) free(normal);

	if (compiled) glDeleteLists(list_id, 1); 

	Reset();
}

#include "../m_foundation/opengl.h"

void Model::Display(void)
{
	if (compiled)
	{
		glCallList(list_id);
	}
	else
	{
		glVertexPointer(3, GL_FLOAT, 0, vtx);
		glTexCoordPointer(2, GL_FLOAT, 0, texCoord);
		glNormalPointer(GL_FLOAT, 0, normal);

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);

		if (found_GL_EXT_compiled_vertex_array)
			glLockArraysEXT(0, vtxCount);

		if (found_GL_EXT_draw_range_elements)
			glDrawRangeElements(GL_TRIANGLES, 0, vtxCount-1, 3*triCount, GL_UNSIGNED_INT, tri);
		else
			glDrawElements(GL_TRIANGLES, 3*triCount, GL_UNSIGNED_INT, tri);

		if (found_GL_EXT_compiled_vertex_array)
			glUnlockArraysEXT();

		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
	}
}

void Model::Compile(void)
{
	if (compiled) return;

	list_id = glGenLists(1); 

	glNewList(list_id, GL_COMPILE);

	glBegin(GL_TRIANGLES);
	for (int i = 0; i < triCount*3; i++)
	{
		int v = tri[i];
		glTexCoord2fv(&texCoord[v*2]);
		glNormal3fv(normal[v].v);
		glVertex3fv(vtx[v].v);
	}
	glEnd();

	glEndList();

	compiled = true;
}

bool Model::Copy(Model *src)
{
	Unload();	//make sure we're clear

	//vertex data
	vtxCount = src->vtxCount;
	if ((vtx = (vector_t *)malloc(sizeof(vector_t) * vtxCount))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(vtx, src->vtx, sizeof(vector_t) * vtxCount);

	if ((texCoord = (float *)malloc(2 * sizeof(float) * vtxCount))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(texCoord, src->texCoord, 2 * sizeof(float) * vtxCount);

	if ((normal = (vector_t *)malloc(sizeof(vector_t) * vtxCount))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(normal, src->normal, sizeof(vector_t) * vtxCount);

	//element data
	triCount = src->triCount;
	if ((tri = (int *)malloc(sizeof(int) * 3 * triCount))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(tri, src->tri, sizeof(int) * 3 * triCount);

	//title
	memcpy(title, src->title, sizeof(title));

	//and range info
	VectorCopy(&src->bbox.min, &bbox.min);
	VectorCopy(&src->bbox.max, &bbox.max);

	radius = src->radius;

	return true;
}



Model *GetModel(LinkedList *p, char *title)
{
	for (Model *m = (Model *)(p->head); m; m = (Model *)(m->chain_next))
		if (!_stricmp(title, m->title)) break;

	return m;
}
