/*

    File: Particle.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>

#include "particle.h"
#include "main.h"

#include "../m_foundation/main.h"
#include "../m_math3d/vector.h"
#include "../m_foundation/chain.h"
#include "../m_foundation/types.h"
#include "../m_foundation/texture.h"

#include "menu.h"

class Particle : public Chain {
public:
	vector_t	position, velocity;
	color_t		color;
	scalar_t	size;
	bool		fade;
	int			life, max_life;

	//rendering specifics
	Texture		*texture;
	texcoord_t	blendfunc;

	//z-sorting binary tree
	scalar_t	dist;
	Particle	*child[2];

	//physics integrator
	bool		updatePhysics;
};

static Particle *particle;
static int particle_max;		//holds the max
static texcoord_t last_blendfunc;
static Texture *last_texture;

static LinkedList particle_liveChain;
static LinkedList particle_deadChain;

static int particle_counter;

static void Reset_ParticleSystem(void)
{
	particle = NULL;
	particle_max = 0;

	last_blendfunc.st[0] = GL_SRC_ALPHA;
	last_blendfunc.st[1] = GL_ONE;

	last_texture = NULL;
}

bool Init_ParticleSystem(int max)
{
	Reset_ParticleSystem();

	//set the count
	particle_max = max;
	
	//allocate the particle array
	particle = (Particle *)malloc(sizeof(Particle) * particle_max);
	if (!particle) return false;

	for (int a = 0; a < particle_max; a++)
	{
		memset(&particle[a], 0, sizeof(Particle));	//zero it
		particle_deadChain.AddLast(&particle[a]);				//link it
	}

	return true;
}

void Shutdown_ParticleSystem(void)
{
	//unload
	particle_liveChain.UnlinkAll();
	particle_deadChain.UnlinkAll();

	if (particle) free(particle);

	//and reset
	Reset_ParticleSystem();
}

void Particle_SetBlend(unsigned short src, unsigned short dest)
{
	last_blendfunc.st[0] = src;
	last_blendfunc.st[1] = dest;
}

void Particle_SetTexture(Texture *t)
{
	last_texture = t;
}

void Update_ParticleSystem(void)
{
	Particle *next;

particle_counter = 0;

	for (Particle *p = (Particle *)particle_liveChain.head; p; p = next)
	{
particle_counter++;
		next = (Particle *)p->chain_next;

		p->life--;
		if (p->life <= 0)
		{
			p->Unlink();
			particle_deadChain.AddLast(p);
			continue;
		}

		if (p->updatePhysics)
			VectorAdd(&p->position, &p->velocity, &p->position);

	}
}

static void LinkParticle(Particle *p, Particle *node)
{
	int ch = p->dist > node->dist;

	if (node->child[ch])
		LinkParticle(p, node->child[ch]);
	else
	{
		node->child[ch] = p;
		p->child[0] = p->child[1] = NULL;
	}
}

static void DisplayParticle(Particle *p)
{
	if (!menu_particle_enable) return;

	if (p->child[0]) DisplayParticle(p->child[0]);

	scalar_t size = p->size * 0.5f;
	color_t color;

	color.l = p->color.l;

	if (p->fade)
	{
		color.rgba[3] = (unsigned char)( (float)p->color.rgba[3] * (float)p->life / (float)p->max_life);
	}

	
//	SetTexture( (p->texture) );		//whats wrong here? if i copy the code out of SetTexture then things work fine...
	glBindTexture(GL_TEXTURE_2D, p->texture ? p->texture->texture_id : 0);

	glColor4ubv(color.rgba);

	glBlendFunc(p->blendfunc.st[0], p->blendfunc.st[1]);

	glBegin(GL_TRIANGLE_STRIP);
	glTexCoord2f(0,0);

#define VIEW_LEFT	main_camera->matrix.vec[1]
#define VIEW_UP		main_camera->matrix.vec[2]
	glVertex3d(	p->position.x - size * (VIEW_LEFT.x + VIEW_UP.x),
				p->position.y - size * (VIEW_LEFT.y + VIEW_UP.y),
				p->position.z - size * (VIEW_LEFT.z + VIEW_UP.z));
	glTexCoord2f(1,0);
	glVertex3d(	p->position.x - size * (VIEW_LEFT.x - VIEW_UP.x),
				p->position.y - size * (VIEW_LEFT.y - VIEW_UP.y),
				p->position.z - size * (VIEW_LEFT.z - VIEW_UP.z));
	glTexCoord2f(0,1);
	glVertex3d(	p->position.x + size * (VIEW_LEFT.x - VIEW_UP.x),
				p->position.y + size * (VIEW_LEFT.y - VIEW_UP.y),
				p->position.z + size * (VIEW_LEFT.z - VIEW_UP.z));
	glTexCoord2f(1,1);
	glVertex3d(	p->position.x + size * (VIEW_LEFT.x + VIEW_UP.x),
				p->position.y + size * (VIEW_LEFT.y + VIEW_UP.y),
				p->position.z + size * (VIEW_LEFT.z + VIEW_UP.z));
	glEnd();

	if (p->child[1]) DisplayParticle(p->child[1]);
}

void Display_ParticleSystem(void)
{
	Particle *p, *root;

	root = (Particle *)particle_liveChain.head;
	if (!root) return;

	for (p = root; p; p = (Particle *)(p->chain_next))
	{
		vector_t del;
		//reposition each particle to be closest to the player
			//calc delta
		world->Delta(&p->position, &main_camera->position, &del);
			//calc depth
		p->dist = VectorDot(&del, &main_camera->matrix.vec[0]);
			//add delta
		VectorAdd(&main_camera->position, &del, &p->position);
	}

	//build binary tree
	p = root;
	p->child[0] = p->child[1] = NULL;
	for (p = (Particle *)(p->chain_next); p; p = (Particle *)(p->chain_next))
		if (p->dist > main_camera->znear)		//more extensive frustum test?
			LinkParticle(p, (Particle *)particle_liveChain.head);

	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glDepthMask(GL_FALSE);

	DisplayParticle(root);

	glPopAttrib();
}

static Particle *NewParticle(void)
{
	Particle *p = (Particle *)particle_deadChain.tail;
	if (p)
	{
		p->Unlink();
		particle_liveChain.AddLast(p);
	}
	return p;
}

void Particle_Create(const vector_t *pos, const vector_t *vel, scalar_t size, int color, int life, bool fade)
{
	if (!menu_particle_enable) return;

	Particle *p;
	p = NewParticle();
	if (!p) return;

	VectorCopy(pos, &p->position);

	p->updatePhysics = (vel->x || vel->y || vel->z);
	if (p->updatePhysics) VectorCopy(vel, &p->velocity);

	p->size = size;
	p->color.l = color;
	p->life = p->max_life = life;

	p->fade = fade;

	p->blendfunc.l = last_blendfunc.l;
	p->texture = last_texture;
}

int Particle_Counter(void)
{
	return particle_counter;
}